@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: bounce.c,v 1.64 2014/04/19 17:27:40 gilles Exp $	*/
 
 /*
  * Copyright (c) 2009 Gilles Chehade <gilles@poolp.org>
@@ -37,6 +37,7 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <limits.h>
 
 #include "smtpd.h"
 #include "log.h"
@@ -59,7 +60,10 @@ enum {
 struct bounce_envelope {
 	TAILQ_ENTRY(bounce_envelope)	 entry;
 	uint64_t			 id;
+	struct mailaddr			 dest;
 	char				*report;
+	uint8_t				 esc_class;
+	uint8_t				 esc_code;
 };
 
 struct bounce_message {
@@ -80,6 +84,7 @@ struct bounce_session {
 	int				 state;
 	struct iobuf			 iobuf;
 	struct io			 io;
+	uint64_t			 boundary;
 };
 
 SPLAY_HEAD(bounce_message_tree, bounce_message);
@@ -98,6 +103,7 @@ static void bounce_io(struct io *, int);
 static void bounce_timeout(int, short, void *);
 static void bounce_free(struct bounce_session *);
 static const char *bounce_strtype(enum bounce_type);
+static const char *action_str(const struct delivery_bounce *);
 
 static struct tree			wait_fd;
 static struct bounce_message_tree	messages;
@@ -124,7 +130,7 @@ bounce_init(void)
 void
 bounce_add(uint64_t evpid)
 {
-	char			 buf[SMTPD_MAXLINESIZE], *line;
+	char			 buf[LINE_MAX], *line;
 	struct envelope		 evp;
 	struct bounce_message	 key, *msg;
 	struct bounce_envelope	*be;
@@ -132,7 +138,7 @@ bounce_add(uint64_t evpid)
 	bounce_init();
 
 	if (queue_envelope_load(evpid, &evp) == 0) {
-		m_create(p_scheduler, IMSG_DELIVERY_PERMFAIL, 0, 0, -1);
+		m_create(p_scheduler, IMSG_QUEUE_DELIVERY_PERMFAIL, 0, 0, -1);
 		m_add_evpid(p_scheduler, evpid);
 		m_close(p_scheduler);
 		return;
@@ -145,6 +151,16 @@ bounce_add(uint64_t evpid)
 	key.msgid = evpid_to_msgid(evpid);
 	key.bounce = evp.agent.bounce;
 	key.smtpname = evp.smtpname;
+
+	if (evp.errorline[0] == '4')
+		key.bounce.type = B_WARNING;
+	else if (evp.errorline[0] == '5')
+		key.bounce.type = B_ERROR;
+	else
+		key.bounce.type = B_DSN;
+
+	key.bounce.dsn_ret = evp.dsn_ret;
+	key.bounce.expire = evp.expire;
 	msg = SPLAY_FIND(bounce_message_tree, &messages, &key);
 	if (msg == NULL) {
 		msg = xcalloc(1, sizeof(*msg), "bounce_add");
@@ -154,7 +170,7 @@ bounce_add(uint64_t evpid)
 		TAILQ_INIT(&msg->envelopes);
 
 		msg->smtpname = xstrdup(evp.smtpname, "bounce_add");
-		snprintf(buf, sizeof(buf), "%s@%s", evp.sender.user,
+		(void)snprintf(buf, sizeof(buf), "%s@%s", evp.sender.user,
 		    evp.sender.domain);
 		msg->to = xstrdup(buf, "bounce_add");
 		nmessage += 1;
@@ -168,12 +184,17 @@ bounce_add(uint64_t evpid)
 	line = evp.errorline;
 	if (strlen(line) > 4 && (*line == '1' || *line == '6'))
 		line += 4;
-	snprintf(buf, sizeof(buf), "%s@%s: %s\n", evp.dest.user,
+	(void)snprintf(buf, sizeof(buf), "%s@%s: %s\n", evp.dest.user,
 	    evp.dest.domain, line);
 
 	be = xmalloc(sizeof *be, "bounce_add");
 	be->id = evpid;
 	be->report = xstrdup(buf, "bounce_add");
+	(void)strlcpy(be->dest.user, evp.dest.user, sizeof(be->dest.user));
+	(void)strlcpy(be->dest.domain, evp.dest.domain,
+	    sizeof(be->dest.domain));
+	be->esc_class = evp.esc_class;
+	be->esc_code = evp.esc_code;
 	TAILQ_INSERT_TAIL(&msg->envelopes, be, entry);
 	buf[strcspn(buf, "\n")] = '\0';
 	log_debug("debug: bounce: adding report %16"PRIx64": %s", be->id, buf);
@@ -211,6 +232,7 @@ bounce_fd(int fd)
 	io_init(&s->io, fd, s, bounce_io, &s->iobuf);
 	io_set_timeout(&s->io, 30000);
 	io_set_read(&s->io);
+	s->boundary = generate_uid();
 
 	log_debug("debug: bounce: new session %p", s);
 	stat_increment("bounce.session", 1);
@@ -268,7 +290,7 @@ bounce_drain()
 		}
 
 		log_debug("debug: bounce: requesting new enqueue socket...");
-		m_compose(p_smtp, IMSG_SMTP_ENQUEUE_FD, 0, 0, -1, NULL, 0);
+		m_compose(p_pony, IMSG_QUEUE_SMTP_SESSION, 0, 0, -1, NULL, 0);
 
 		running += 1;
 	}
@@ -294,25 +316,30 @@ bounce_send(struct bounce_session *s, const char *fmt, ...)
 }
 
 static const char *
-bounce_duration(long long int d) {
+bounce_duration(long long int d)
+{
 	static char buf[32];
 
 	if (d < 60) {
-		snprintf(buf, sizeof buf, "%lld second%s", d, (d == 1)?"":"s");
+		(void)snprintf(buf, sizeof buf, "%lld second%s", d,
+		    (d == 1)?"":"s");
 	} else if (d < 3600) {
 		d = d / 60;
-		snprintf(buf, sizeof buf, "%lld minute%s", d, (d == 1)?"":"s");
+		(void)snprintf(buf, sizeof buf, "%lld minute%s", d,
+		    (d == 1)?"":"s");
 	}
 	else if (d < 3600 * 24) {
 		d = d / 3600;
-		snprintf(buf, sizeof buf, "%lld hour%s", d, (d == 1)?"":"s");
+		(void)snprintf(buf, sizeof buf, "%lld hour%s", d,
+		    (d == 1)?"":"s");
 	}
 	else {
 		d = d / (3600 * 24);
-		snprintf(buf, sizeof buf, "%lld day%s", d, (d == 1)?"":"s");
+		(void)snprintf(buf, sizeof buf, "%lld day%s", d,
+		    (d == 1)?"":"s");
 	}
 	return (buf);
-};
+}
 
 #define NOTICE_INTRO							    \
 	"    Hi!\n\n"							    \
@@ -341,7 +368,7 @@ static int
 bounce_next_message(struct bounce_session *s)
 {
 	struct bounce_message	*msg;
-	char			 buf[SMTPD_MAXLINESIZE];
+	char			 buf[LINE_MAX];
 	int			 fd;
 	time_t			 now;
 
@@ -363,16 +390,16 @@ bounce_next_message(struct bounce_session *s)
 	SPLAY_REMOVE(bounce_message_tree, &messages, msg);
 
 	if ((fd = queue_message_fd_r(msg->msgid)) == -1) {
-		bounce_delivery(msg, IMSG_DELIVERY_TEMPFAIL,
+		bounce_delivery(msg, IMSG_QUEUE_DELIVERY_TEMPFAIL,
 		    "Could not open message fd");
 		goto again;		
 	}
 
 	if ((s->msgfp = fdopen(fd, "r")) == NULL) {
-		snprintf(buf, sizeof(buf), "fdopen: %s", strerror(errno));
+		(void)snprintf(buf, sizeof(buf), "fdopen: %s", strerror(errno));
 		log_warn("warn: bounce: fdopen");
 		close(fd);
-		bounce_delivery(msg, IMSG_DELIVERY_TEMPFAIL, buf);
+		bounce_delivery(msg, IMSG_QUEUE_DELIVERY_TEMPFAIL, buf);
 		goto again;
 	}
 
@@ -427,13 +454,22 @@ bounce_next(struct bounce_session *s)
 		    "To: %s\n"
 		    "Date: %s\n"
 		    "\n"
-		    NOTICE_INTRO
+		    "This is a MIME-encapsulated message.\n"
 		    "\n",
 		    bounce_strtype(s->msg->bounce.type),
 		    s->smtpname,
 		    s->msg->to,
 		    time_to_text(time(NULL)));
 
+		iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
+		    "--%16" PRIu64 "/%s\n"
+		    "Content-Description: Notification\n"
+		    "Content-Type: text/plain; charset=us-ascii\n"
+		    "\n"
+		    NOTICE_INTRO
+		    "\n",
+		    s->boundary, s->smtpname);
+
 		switch (s->msg->bounce.type) {
 		case B_ERROR:
 			iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
@@ -469,6 +505,34 @@ bounce_next(struct bounce_session *s)
 		    "    Below is a copy of the original message:\n"
 		    "\n");
 
+		iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
+		    "--%16" PRIu64 "/%s\n"
+		    "Content-Description: Delivery Report\n"
+		    "Content-Type: message/delivery-status\n"
+		    "\n",
+		    s->boundary, s->smtpname);
+
+		iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
+	    	    "Reporting-MTA: dns; %s\n"
+	    	    "Arrival-Date: %s\n"
+		    "\n",
+	    	    s->smtpname,
+	    	    time_to_text(time(NULL)));
+
+		TAILQ_FOREACH(evp, &s->msg->envelopes, entry) {
+			iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
+	    	    	    "Final-Recipient: rfc822; %s@%s\n"
+			    "Action: %s\n"
+	    	    	    "Status: %s\n"
+	    	    	    "Diagnostic-Code: smtp; %s\n"
+	    	    	    "\n",
+			    evp->dest.user,
+			    evp->dest.domain,
+			    action_str(&s->msg->bounce),
+	    	    	    esc_code(evp->esc_class, evp->esc_code),
+	    	    	    esc_description(evp->esc_code));
+		}
+
 		log_trace(TRACE_BOUNCE, "bounce: %p: >>> [... %zu bytes ...]",
 		    s, iobuf_queued(&s->iobuf));
 
@@ -476,9 +540,13 @@ bounce_next(struct bounce_session *s)
 		break;
 
 	case BOUNCE_DATA_MESSAGE:
+		iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
+		    "--%16" PRIu64 "/%s\n"
+	    	    "Content-Type: message/rfc822\n"
+	    	    "\n",
+	    	    s->boundary, s->smtpname);
 
 		n = iobuf_queued(&s->iobuf);
-
 		while (iobuf_queued(&s->iobuf) < BOUNCE_HIWAT) {
 			line = fgetln(s->msgfp, &len);
 			if (line == NULL)
@@ -488,6 +556,9 @@ bounce_next(struct bounce_session *s)
 			    s->msg->bounce.dsn_ret ==  DSN_RETHDRS) {
 				fclose(s->msgfp);
 				s->msgfp = NULL;
+				iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
+	    	    		    "\n--%16" PRIu64 "/%s--\n", s->boundary,
+				    s->smtpname);
 				bounce_send(s, ".");
 				s->state = BOUNCE_DATA_END;
 				return (0);
@@ -501,12 +572,15 @@ bounce_next(struct bounce_session *s)
 		if (ferror(s->msgfp)) {
 			fclose(s->msgfp);
 			s->msgfp = NULL;
-			bounce_delivery(s->msg, IMSG_DELIVERY_TEMPFAIL,
+			bounce_delivery(s->msg, IMSG_QUEUE_DELIVERY_TEMPFAIL,
 			    "Error reading message");
 			s->msg = NULL;
 			return (-1);
 		}
 
+		iobuf_xfqueue(&s->iobuf, "bounce_next: BODY",
+	    	    "\n--%16" PRIu64 "/%s--\n", s->boundary, s->smtpname);
+
 		log_trace(TRACE_BOUNCE, "bounce: %p: >>> [... %zu bytes ...]",
 		    s, iobuf_queued(&s->iobuf) - n);
 
@@ -541,7 +615,7 @@ bounce_delivery(struct bounce_message *msg, int delivery, const char *status)
 
 	n = 0;
 	while ((be = TAILQ_FIRST(&msg->envelopes))) {
-		if (delivery == IMSG_DELIVERY_TEMPFAIL) {
+		if (delivery == IMSG_QUEUE_DELIVERY_TEMPFAIL) {
 			if (queue_envelope_load(be->id, &evp) == 0) {
 				fatalx("could not reload envelope!");
 			}
@@ -565,16 +639,16 @@ bounce_delivery(struct bounce_message *msg, int delivery, const char *status)
 	}
 
 
-	if (delivery == IMSG_DELIVERY_TEMPFAIL)
+	if (delivery == IMSG_QUEUE_DELIVERY_TEMPFAIL)
 		f = "TempFail";
-	else if (delivery == IMSG_DELIVERY_PERMFAIL)
+	else if (delivery == IMSG_QUEUE_DELIVERY_PERMFAIL)
 		f = "PermFail";
 	else
 		f = NULL;
 
 	if (f)
-		log_warnx("warn: %s injecting failure report on message %08"PRIx32
-		    " to <%s> for %zu envelope%s: %s",
+		log_warnx("warn: %s injecting failure report on message %08"
+		    PRIx32 " to <%s> for %zu envelope%s: %s",
 		    f, msg->msgid, msg->to, n, n > 1 ? "s":"", status);
 
 	nmessage -= 1;
@@ -602,11 +676,11 @@ bounce_status(struct bounce_session *s, const char *fmt, ...)
 	va_end(ap);
 
 	if (*status == '2')
-		delivery = IMSG_DELIVERY_OK;
+		delivery = IMSG_QUEUE_DELIVERY_OK;
 	else if (*status == '5' || *status == '6')
-		delivery = IMSG_DELIVERY_PERMFAIL;
+		delivery = IMSG_QUEUE_DELIVERY_PERMFAIL;
 	else
-		delivery = IMSG_DELIVERY_TEMPFAIL;
+		delivery = IMSG_QUEUE_DELIVERY_TEMPFAIL;
 
 	bounce_delivery(s->msg, delivery, status);
 	s->msg = NULL;
@@ -648,7 +722,7 @@ bounce_io(struct io *io, int evt)
 	case IO_DATAIN:
 	    nextline:
 		line = iobuf_getline(&s->iobuf, &len);
-		if (line == NULL && iobuf_len(&s->iobuf) >= SMTPD_MAXLINESIZE) {
+		if (line == NULL && iobuf_len(&s->iobuf) >= LINE_MAX) {
 			bounce_status(s, "Input too long");
 			bounce_free(s);
 			return;
@@ -722,6 +796,25 @@ bounce_message_cmp(const struct bounce_message *a,
 	return memcmp(&a->bounce, &b->bounce, sizeof (a->bounce));
 }
 
+static const char *
+action_str(const struct delivery_bounce *b)
+{
+	switch (b->type) {
+	case B_ERROR:
+		return ("failed");
+	case B_WARNING:
+		return ("delayed");
+	case B_DSN:
+		if (b->mta_without_dsn)
+			return ("relayed");
+		
+		return ("delivered");
+	default:
+		log_warn("warn: bounce: unknown bounce_type");
+		return ("");
+	}
+}
+
 static const char *
 bounce_strtype(enum bounce_type t)
 {
