@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: control.c,v 1.100 2014/04/19 11:17:14 gilles Exp $	*/
 
 /*
  * Copyright (c) 2012 Gilles Chehade <gilles@poolp.org>
@@ -40,6 +40,7 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <limits.h>
 
 #include "smtpd.h"
 #include "log.h"
@@ -68,15 +69,18 @@ static void control_close(struct ctl_conn *);
 static void control_sig_handler(int, short, void *);
 static void control_dispatch_ext(struct mproc *, struct imsg *);
 static void control_digest_update(const char *, size_t, int);
+static void control_broadcast_verbose(int, int);
 
 static struct stat_backend *stat_backend = NULL;
 extern const char *backend_stat;
 
-static uint32_t			connid = 0;
+static uint64_t			connid = 0;
 static struct tree		ctl_conns;
+static struct tree		ctl_count;
 static struct stat_digest	digest;
 
-#define	CONTROL_FD_RESERVE	5
+#define	CONTROL_FD_RESERVE		5
+#define	CONTROL_MAXCONN_PER_CLIENT	32
 
 static void
 control_imsg(struct mproc *p, struct imsg *imsg)
@@ -88,9 +92,9 @@ control_imsg(struct mproc *p, struct imsg *imsg)
 	const void		*data;
 	size_t			 sz;
 
-	if (p->proc == PROC_SMTP) {
+	if (p->proc == PROC_PONY) {
 		switch (imsg->hdr.type) {
-		case IMSG_SMTP_ENQUEUE_FD:
+		case IMSG_CTL_SMTP_SESSION:
 			c = tree_get(&ctl_conns, imsg->hdr.peerid);
 			if (c == NULL)
 				return;
@@ -115,6 +119,9 @@ control_imsg(struct mproc *p, struct imsg *imsg)
 	if (p->proc == PROC_QUEUE) {
 		switch (imsg->hdr.type) {
 		case IMSG_CTL_LIST_ENVELOPES:
+		case IMSG_CTL_DISCOVER_EVPID:
+		case IMSG_CTL_DISCOVER_MSGID:
+		case IMSG_CTL_UNCORRUPT_MSGID:
 			c = tree_get(&ctl_conns, imsg->hdr.peerid);
 			if (c == NULL)
 				return;
@@ -122,7 +129,7 @@ control_imsg(struct mproc *p, struct imsg *imsg)
 			return;
 		}
 	}
-	if (p->proc == PROC_MTA) {
+	if (p->proc == PROC_PONY) {
 		switch (imsg->hdr.type) {
 		case IMSG_CTL_OK:
 		case IMSG_CTL_FAIL:
@@ -280,18 +287,17 @@ control(void)
 	signal(SIGHUP, SIG_IGN);
 
 	tree_init(&ctl_conns);
+	tree_init(&ctl_count);
 
 	memset(&digest, 0, sizeof digest);
 	digest.startup = time(NULL);
 
 	config_peer(PROC_SCHEDULER);
 	config_peer(PROC_QUEUE);
-	config_peer(PROC_SMTP);
-	config_peer(PROC_MFA);
 	config_peer(PROC_PARENT);
 	config_peer(PROC_LKA);
-	config_peer(PROC_MDA);
-	config_peer(PROC_MTA);
+	config_peer(PROC_PONY);
+	config_peer(PROC_CA);
 	config_done();
 
 	control_listen();
@@ -307,7 +313,6 @@ static void
 control_shutdown(void)
 {
 	log_info("info: control process exiting");
-	unlink(SMTPD_SOCKET);
 	_exit(0);
 }
 
@@ -330,6 +335,9 @@ control_accept(int listenfd, short event, void *arg)
 	socklen_t		 len;
 	struct sockaddr_un	 sun;
 	struct ctl_conn		*c;
+	size_t			*count;
+	uid_t			 euid;
+	gid_t			 egid;
 
 	if (available_fds(CONTROL_FD_RESERVE))
 		goto pause;
@@ -346,10 +354,31 @@ control_accept(int listenfd, short event, void *arg)
 
 	session_socket_blockmode(connfd, BM_NONBLOCK);
 
-	c = xcalloc(1, sizeof(*c), "control_accept");
-	if (getpeereid(connfd, &c->euid, &c->egid) == -1)
+	if (getpeereid(connfd, &euid, &egid) == -1)
 		fatal("getpeereid");
-	c->id = ++connid;
+
+	count = tree_get(&ctl_count, euid);
+	if (count == NULL) {
+		count = xcalloc(1, sizeof *count, "control_accept");
+		tree_xset(&ctl_count, euid, count);
+	}
+
+	if (*count == CONTROL_MAXCONN_PER_CLIENT) {
+		close(connfd);
+		log_warnx("warn: too many connections to control socket "
+		    "from user with uid %lu", (unsigned long int)euid);
+		return;
+	}
+	(*count)++;
+
+	do {
+		++connid;
+	} while (tree_get(&ctl_conns, connid));
+
+	c = xcalloc(1, sizeof(*c), "control_accept");
+	c->euid = euid;
+	c->egid = egid;
+	c->id = connid;
 	c->mproc.proc = PROC_CLIENT;
 	c->mproc.handler = control_dispatch_ext;
 	c->mproc.data = c;
@@ -368,6 +397,14 @@ pause:
 static void
 control_close(struct ctl_conn *c)
 {
+	size_t	*count;
+
+	count = tree_xget(&ctl_count, c->euid);
+	(*count)--;
+	if (*count == 0) {
+		tree_xpop(&ctl_count, c->euid);
+		free(count);
+	}
 	tree_xpop(&ctl_conns, c->id);
 	mproc_clear(&c->mproc);
 	free(c);
@@ -437,6 +474,8 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 	char			*key;
 	struct stat_value	 val;
 	size_t			 len;
+	uint64_t		 evpid;
+	uint32_t		 msgid;
 
 	c = p->data;
 
@@ -451,39 +490,33 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 	}
 
 	switch (imsg->hdr.type) {
-	case IMSG_SMTP_ENQUEUE_FD:
+	case IMSG_CTL_SMTP_SESSION:
 		if (env->sc_flags & (SMTPD_SMTP_PAUSED | SMTPD_EXITING)) {
 			m_compose(p, IMSG_CTL_FAIL, 0, 0, -1, NULL, 0);
 			return;
 		}
-		m_compose(p_smtp, IMSG_SMTP_ENQUEUE_FD, c->id, 0, -1,
+		m_compose(p_pony, IMSG_CTL_SMTP_SESSION, c->id, 0, -1,
 		    &c->euid, sizeof(c->euid));
 		return;
 
-	case IMSG_STATS:
-		if (c->euid)
-			goto badcred;
-		m_compose(p, IMSG_STATS, 0, 0, -1, NULL, 0);
-		return;
-
-	case IMSG_DIGEST:
+	case IMSG_CTL_GET_DIGEST:
 		if (c->euid)
 			goto badcred;
 		digest.timestamp = time(NULL);
-		m_compose(p, IMSG_DIGEST, 0, 0, -1, &digest, sizeof digest);
+		m_compose(p, IMSG_CTL_GET_DIGEST, 0, 0, -1, &digest, sizeof digest);
 		return;
 
-	case IMSG_STATS_GET:
+	case IMSG_CTL_GET_STATS:
 		if (c->euid)
 			goto badcred;
 		kvp = imsg->data;
 		if (! stat_backend->iter(&kvp->iter, &key, &val))
 			kvp->iter = NULL;
 		else {
-			strlcpy(kvp->key, key, sizeof kvp->key);
+			(void)strlcpy(kvp->key, key, sizeof kvp->key);
 			kvp->val = val;
 		}
-		m_compose(p, IMSG_STATS_GET, 0, 0, -1, kvp, sizeof *kvp);
+		m_compose(p, IMSG_CTL_GET_STATS, 0, 0, -1, kvp, sizeof *kvp);
 		return;
 
 	case IMSG_CTL_SHUTDOWN:
@@ -513,14 +546,12 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		verbose = v;
 		log_verbose(verbose);
 
-		m_create(p_parent, IMSG_CTL_VERBOSE, 0, 0, -1);
-		m_add_int(p_parent, verbose);
-		m_close(p_parent);
+		control_broadcast_verbose(IMSG_CTL_VERBOSE, verbose);
 
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
-	case IMSG_CTL_TRACE:
+	case IMSG_CTL_TRACE_ENABLE:
 		if (c->euid)
 			goto badcred;
 
@@ -531,14 +562,12 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		verbose |= v;
 		log_verbose(verbose);
 
-		m_create(p_parent, IMSG_CTL_TRACE, 0, 0, -1);
-		m_add_int(p_parent, v);
-		m_close(p_parent);
+		control_broadcast_verbose(IMSG_CTL_VERBOSE, verbose);
 
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
-	case IMSG_CTL_UNTRACE:
+	case IMSG_CTL_TRACE_DISABLE:
 		if (c->euid)
 			goto badcred;
 
@@ -549,14 +578,12 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		verbose &= ~v;
 		log_verbose(verbose);
 
-		m_create(p_parent, IMSG_CTL_UNTRACE, 0, 0, -1);
-		m_add_int(p_parent, v);
-		m_close(p_parent);
+		control_broadcast_verbose(IMSG_CTL_VERBOSE, verbose);
 
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
-	case IMSG_CTL_PROFILE:
+	case IMSG_CTL_PROFILE_ENABLE:
 		if (c->euid)
 			goto badcred;
 
@@ -566,14 +593,12 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		memcpy(&v, imsg->data, sizeof(v));
 		profiling |= v;
 
-		m_create(p_parent, IMSG_CTL_PROFILE, 0, 0, -1);
-		m_add_int(p_parent, v);
-		m_close(p_parent);
+		control_broadcast_verbose(IMSG_CTL_PROFILE, profiling);
 
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
-	case IMSG_CTL_UNPROFILE:
+	case IMSG_CTL_PROFILE_DISABLE:
 		if (c->euid)
 			goto badcred;
 
@@ -583,9 +608,7 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		memcpy(&v, imsg->data, sizeof(v));
 		profiling &= ~v;
 
-		m_create(p_parent, IMSG_CTL_UNPROFILE, 0, 0, -1);
-		m_add_int(p_parent, v);
-		m_close(p_parent);
+		control_broadcast_verbose(IMSG_CTL_PROFILE, profiling);
 
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
@@ -636,7 +659,7 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		}
 		log_info("info: smtp paused");
 		env->sc_flags |= SMTPD_SMTP_PAUSED;
-		m_compose(p_smtp, IMSG_CTL_PAUSE_SMTP, 0, 0, -1, NULL, 0);
+		m_compose(p_pony, IMSG_CTL_PAUSE_SMTP, 0, 0, -1, NULL, 0);
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
@@ -686,7 +709,7 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		}
 		log_info("info: smtp resumed");
 		env->sc_flags &= ~SMTPD_SMTP_PAUSED;
-		m_forward(p_smtp, imsg);
+		m_forward(p_pony, imsg);
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
@@ -694,7 +717,7 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		if (c->euid)
 			goto badcred;
 
-		m_forward(p_mta, imsg);
+		m_forward(p_pony, imsg);
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
@@ -721,7 +744,7 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 			goto badcred;
 
 		imsg->hdr.peerid = c->id;
-		m_forward(p_mta, imsg);
+		m_forward(p_pony, imsg);
 		return;
 
 	case IMSG_CTL_SHOW_STATUS:
@@ -740,10 +763,10 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		if (imsg->hdr.len - IMSG_HEADER_SIZE <= sizeof(ss))
 			goto invalid;
 		memmove(&ss, imsg->data, sizeof(ss));
-		m_create(p_mta, imsg->hdr.type, c->id, 0, -1);
-		m_add_sockaddr(p_mta, (struct sockaddr *)&ss);
-		m_add_string(p_mta, (char *)imsg->data + sizeof(ss));
-		m_close(p_mta);
+		m_create(p_pony, imsg->hdr.type, c->id, 0, -1);
+		m_add_sockaddr(p_pony, (struct sockaddr *)&ss);
+		m_add_string(p_pony, (char *)imsg->data + sizeof(ss));
+		m_close(p_pony);
 		return;
 
 	case IMSG_CTL_SCHEDULE:
@@ -762,19 +785,46 @@ control_dispatch_ext(struct mproc *p, struct imsg *imsg)
 		m_forward(p_scheduler, imsg);
 		return;
 
-	case IMSG_LKA_UPDATE_TABLE:
+	case IMSG_CTL_UPDATE_TABLE:
 		if (c->euid)
 			goto badcred;
 
 		/* table name too long */
 		len = strlen(imsg->data);
-		if (len >= SMTPD_MAXLINESIZE)
+		if (len >= LINE_MAX)
 			goto invalid;
 
 		m_forward(p_lka, imsg);
 		m_compose(p, IMSG_CTL_OK, 0, 0, -1, NULL, 0);
 		return;
 
+	case IMSG_CTL_DISCOVER_EVPID:
+		if (c->euid)
+			goto badcred;
+
+		if (imsg->hdr.len - IMSG_HEADER_SIZE != sizeof evpid)
+			goto invalid;
+
+		memmove(&evpid, imsg->data, sizeof evpid);
+		m_create(p_queue, imsg->hdr.type, c->id, 0, -1);
+		m_add_evpid(p_queue, evpid);
+		m_close(p_queue);
+		return;
+
+	case IMSG_CTL_DISCOVER_MSGID:
+	case IMSG_CTL_UNCORRUPT_MSGID:
+		if (c->euid)
+			goto badcred;
+
+		if (imsg->hdr.len - IMSG_HEADER_SIZE != sizeof msgid)
+			goto invalid;
+
+		memmove(&msgid, imsg->data, sizeof msgid);
+		m_create(p_queue, imsg->hdr.type, c->id, 0, -1);
+		m_add_msgid(p_queue, msgid);
+		m_close(p_queue);
+		return;
+
 	default:
 		log_debug("debug: control_dispatch_ext: "
 		    "error handling %s imsg",
@@ -785,3 +835,31 @@ badcred:
 invalid:
 	m_compose(p, IMSG_CTL_FAIL, 0, 0, -1, NULL, 0);
 }
+
+static void
+control_broadcast_verbose(int msg, int v)
+{
+	m_create(p_lka, msg, 0, 0, -1);
+	m_add_int(p_lka, v);
+	m_close(p_lka);
+
+	m_create(p_pony, msg, 0, 0, -1);
+	m_add_int(p_pony, v);
+	m_close(p_pony);
+
+	m_create(p_queue, msg, 0, 0, -1);
+	m_add_int(p_queue, v);
+	m_close(p_queue);
+
+	m_create(p_ca, msg, 0, 0, -1);
+	m_add_int(p_ca, v);
+	m_close(p_ca);
+
+	m_create(p_scheduler, msg, 0, 0, -1);
+	m_add_int(p_scheduler, v);
+	m_close(p_scheduler);
+
+	m_create(p_parent, msg, 0, 0, -1);
+	m_add_int(p_parent, v);
+	m_close(p_parent);
+}
