@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: lka_session.c,v 1.66 2014/04/19 12:55:23 gilles Exp $	*/
 
 /*
  * Copyright (c) 2011 Gilles Chehade <gilles@poolp.org>
@@ -38,12 +38,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <limits.h>
 
 #include "smtpd.h"
 #include "log.h"
 
-#define TAG_CHAR	'+'	/* gilles+tag@ */
-
 #define	EXPAND_DEPTH	10
 
 #define	F_WAITING	0x01
@@ -72,7 +71,6 @@ static void lka_resume(struct lka_session *);
 static size_t lka_expand_format(char *, size_t, const struct envelope *,
     const struct userinfo *);
 static void mailaddr_to_username(const struct mailaddr *, char *, size_t);
-static int mailaddr_tag(const struct mailaddr *, char *, size_t);
 
 static int mod_lowercase(char *, size_t);
 static int mod_uppercase(char *, size_t);
@@ -87,8 +85,6 @@ struct modifiers {
 	{ "strip",	mod_strip },
 	{ "raw",	NULL },		/* special case, must stay last */
 };
-static const char	*unsafe = MAILADDR_ESCAPE;
-
 
 static int		init;
 static struct tree	sessions;
@@ -231,20 +227,20 @@ lka_resume(struct lka_session *lks)
 	}
     error:
 	if (lks->error) {
-		m_create(p_smtp, IMSG_LKA_EXPAND_RCPT, 0, 0, -1);
-		m_add_id(p_smtp, lks->id);
-		m_add_int(p_smtp, lks->error);
+		m_create(p_pony, IMSG_SMTP_EXPAND_RCPT, 0, 0, -1);
+		m_add_id(p_pony, lks->id);
+		m_add_int(p_pony, lks->error);
 
 		if (lks->errormsg)
-			m_add_string(p_smtp, lks->errormsg);
+			m_add_string(p_pony, lks->errormsg);
 		else {
 			if (lks->error == LKA_PERMFAIL)
-				m_add_string(p_smtp, "550 Invalid recipient");
+				m_add_string(p_pony, "550 Invalid recipient");
 			else if (lks->error == LKA_TEMPFAIL)
-				m_add_string(p_smtp, "451 Temporary failure");
+				m_add_string(p_pony, "451 Temporary failure");
 		}
 
-		m_close(p_smtp);
+		m_close(p_pony);
 		while ((ep = TAILQ_FIRST(&lks->deliverylist)) != NULL) {
 			TAILQ_REMOVE(&lks->deliverylist, ep, entry);
 			free(ep);
@@ -254,14 +250,14 @@ lka_resume(struct lka_session *lks)
 		/* Process the delivery list and submit envelopes to queue */
 		while ((ep = TAILQ_FIRST(&lks->deliverylist)) != NULL) {
 			TAILQ_REMOVE(&lks->deliverylist, ep, entry);
-			m_create(p_queue, IMSG_QUEUE_SUBMIT_ENVELOPE, 0, 0, -1);
+			m_create(p_queue, IMSG_LKA_ENVELOPE_SUBMIT, 0, 0, -1);
 			m_add_id(p_queue, lks->id);
 			m_add_envelope(p_queue, ep);
 			m_close(p_queue);
 			free(ep);
 		}
 
-		m_create(p_queue, IMSG_QUEUE_COMMIT_ENVELOPES, 0, 0, -1);
+		m_create(p_queue, IMSG_LKA_ENVELOPE_COMMIT, 0, 0, -1);
 		m_add_id(p_queue, lks->id);
 		m_close(p_queue);
 	}
@@ -280,6 +276,7 @@ lka_expand(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 	struct mailaddr		maddr;
 	int			r;
 	union lookup		lk;
+	char		       *tag;
 
 	if (xn->depth >= EXPAND_DEPTH) {
 		log_trace(TRACE_EXPAND, "expand: lka_expand: node too deep.");
@@ -384,15 +381,21 @@ lka_expand(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 				break;
 		}
 
+		/* gilles+hackers@ -> gilles@ */
+		if ((tag = strchr(xn->u.user, TAG_CHAR)) != NULL)
+			*tag++ = '\0';
+
 		/* A username should not exceed the size of a system user */
+		/*
 		if (strlen(xn->u.user) >= sizeof fwreq.user) {
 			log_trace(TRACE_EXPAND, "expand: lka_expand: "
 			    "user-part too long to be a system user");
 			lks->error = LKA_PERMFAIL;
 			break;
 		}
+		*/
 
-		r = table_lookup(rule->r_userbase, xn->u.user, K_USERINFO, &lk);
+		r = table_lookup(rule->r_userbase, NULL, xn->u.user, K_USERINFO, &lk);
 		if (r == -1) {
 			log_trace(TRACE_EXPAND, "expand: lka_expand: "
 			    "backend error while searching user");
@@ -417,7 +420,7 @@ lka_expand(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 		fwreq.uid = lk.userinfo.uid;
 		fwreq.gid = lk.userinfo.gid;
 
-		m_compose(p_parent, IMSG_PARENT_FORWARD_OPEN, 0, 0, -1,
+		m_compose(p_parent, IMSG_LKA_OPEN_FORWARD, 0, 0, -1,
 		    &fwreq, sizeof(fwreq));
 		lks->flags |= F_WAITING;
 		break;
@@ -458,6 +461,27 @@ lka_expand(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 		    "[depth=%d]", xn->u.buffer, xn->depth);
 		lka_submit(lks, rule, xn);
 		break;
+
+	case EXPAND_MAILDIR:
+		log_trace(TRACE_EXPAND, "expand: lka_expand: maildir: %s "
+		    "[depth=%d]", xn->u.buffer, xn->depth);
+		r = table_lookup(rule->r_userbase, NULL,
+		    xn->parent->u.user, K_USERINFO, &lk);
+		if (r == -1) {
+			log_trace(TRACE_EXPAND, "expand: lka_expand: maildir: "
+			    "backend error while searching user");
+			lks->error = LKA_TEMPFAIL;
+			break;
+		}
+		if (r == 0) {
+			log_trace(TRACE_EXPAND, "expand: lka_expand: maildir: "
+			    "user-part does not match system user");
+			lks->error = LKA_PERMFAIL;
+			break;
+		}
+
+		lka_submit(lks, rule, xn);
+		break;
 	}
 }
 
@@ -481,7 +505,6 @@ lka_submit(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 	struct envelope		*ep;
 	struct expandnode	*xn2;
 	int			 r;
-	char			 tag[EXPAND_BUFFER];
 
 	ep = xmemdup(&lks->envelope, sizeof *ep, "lka_submit");
 	ep->expire = rule->r_qexpire;
@@ -497,10 +520,10 @@ lka_submit(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 
 		/* only rewrite if not a bounce */
 		if (ep->sender.user[0] && rule->r_as && rule->r_as->user[0])
-			strlcpy(ep->sender.user, rule->r_as->user,
+			(void)strlcpy(ep->sender.user, rule->r_as->user,
 			    sizeof ep->sender.user);
 		if (ep->sender.user[0] && rule->r_as && rule->r_as->domain[0])
-			strlcpy(ep->sender.domain, rule->r_as->domain,
+			(void)strlcpy(ep->sender.domain, rule->r_as->domain,
 			    sizeof ep->sender.domain);
 		break;
 	case A_NONE:
@@ -515,54 +538,48 @@ lka_submit(struct lka_session *lks, struct rule *rule, struct expandnode *xn)
 		/* set username */
 		if ((xn->type == EXPAND_FILTER || xn->type == EXPAND_FILENAME)
 		    && xn->alias) {
-			strlcpy(ep->agent.mda.username, SMTPD_USER,
+			(void)strlcpy(ep->agent.mda.username, SMTPD_USER,
 			    sizeof(ep->agent.mda.username));
 		}
 		else {
 			xn2 = lka_find_ancestor(xn, EXPAND_USERNAME);
-			strlcpy(ep->agent.mda.username, xn2->u.user,
+			(void)strlcpy(ep->agent.mda.username, xn2->u.user,
 			    sizeof(ep->agent.mda.username));
 		}
 
-		r = table_lookup(rule->r_userbase, ep->agent.mda.username,
+		r = table_lookup(rule->r_userbase, NULL, ep->agent.mda.username,
 		    K_USERINFO, &lk);
 		if (r <= 0) {
 			lks->error = (r == -1) ? LKA_TEMPFAIL : LKA_PERMFAIL;
 			free(ep);
 			return;
 		}
-		strlcpy(ep->agent.mda.usertable, rule->r_userbase->t_name,
+		(void)strlcpy(ep->agent.mda.usertable, rule->r_userbase->t_name,
 		    sizeof ep->agent.mda.usertable);
-		strlcpy(ep->agent.mda.username, lk.userinfo.username,
+		(void)strlcpy(ep->agent.mda.username, lk.userinfo.username,
 		    sizeof ep->agent.mda.username);
+		strlcpy(ep->agent.mda.delivery_user, rule->r_delivery_user,
+		    sizeof ep->agent.mda.delivery_user);
 
 		if (xn->type == EXPAND_FILENAME) {
 			ep->agent.mda.method = A_FILENAME;
-			strlcpy(ep->agent.mda.buffer, xn->u.buffer,
+			(void)strlcpy(ep->agent.mda.buffer, xn->u.buffer,
 			    sizeof ep->agent.mda.buffer);
 		}
 		else if (xn->type == EXPAND_FILTER) {
 			ep->agent.mda.method = A_MDA;
-			strlcpy(ep->agent.mda.buffer, xn->u.buffer,
+			(void)strlcpy(ep->agent.mda.buffer, xn->u.buffer,
 			    sizeof ep->agent.mda.buffer);
 		}
 		else if (xn->type == EXPAND_USERNAME) {
 			ep->agent.mda.method = rule->r_action;
-			strlcpy(ep->agent.mda.buffer, rule->r_value.buffer,
+			(void)strlcpy(ep->agent.mda.buffer, rule->r_value.buffer,
+			    sizeof ep->agent.mda.buffer);
+		}
+		else if (xn->type == EXPAND_MAILDIR) {
+			ep->agent.mda.method = A_MAILDIR;
+			(void)strlcpy(ep->agent.mda.buffer, xn->u.buffer,
 			    sizeof ep->agent.mda.buffer);
-
-			memset(tag, 0, sizeof tag);
-			if (! mailaddr_tag(&ep->dest, tag, sizeof tag)) {
-				lks->error = LKA_PERMFAIL;
-				free(ep);
-				return;
-			}
-			if (rule->r_action == A_MAILDIR && tag[0]) {
-				strlcat(ep->agent.mda.buffer, "/.",
-				    sizeof(ep->agent.mda.buffer));
-				strlcat(ep->agent.mda.buffer, tag,
-				    sizeof(ep->agent.mda.buffer));
-			}
 		}
 		else
 			fatalx("lka_deliver: bad node type");
@@ -709,7 +726,7 @@ lka_expand_token(char *dest, size_t len, const char *token,
 	
 	if (! raw && replace)
 		for (i = 0; (size_t)i < strlen(tmp); ++i)
-			if (strchr(unsafe, tmp[i]))
+			if (strchr(MAILADDR_ESCAPE, tmp[i]))
 				tmp[i] = ':';
 
 	/* expanded string is empty */
@@ -765,8 +782,10 @@ lka_expand_format(char *buf, size_t len, const struct envelope *ep,
 	char		token[MAXTOKENLEN];
 	size_t		ret, tmpret;
 
-	if (len < sizeof tmpbuf)
-		fatalx("lka_expand_format: tmp buffer < rule buffer");
+	if (len < sizeof tmpbuf) {
+		log_warnx("lka_expand_format: tmp buffer < rule buffer");
+		return 0;
+	}
 
 	memset(tmpbuf, 0, sizeof tmpbuf);
 	pbuf = buf;
@@ -820,6 +839,10 @@ lka_expand_format(char *buf, size_t len, const struct envelope *ep,
 		if (exptoklen == 0)
 			return 0;
 
+		/* writing expanded token at ptmp will overflow tmpbuf */
+		if (sizeof (tmpbuf) - (ptmp - tmpbuf) <= exptoklen)
+			return 0;
+
 		memcpy(ptmp, exptok, exptoklen);
 		pbuf   = ebuf + 1;
 		ptmp  += exptoklen;
@@ -837,35 +860,7 @@ lka_expand_format(char *buf, size_t len, const struct envelope *ep,
 static void
 mailaddr_to_username(const struct mailaddr *maddr, char *dst, size_t len)
 {
-	char	*tag;
-
 	xlowercase(dst, maddr->user, len);
-
-	/* gilles+hackers@ -> gilles@ */
-	if ((tag = strchr(dst, TAG_CHAR)) != NULL)
-		*tag++ = '\0';
-}
-
-static int
-mailaddr_tag(const struct mailaddr *maddr, char *dest, size_t len)
-{
-	char		*tag;
-	char		*sanitized;
-
-	if ((tag = strchr(maddr->user, TAG_CHAR))) {
-		tag++;
-		while (*tag == '.')
-			tag++;
-	}
-	if (tag == NULL)
-		return 1;
-
-	if (strlcpy(dest, tag, len) >= len)
-		return 0;
-	for (sanitized = dest; *sanitized; sanitized++)
-		if (strchr(unsafe, *sanitized))
-			*sanitized = ':';
-	return 1;
 }
 
 static int 
