@@ -45,6 +45,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <limits.h>
 #ifdef HAVE_UTIL_H
 #include <util.h>
 #endif
@@ -96,11 +97,17 @@ purge_config(uint8_t what)
 	memset(env, 0, sizeof(struct smtpd));
 }
 
+int
+fork_proc_backend(const char *backend, const char *conf, const char *procname)
+{
+	return (-1);
+}
+
 int
 main(int argc, char *argv[])
 {
 	struct stat	 sb;
-	char		 dbname[SMTPD_MAXPATHLEN];
+	char		 dbname[PATH_MAX];
 	char		*opts;
 	char		*conf;
 	int		 ch;
@@ -253,7 +260,6 @@ parse_map(char *filename)
 	char	*line;
 	size_t	 len;
 	size_t	 lineno = 0;
-	char	 delim[] = { '\\', 0, 0 };
 
 	if (strcmp(filename, "-") == 0)
 		fp = fdopen(0, "r");
@@ -273,7 +279,8 @@ parse_map(char *filename)
 		return 0;
 	}
 
-	while ((line = fparseln(fp, &len, &lineno, delim, 0)) != NULL) {
+	while ((line = fparseln(fp, &len, &lineno,
+	    NULL, FPARSELN_UNESCCOMM)) != NULL) {
 		if (! parse_entry(line, len, lineno)) {
 			free(line);
 			fclose(fp);
@@ -310,7 +317,7 @@ parse_mapentry(char *line, size_t len, size_t lineno)
 	keyp = line;
 	while (isspace((unsigned char)*keyp))
 		keyp++;
-	if (*keyp == '\0' || *keyp == '#')
+	if (*keyp == '\0')
 		return 1;
 
 	valp = keyp;
@@ -319,7 +326,7 @@ parse_mapentry(char *line, size_t len, size_t lineno)
 		goto bad;
 	while (*valp == ':' || isspace((unsigned char)*valp))
 		valp++;
-	if (*valp == '\0' || *valp == '#')
+	if (*valp == '\0')
 		goto bad;
 
 	/* Check for dups. */
@@ -367,7 +374,7 @@ parse_setentry(char *line, size_t len, size_t lineno)
 	keyp = line;
 	while (isspace((unsigned char)*keyp))
 		keyp++;
-	if (*keyp == '\0' || *keyp == '#')
+	if (*keyp == '\0')
 		return 1;
 
 	val.data  = "<set>";
@@ -406,7 +413,6 @@ make_aliases(DBT *val, char *text)
 {
 	struct expandnode	xn;
 	char		       *subrcpt;
-	char		       *endp;
 	char		       *origtext;
 
 	val->data = NULL;
@@ -415,17 +421,11 @@ make_aliases(DBT *val, char *text)
 	origtext = xstrdup(text, "make_aliases");
 
 	while ((subrcpt = strsep(&text, ",")) != NULL) {
-		/* subrcpt: strip initial whitespace. */
-		while (isspace((unsigned char)*subrcpt))
-			++subrcpt;
+		/* subrcpt: strip initial and trailing whitespace. */
+		subrcpt = strip(subrcpt);
 		if (*subrcpt == '\0')
 			goto error;
 
-		/* subrcpt: strip trailing whitespace. */
-		endp = subrcpt + strlen(subrcpt) - 1;
-		while (subrcpt < endp && isspace((unsigned char)*endp))
-			*endp-- = '\0';
-
 		if (! text_to_expandnode(&xn, subrcpt))
 			goto error;
 	}
