@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: mda.c,v 1.105 2014/04/19 17:42:18 gilles Exp $	*/
 
 /*
  * Copyright (c) 2008 Gilles Chehade <gilles@poolp.org>
@@ -40,6 +40,7 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <limits.h>
 #if defined(HAVE_STRNVIS) && defined(HAVE_VIS_H)
 #include <vis.h>
 #endif
@@ -70,8 +71,8 @@ struct mda_user {
 	uint64_t			id;
 	TAILQ_ENTRY(mda_user)		entry;
 	TAILQ_ENTRY(mda_user)		entry_runnable;
-	char				name[SMTPD_MAXLOGNAME];
-	char				usertable[SMTPD_MAXPATHLEN];
+	char				name[LOGIN_NAME_MAX];
+	char				usertable[PATH_MAX];
 	size_t				evpcount;
 	TAILQ_HEAD(, mda_envelope)	envelopes;
 	int				flags;
@@ -88,16 +89,17 @@ struct mda_session {
 	FILE			*datafp;
 };
 
-static void mda_imsg(struct mproc *, struct imsg *);
 static void mda_io(struct io *, int);
-static void mda_shutdown(void);
-static void mda_sig_handler(int, short, void *);
 static int mda_check_loop(FILE *, struct mda_envelope *);
 static int mda_getlastline(int, char *, size_t);
 static void mda_done(struct mda_session *);
 static void mda_fail(struct mda_user *, int, const char *, enum enhanced_status_code);
 static void mda_drain(void);
 static void mda_log(const struct mda_envelope *, const char *, const char *);
+static void mda_queue_ok(uint64_t);
+static void mda_queue_tempfail(uint64_t, const char *, enum enhanced_status_code);
+static void mda_queue_permfail(uint64_t, const char *, enum enhanced_status_code);
+static void mda_queue_loop(uint64_t);
 static struct mda_user *mda_user(const struct envelope *);
 static void mda_user_free(struct mda_user *);
 static const char *mda_user_to_text(const struct mda_user *);
@@ -110,7 +112,7 @@ static struct tree	users;
 
 static TAILQ_HEAD(, mda_user)	runnable;
 
-static void
+void
 mda_imsg(struct mproc *p, struct imsg *imsg)
 {
 	struct mda_session	*s;
@@ -123,14 +125,15 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 	const void		*data;
 	const char		*error, *parent_error;
 	uint64_t		 reqid;
+	time_t			 now;
 	size_t			 sz;
-	char			 out[256], buf[SMTPD_MAXLINESIZE];
-	int			 n, v;
+	char			 out[256], buf[LINE_MAX];
+	int			 n;
 	enum lka_resp_status	status;
 
 	if (p->proc == PROC_LKA) {
 		switch (imsg->hdr.type) {
-		case IMSG_LKA_USERINFO:
+		case IMSG_MDA_LOOKUP_USERINFO:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_get_int(&m, (int *)&status);
@@ -162,7 +165,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 	if (p->proc == PROC_QUEUE) {
 		switch (imsg->hdr.type) {
 
-		case IMSG_MDA_DELIVER:
+		case IMSG_QUEUE_DELIVER:
 			m_msg(&m, imsg);
 			m_get_envelope(&m, &evp);
 			m_end(&m);
@@ -180,7 +183,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 
 			if (u->flags & USER_ONHOLD) {
 				u->flags |= USER_HOLDQ;
-				m_create(p_queue, IMSG_DELIVERY_HOLD, 0, 0, -1);
+				m_create(p_queue, IMSG_MDA_DELIVERY_HOLD, 0, 0, -1);
 				m_add_evpid(p_queue, evp.id);
 				m_add_id(p_queue, u->id);
 				m_close(p_queue);
@@ -201,7 +204,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 			mda_drain();
 			return;
 
-		case IMSG_QUEUE_MESSAGE_FD:
+		case IMSG_MDA_OPEN_MESSAGE:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_end(&m);
@@ -211,7 +214,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 
 			if (imsg->fd == -1) {
 				log_debug("debug: mda: cannot get message fd");
-				queue_tempfail(e->id, "Cannot get message fd",
+				mda_queue_tempfail(e->id, "Cannot get message fd",
 				    ESC_OTHER_MAIL_SYSTEM_STATUS);
 				mda_log(e, "TempFail", "Cannot get message fd");
 				mda_done(s);
@@ -225,7 +228,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 			if ((s->datafp = fdopen(imsg->fd, "r")) == NULL) {
 				log_warn("warn: mda: fdopen");
 				close(imsg->fd);
-				queue_tempfail(e->id, "fdopen failed",
+				mda_queue_tempfail(e->id, "fdopen failed",
 				    ESC_OTHER_MAIL_SYSTEM_STATUS);
 				mda_log(e, "TempFail", "fdopen failed");
 				mda_done(s);
@@ -235,26 +238,39 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 			/* check delivery loop */
 			if (mda_check_loop(s->datafp, e)) {
 				log_debug("debug: mda: loop detected");
-				queue_loop(e->id);
+				mda_queue_loop(e->id);
 				mda_log(e, "PermFail", "Loop detected");
 				mda_done(s);
 				return;
 			}
 
-			/* start queueing delivery headers */
-			if (e->sender[0])
-				/* XXX: remove exising Return-Path, if any */
-				n = iobuf_fqueue(&s->iobuf,
-				    "Return-Path: %s\nDelivered-To: %s\n",
-				    e->sender, e->rcpt ? e->rcpt : e->dest);
-			else
-				n = iobuf_fqueue(&s->iobuf,
-				    "Delivered-To: %s\n",
-				    e->rcpt ? e->rcpt : e->dest);
+			n = 0;
+			/* prepend "From " separator ... for A_MDA and A_FILENAME backends only */
+			if (e->method == A_MDA || e->method == A_FILENAME) {
+				time(&now);
+				if (e->sender[0])
+					n = iobuf_fqueue(&s->iobuf,
+					    "From %s %s", e->sender, ctime(&now));
+				else
+					n = iobuf_fqueue(&s->iobuf,
+					    "From MAILER-DAEMON@%s %s", env->sc_hostname, ctime(&now));
+			}
+			if (n != -1) {
+				/* start queueing delivery headers */
+				if (e->sender[0])
+					/* XXX: remove existing Return-Path, if any */
+					n = iobuf_fqueue(&s->iobuf,
+					    "Return-Path: %s\nDelivered-To: %s\n",
+					    e->sender, e->rcpt ? e->rcpt : e->dest);
+				else
+					n = iobuf_fqueue(&s->iobuf,
+					    "Delivered-To: %s\n",
+					    e->rcpt ? e->rcpt : e->dest);
+			}
 			if (n == -1) {
 				log_warn("warn: mda: "
 				    "fail to write delivery info");
-				queue_tempfail(e->id, "Out of memory",
+				mda_queue_tempfail(e->id, "Out of memory",
 				    ESC_OTHER_MAIL_SYSTEM_STATUS);
 				mda_log(e, "TempFail", "Out of memory");
 				mda_done(s);
@@ -268,10 +284,19 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 			case A_MDA:
 				deliver.mode = A_MDA;
 				deliver.userinfo = *userinfo;
-				strlcpy(deliver.user, userinfo->username,
+				(void)strlcpy(deliver.user, userinfo->username,
 				    sizeof(deliver.user));
-				strlcpy(deliver.to, e->buffer,
-				    sizeof(deliver.to));
+				if (strlcpy(deliver.to, e->buffer,
+					sizeof(deliver.to))
+				    >= sizeof(deliver.to)) {
+					mda_queue_tempfail(e->id,
+					    "mda command too long",
+					    ESC_OTHER_MAIL_SYSTEM_STATUS);
+					mda_log(e, "TempFail",
+					    "mda command too long");
+					mda_done(s);
+					return;
+				}
 				break;
 
 			case A_MBOX:
@@ -281,41 +306,76 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 				deliver.mode = A_MBOX;
 				deliver.userinfo = *userinfo;
 				deliver.userinfo.uid = 0;
-				strlcpy(deliver.user, "root",
+				(void)strlcpy(deliver.user, "root",
 				    sizeof(deliver.user));
-				strlcpy(deliver.to, userinfo->username,
-				    sizeof(deliver.to));
-				strlcpy(deliver.from, e->sender,
+				(void)strlcpy(deliver.from, e->sender,
 				    sizeof(deliver.from));
+				(void)strlcpy(deliver.to, userinfo->username,
+				    sizeof(deliver.to));
 				break;
 
 			case A_MAILDIR:
 				deliver.mode = A_MAILDIR;
 				deliver.userinfo = *userinfo;
-				strlcpy(deliver.user, userinfo->username,
+				(void)strlcpy(deliver.user, userinfo->username,
 				    sizeof(deliver.user));
-				strlcpy(deliver.to, e->buffer,
-				    sizeof(deliver.to));
+				(void)strlcpy(deliver.dest, e->dest,
+				    sizeof(deliver.dest));
+				if (strlcpy(deliver.to, e->buffer,
+					sizeof(deliver.to))
+				    >= sizeof(deliver.to)) {
+					log_warn("warn: mda: "
+					    "deliver buffer too large");
+					mda_queue_tempfail(e->id,
+					    "Maildir path too long",
+					    ESC_OTHER_MAIL_SYSTEM_STATUS);
+					mda_log(e, "TempFail",
+					    "Maildir path too long");
+					mda_done(s);
+					return;
+				}
 				break;
 
 			case A_FILENAME:
 				deliver.mode = A_FILENAME;
 				deliver.userinfo = *userinfo;
-				strlcpy(deliver.user, userinfo->username,
+				(void)strlcpy(deliver.user, userinfo->username,
 				    sizeof deliver.user);
-				strlcpy(deliver.to, e->buffer,
-				    sizeof deliver.to);
+				if (strlcpy(deliver.to, e->buffer,
+					sizeof(deliver.to))
+				    >= sizeof(deliver.to)) {
+					log_warn("warn: mda: "
+					    "deliver buffer too large");
+					mda_queue_tempfail(e->id,
+					    "filename path too long",
+					    ESC_OTHER_MAIL_SYSTEM_STATUS);
+					mda_log(e, "TempFail",
+					    "filename path too long");
+					mda_done(s);
+					return;
+				}
 				break;
 
 			case A_LMTP:
 				deliver.mode = A_LMTP;
 				deliver.userinfo = *userinfo;
-				strlcpy(deliver.user, userinfo->username,
+				(void)strlcpy(deliver.user, e->user,
 				    sizeof(deliver.user));
-				strlcpy(deliver.to, e->buffer,
-				    sizeof(deliver.to));
-				strlcpy(deliver.from, e->sender,
+				(void)strlcpy(deliver.from, e->sender,
 				    sizeof(deliver.from));
+				if (strlcpy(deliver.to, e->buffer,
+					sizeof(deliver.to))
+				    >= sizeof(deliver.to)) {
+					log_warn("warn: mda: "
+					    "deliver buffer too large");
+					mda_queue_tempfail(e->id,
+					    "socket path too long",
+					    ESC_OTHER_MAIL_SYSTEM_STATUS);
+					mda_log(e, "TempFail",
+					    "socket path too long");
+					mda_done(s);
+					return;
+				}
 				break;
 
 			default:
@@ -327,7 +387,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 			    "for session %016"PRIx64 " evpid %016"PRIx64,
 			    s->id, s->evp->id);
 
-			m_create(p_parent, IMSG_PARENT_FORK_MDA, 0, 0, -1);
+			m_create(p_parent, IMSG_MDA_FORK, 0, 0, -1);
 			m_add_id(p_parent, reqid);
 			m_add_data(p_parent, &deliver, sizeof(deliver));
 			m_close(p_parent);
@@ -337,7 +397,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 
 	if (p->proc == PROC_PARENT) {
 		switch (imsg->hdr.type) {
-		case IMSG_PARENT_FORK_MDA:
+		case IMSG_MDA_FORK:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_end(&m);
@@ -346,7 +406,7 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 			e = s->evp;
 			if (imsg->fd == -1) {
 				log_warn("warn: mda: fail to retrieve mda fd");
-				queue_tempfail(e->id, "Cannot get mda fd",
+				mda_queue_tempfail(e->id, "Cannot get mda fd",
 				    ESC_OTHER_MAIL_SYSTEM_STATUS);
 				mda_log(e, "TempFail", "Cannot get mda fd");
 				mda_done(s);
@@ -390,117 +450,34 @@ mda_imsg(struct mproc *p, struct imsg *imsg)
 
 			/* update queue entry */
 			if (error) {
-				queue_tempfail(e->id, error,
+				mda_queue_tempfail(e->id, error,
 				    ESC_OTHER_MAIL_SYSTEM_STATUS);
-				snprintf(buf, sizeof buf, "Error (%s)", error);
+				(void)snprintf(buf, sizeof buf, "Error (%s)", error);
 				mda_log(e, "TempFail", buf);
 			}
 			else {
-				queue_ok(e->id);
+				mda_queue_ok(e->id);
 				mda_log(e, "Ok", "Delivered");
 			}
 			mda_done(s);
 			return;
-
-		case IMSG_CTL_VERBOSE:
-			m_msg(&m, imsg);
-			m_get_int(&m, &v);
-			m_end(&m);
-			log_verbose(v);
-			return;
-
-		case IMSG_CTL_PROFILE:
-			m_msg(&m, imsg);
-			m_get_int(&m, &v);
-			m_end(&m);
-			profiling = v;
-			return;
 		}
 	}
 
 	errx(1, "mda_imsg: unexpected %s imsg", imsg_to_str(imsg->hdr.type));
 }
 
-static void
-mda_sig_handler(int sig, short event, void *p)
+void
+mda_postfork()
 {
-	switch (sig) {
-	case SIGINT:
-	case SIGTERM:
-		mda_shutdown();
-		break;
-	default:
-		fatalx("mda_sig_handler: unexpected signal");
-	}
 }
 
-static void
-mda_shutdown(void)
+void
+mda_postprivdrop()
 {
-	log_info("info: mail delivery agent exiting");
-	_exit(0);
-}
-
-pid_t
-mda(void)
-{
-	pid_t		 pid;
-	struct passwd	*pw;
-	struct event	 ev_sigint;
-	struct event	 ev_sigterm;
-
-	switch (pid = fork()) {
-	case -1:
-		fatal("mda: cannot fork");
-	case 0:
-		post_fork(PROC_MDA);
-		break;
-	default:
-		return (pid);
-	}
-
-	purge_config(PURGE_EVERYTHING);
-
-	if ((pw = getpwnam(SMTPD_USER)) == NULL)
-		fatalx("unknown user " SMTPD_USER);
-
-	if (chroot(PATH_CHROOT) == -1)
-		fatal("mda: chroot");
-	if (chdir("/") == -1)
-		fatal("mda: chdir(\"/\")");
-
-	config_process(PROC_MDA);
-
-	if (setgroups(1, &pw->pw_gid) ||
-	    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) ||
-	    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid))
-		fatal("mda: cannot drop privileges");
-
 	tree_init(&sessions);
 	tree_init(&users);
 	TAILQ_INIT(&runnable);
-
-	imsg_callback = mda_imsg;
-	event_init();
-
-	signal_set(&ev_sigint, SIGINT, mda_sig_handler, NULL);
-	signal_set(&ev_sigterm, SIGTERM, mda_sig_handler, NULL);
-	signal_add(&ev_sigint, NULL);
-	signal_add(&ev_sigterm, NULL);
-	signal(SIGPIPE, SIG_IGN);
-	signal(SIGHUP, SIG_IGN);
-
-	config_peer(PROC_PARENT);
-	config_peer(PROC_QUEUE);
-	config_peer(PROC_LKA);
-	config_peer(PROC_CONTROL);
-	config_done();
-
-	if (event_dispatch() < 0)
-		fatal("event_dispatch");
-	mda_shutdown();
-
-	return (0);
 }
 
 static void
@@ -530,7 +507,7 @@ mda_io(struct io *io, int evt)
 			if ((ln = fgetln(s->datafp, &len)) == NULL)
 				break;
 			if (iobuf_queue(&s->iobuf, ln, len) == -1) {
-				m_create(p_parent, IMSG_PARENT_KILL_MDA,
+				m_create(p_parent, IMSG_MDA_KILL,
 				    0, 0, -1);
 				m_add_id(p_parent, s->id);
 				m_add_string(p_parent, "Out of memory");
@@ -548,7 +525,7 @@ mda_io(struct io *io, int evt)
 		if (ferror(s->datafp)) {
 			log_debug("debug: mda: ferror on session %016"PRIx64,
 			    s->id);
-			m_create(p_parent, IMSG_PARENT_KILL_MDA, 0, 0, -1);
+			m_create(p_parent, IMSG_MDA_KILL, 0, 0, -1);
 			m_add_id(p_parent, s->id);
 			m_add_string(p_parent, "Error reading body");
 			m_close(p_parent);
@@ -637,7 +614,7 @@ static int
 mda_getlastline(int fd, char *dst, size_t dstsz)
 {
 	FILE	*fp;
-	char	*ln, buf[SMTPD_MAXLINESIZE];
+	char	*ln, buf[LINE_MAX];
 	size_t	 len;
 
 	memset(buf, 0, sizeof buf);
@@ -665,9 +642,9 @@ mda_getlastline(int fd, char *dst, size_t dstsz)
 	fclose(fp);
 
 	if (buf[0]) {
-		strlcpy(dst, "\"", dstsz);
-		strnvis(dst + 1, buf, dstsz - 2, VIS_SAFE | VIS_CSTYLE);
-		strlcat(dst, "\"", dstsz);
+		(void)strlcpy(dst, "\"", dstsz);
+		(void)strnvis(dst + 1, buf, dstsz - 2, VIS_SAFE | VIS_CSTYLE);
+		(void)strlcat(dst, "\"", dstsz);
 	}
 
 	return (0);
@@ -682,11 +659,11 @@ mda_fail(struct mda_user *user, int permfail, const char *error, enum enhanced_s
 		TAILQ_REMOVE(&user->envelopes, e, entry);
 		if (permfail) {
 			mda_log(e, "PermFail", error);
-			queue_permfail(e->id, error, code);
+			mda_queue_permfail(e->id, error, code);
 		}
 		else {
 			mda_log(e, "TempFail", error);
-			queue_tempfail(e->id, error, code);
+			mda_queue_tempfail(e->id, error, code);
 		}
 		mda_envelope_free(e);
 	}
@@ -741,7 +718,7 @@ mda_drain(void)
 				u->flags &= ~USER_ONHOLD;
 			}
 			if (u->flags & USER_HOLDQ) {
-				m_create(p_queue, IMSG_DELIVERY_RELEASE, 0, 0, -1);
+				m_create(p_queue, IMSG_MDA_HOLDQ_RELEASE, 0, 0, -1);
 				m_add_id(p_queue, u->id);
 				m_add_int(p_queue, env->sc_mda_task_release);
 				m_close(p_queue);
@@ -785,12 +762,12 @@ mda_done(struct mda_session *s)
 static void
 mda_log(const struct mda_envelope *evp, const char *prefix, const char *status)
 {
-	char rcpt[SMTPD_MAXLINESIZE];
+	char rcpt[LINE_MAX];
 	const char *method;
 
 	rcpt[0] = '\0';
 	if (evp->rcpt)
-		snprintf(rcpt, sizeof rcpt, "rcpt=<%s>, ", evp->rcpt);
+		(void)snprintf(rcpt, sizeof rcpt, "rcpt=<%s>, ", evp->rcpt);
 
 	if (evp->method == A_MAILDIR)
 		method = "maildir";
@@ -818,6 +795,42 @@ mda_log(const struct mda_envelope *evp, const char *prefix, const char *status)
 	    status);
 }
 
+static void
+mda_queue_ok(uint64_t evpid)
+{
+	m_create(p_queue, IMSG_MDA_DELIVERY_OK, 0, 0, -1);
+	m_add_evpid(p_queue, evpid);
+	m_close(p_queue);
+}
+
+static void
+mda_queue_tempfail(uint64_t evpid, const char *reason, enum enhanced_status_code code)
+{
+	m_create(p_queue, IMSG_MDA_DELIVERY_TEMPFAIL, 0, 0, -1);
+	m_add_evpid(p_queue, evpid);
+	m_add_string(p_queue, reason);
+	m_add_int(p_queue, (int)code);
+	m_close(p_queue);
+}
+
+static void
+mda_queue_permfail(uint64_t evpid, const char *reason, enum enhanced_status_code code)
+{
+	m_create(p_queue, IMSG_MDA_DELIVERY_PERMFAIL, 0, 0, -1);
+	m_add_evpid(p_queue, evpid);
+	m_add_string(p_queue, reason);
+	m_add_int(p_queue, (int)code);
+	m_close(p_queue);
+}
+
+static void
+mda_queue_loop(uint64_t evpid)
+{
+	m_create(p_queue, IMSG_MDA_DELIVERY_LOOP, 0, 0, -1);
+	m_add_evpid(p_queue, evpid);
+	m_close(p_queue);
+}
+
 static struct mda_user *
 mda_user(const struct envelope *evp)
 {
@@ -834,22 +847,29 @@ mda_user(const struct envelope *evp)
 	u = xcalloc(1, sizeof *u, "mda_user");
 	u->id = generate_uid();
 	TAILQ_INIT(&u->envelopes);
-	strlcpy(u->name, evp->agent.mda.username, sizeof(u->name));
-	strlcpy(u->usertable, evp->agent.mda.usertable,
+	(void)strlcpy(u->name, evp->agent.mda.username, sizeof(u->name));
+	(void)strlcpy(u->usertable, evp->agent.mda.usertable,
 	    sizeof(u->usertable));
 
 	tree_xset(&users, u->id, u);
 
-	m_create(p_lka, IMSG_LKA_USERINFO, 0, 0, -1);
+	m_create(p_lka, IMSG_MDA_LOOKUP_USERINFO, 0, 0, -1);
 	m_add_id(p_lka, u->id);
 	m_add_string(p_lka, evp->agent.mda.usertable);
-	m_add_string(p_lka, evp->agent.mda.username);
+	if (evp->agent.mda.delivery_user[0])
+		m_add_string(p_lka, evp->agent.mda.delivery_user);
+	else
+		m_add_string(p_lka, evp->agent.mda.username);
 	m_close(p_lka);
 	u->flags |= USER_WAITINFO;
 
 	stat_increment("mda.user", 1);
 
-	log_debug("mda: new user %llx for \"%s\"", u->id, mda_user_to_text(u));
+	if (evp->agent.mda.delivery_user[0])
+		log_debug("mda: new user %016" PRIx64 " for \"%s\" delivering as \"%s\"",
+		    u->id, mda_user_to_text(u), evp->agent.mda.delivery_user);
+	else
+		log_debug("mda: new user %016" PRIx64 " for \"%s\"", u->id, mda_user_to_text(u));
 
 	return (u);
 }
@@ -860,7 +880,7 @@ mda_user_free(struct mda_user *u)
 	tree_xpop(&users, u->id);
 
 	if (u->flags & USER_HOLDQ) {
-		m_create(p_queue, IMSG_DELIVERY_RELEASE, 0, 0, -1);
+		m_create(p_queue, IMSG_MDA_HOLDQ_RELEASE, 0, 0, -1);
 		m_add_id(p_queue, u->id);
 		m_add_int(p_queue, 0);
 		m_close(p_queue);
@@ -875,7 +895,7 @@ mda_user_to_text(const struct mda_user *u)
 {
 	static char buf[1024];
 
-	snprintf(buf, sizeof(buf), "%s:%s", u->usertable, u->name);
+	(void)snprintf(buf, sizeof(buf), "%s:%s", u->usertable, u->name);
 
 	return (buf);
 }
@@ -884,19 +904,19 @@ static struct mda_envelope *
 mda_envelope(const struct envelope *evp)
 {
 	struct mda_envelope	*e;
-	char			 buf[SMTPD_MAXLINESIZE];
+	char			 buf[LINE_MAX];
 
 	e = xcalloc(1, sizeof *e, "mda_envelope");
 	e->id = evp->id;
 	e->creation = evp->creation;
 	buf[0] = '\0';
 	if (evp->sender.user[0] && evp->sender.domain[0])
-		snprintf(buf, sizeof buf, "%s@%s",
+		(void)snprintf(buf, sizeof buf, "%s@%s",
 		    evp->sender.user, evp->sender.domain);
 	e->sender = xstrdup(buf, "mda_envelope:sender");
-	snprintf(buf, sizeof buf, "%s@%s", evp->dest.user, evp->dest.domain);
+	(void)snprintf(buf, sizeof buf, "%s@%s", evp->dest.user, evp->dest.domain);
 	e->dest = xstrdup(buf, "mda_envelope:dest");
-	snprintf(buf, sizeof buf, "%s@%s", evp->rcpt.user, evp->rcpt.domain);
+	(void)snprintf(buf, sizeof buf, "%s@%s", evp->rcpt.user, evp->rcpt.domain);
 	if (strcmp(buf, e->dest))
 		e->rcpt = xstrdup(buf, "mda_envelope:rcpt");
 	e->method = evp->agent.mda.method;
@@ -947,7 +967,7 @@ mda_session(struct mda_user * u)
 	    " for user \"%s\" evpid %016" PRIx64, s->id,
 	    mda_user_to_text(u), s->evp->id);
 
-	m_create(p_queue, IMSG_QUEUE_MESSAGE_FD, 0, 0, -1);
+	m_create(p_queue, IMSG_MDA_OPEN_MESSAGE, 0, 0, -1);
 	m_add_id(p_queue, s->id);
 	m_add_msgid(p_queue, evpid_to_msgid(s->evp->id));
 	m_close(p_queue);
