@@ -37,6 +37,7 @@
 #include <openssl/ssl.h>
 #include <pwd.h>
 #include <resolv.h>
+#include <limits.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -82,6 +83,7 @@ enum mta_state {
 #define MTA_WANT_SECURE		0x0010
 #define MTA_USE_AUTH		0x0020
 #define MTA_USE_CERT		0x0040
+#define MTA_DOWNGRADE_PLAIN    	0x0080
 
 #define MTA_TLS_TRIED		0x0080
 
@@ -92,6 +94,7 @@ enum mta_state {
 #define MTA_LMTP		0x0800
 #define MTA_WAIT		0x1000
 #define MTA_HANGON		0x2000
+#define MTA_RECONN		0x4000
 
 #define MTA_EXT_STARTTLS	0x01
 #define MTA_EXT_PIPELINING	0x02
@@ -128,6 +131,8 @@ struct mta_session {
 	FILE			*datafp;
 
 	size_t			 failures;
+
+	char			 replybuf[2048];
 };
 
 static void mta_session_init(void);
@@ -144,7 +149,6 @@ static void mta_send(struct mta_session *, char *, ...);
 static ssize_t mta_queue_data(struct mta_session *);
 static void mta_response(struct mta_session *, char *);
 static const char * mta_strstate(int);
-static int mta_check_loop(FILE *);
 static void mta_start_tls(struct mta_session *);
 static int mta_verify_certificate(struct mta_session *);
 static struct mta_session *mta_tree_pop(struct tree *, uint64_t);
@@ -220,6 +224,9 @@ mta_session(struct mta_relay *relay, struct mta_route *route)
 			s->flags |= MTA_FORCE_PLAIN;
 	}
 
+	if (relay->flags & RELAY_BACKUP)
+		s->flags &= ~MTA_FORCE_PLAIN;
+
 	log_debug("debug: mta: %p: spawned for relay %s", s,
 	    mta_relay_to_text(relay));
 	stat_increment("mta.session", 1);
@@ -233,7 +240,10 @@ mta_session(struct mta_relay *relay, struct mta_route *route)
 		evtimer_set(&s->io.ev, mta_start, s);
 		evtimer_add(&s->io.ev, &tv);
 	} else if (waitq_wait(&route->dst->ptrname, mta_on_ptr, s)) {
-		dns_query_ptr(s->id, s->route->dst->sa);
+		m_create(p_lka,  IMSG_MTA_DNS_PTR, 0, 0, -1);
+		m_add_id(p_lka, s->id);
+		m_add_sockaddr(p_lka, s->route->dst->sa);
+		m_close(p_lka);
 		tree_xset(&wait_ptr, s->id, s);
 		s->flags |= MTA_WAIT;
 	}
@@ -254,7 +264,7 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 
 	switch (imsg->hdr.type) {
 
-	case IMSG_QUEUE_MESSAGE_FD:
+	case IMSG_MTA_OPEN_MESSAGE:
 		m_msg(&m, imsg);
 		m_get_id(&m, &reqid);
 		m_end(&m);
@@ -268,7 +278,7 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 
 		if (imsg->fd == -1) {
 			log_debug("debug: mta: failed to obtain msg fd");
-			mta_flush_task(s, IMSG_DELIVERY_TEMPFAIL,
+			mta_flush_task(s, IMSG_MTA_DELIVERY_TEMPFAIL,
 			    "Could not get message fd", 0, 0);
 			mta_enter_state(s, MTA_READY);
 			io_reload(&s->io);
@@ -279,20 +289,11 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 		if (s->datafp == NULL)
 			fatal("mta: fdopen");
 
-		if (mta_check_loop(s->datafp)) {
-			log_debug("debug: mta: loop detected");
-			fclose(s->datafp);
-			s->datafp = NULL;
-			mta_flush_task(s, IMSG_DELIVERY_LOOP,
-			    "Loop detected", 0, 0);
-			mta_enter_state(s, MTA_READY);
-		} else {
-			mta_enter_state(s, MTA_MAIL);
-		}
+		mta_enter_state(s, MTA_MAIL);
 		io_reload(&s->io);
 		return;
 
-	case IMSG_DNS_PTR:
+	case IMSG_MTA_DNS_PTR:
 		m_msg(&m, imsg);
 		m_get_id(&m, &reqid);
 		m_get_int(&m, &dnserror);
@@ -312,7 +313,7 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 		waitq_run(&h->ptrname, h->ptrname);
 		return;
 
-	case IMSG_LKA_SSL_INIT:
+	case IMSG_MTA_TLS_INIT:
 		resp_ca_cert = imsg->data;
 		s = mta_tree_pop(&wait_ssl_init, resp_ca_cert->reqid);
 		if (s == NULL)
@@ -320,13 +321,13 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 
 		if (resp_ca_cert->status == CA_FAIL) {
 			if (s->relay->pki_name) {
-				log_info("smtp-out: Disconnecting session %016"PRIx64
-				    ": CA failure", s->id);
+				log_info("smtp-out: session %016"PRIx64
+				    ": disconnected (CA failure)", s->id);
 				mta_free(s);
 				return;
 			}
 			else {
-				ssl = ssl_mta_init(NULL, 0, NULL, 0);
+				ssl = ssl_mta_init(NULL, NULL, 0, env->sc_tls_ciphers);
 				if (ssl == NULL)
 					fatal("mta: ssl_mta_init");
 				io_start_tls(&s->io, ssl);
@@ -337,28 +338,21 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 		resp_ca_cert = xmemdup(imsg->data, sizeof *resp_ca_cert, "mta:ca_cert");
 		resp_ca_cert->cert = xstrdup((char *)imsg->data +
 		    sizeof *resp_ca_cert, "mta:ca_cert");
-		resp_ca_cert->key = xstrdup((char *)imsg->data +
-		    sizeof *resp_ca_cert + resp_ca_cert->cert_len,
-		    "mta:ca_key");
-		ssl = ssl_mta_init(resp_ca_cert->cert, resp_ca_cert->cert_len,
-		    resp_ca_cert->key, resp_ca_cert->key_len);
+		ssl = ssl_mta_init(resp_ca_cert->name,
+		    resp_ca_cert->cert, resp_ca_cert->cert_len, env->sc_tls_ciphers);
 		if (ssl == NULL)
 			fatal("mta: ssl_mta_init");
 		io_start_tls(&s->io, ssl);
-
-		memset(resp_ca_cert->cert, 0, resp_ca_cert->cert_len);
-		memset(resp_ca_cert->key, 0, resp_ca_cert->key_len);
+		explicit_bzero(resp_ca_cert->cert, resp_ca_cert->cert_len);
 		free(resp_ca_cert->cert);
-		free(resp_ca_cert->key);
 		free(resp_ca_cert);
 		return;
 
-	case IMSG_LKA_SSL_VERIFY:
+	case IMSG_MTA_TLS_VERIFY:
 		resp_ca_vrfy = imsg->data;
 		s = mta_tree_pop(&wait_ssl_verify, resp_ca_vrfy->reqid);
 		if (s == NULL)
 			return;
-
 		if (resp_ca_vrfy->status == CA_OK)
 			s->flags |= MTA_VERIFIED;
 		else if (s->relay->flags & F_TLS_VERIFY) {
@@ -367,13 +361,12 @@ mta_session_imsg(struct mproc *p, struct imsg *imsg)
 			mta_free(s);
 			return;
 		}
-
 		mta_io(&s->io, IO_TLSVERIFIED);
 		io_resume(&s->io, IO_PAUSE_IN);
 		io_reload(&s->io);
 		return;
 
-	case IMSG_LKA_HELO:
+	case IMSG_MTA_LOOKUP_HELO:
 		m_msg(&m, imsg);
 		m_get_id(&m, &reqid);
 		m_get_int(&m, &status);
@@ -490,7 +483,7 @@ mta_connect(struct mta_session *s)
 
 	if (s->helo == NULL) {
 		if (s->relay->helotable && s->route->src->sa) {
-			m_create(p_lka, IMSG_LKA_HELO, 0, 0, -1);
+			m_create(p_lka, IMSG_MTA_LOOKUP_HELO, 0, 0, -1);
 			m_add_id(p_lka, s->id);
 			m_add_string(p_lka, s->relay->helotable);
 			m_add_sockaddr(p_lka, s->route->src->sa);
@@ -524,6 +517,10 @@ mta_connect(struct mta_session *s)
 			s->use_smtps = 1;	/* tls+smtps */
 			break;
 		}
+		else if (s->flags & MTA_DOWNGRADE_PLAIN) {
+			/* smtp+tls, with tls failure */
+			break;
+		}
 	default:
 		mta_free(s);
 		return;
@@ -543,7 +540,6 @@ mta_connect(struct mta_session *s)
 		((struct sockaddr_in6 *)sa)->sin6_port = htons(portno);
 
 	s->attempt += 1;
-
 	if (s->use_smtp_tls)
 		schema = "smtp+tls://";
 	else if (s->use_starttls)
@@ -555,9 +551,10 @@ mta_connect(struct mta_session *s)
 	else
 		schema = "smtp://";
 
-	log_info("smtp-out: Connecting to %s%s:%d (%s) on session"
-	    " %016"PRIx64"...", schema, sa_to_text(s->route->dst->sa),
-	    portno, s->route->dst->ptrname, s->id);
+	log_info("smtp-out: session %016" PRIx64
+	    ": connecting to %s%s:%d (%s)",
+	    s->id, schema, sa_to_text(s->route->dst->sa),
+	    portno, s->route->dst->ptrname);
 
 	mta_enter_state(s, MTA_INIT);
 	iobuf_xinit(&s->iobuf, 0, 0, "mta_connect");
@@ -584,8 +581,8 @@ mta_enter_state(struct mta_session *s, int newstate)
 	size_t			 envid_sz;
 	int			 oldstate;
 	ssize_t			 q;
-	char			 ibuf[SMTPD_MAXLINESIZE];
-	char			 obuf[SMTPD_MAXLINESIZE];
+	char			 ibuf[LINE_MAX];
+	char			 obuf[LINE_MAX];
 	int			 offset;
 
     again:
@@ -597,6 +594,8 @@ mta_enter_state(struct mta_session *s, int newstate)
 
 	s->state = newstate;
 
+	memset(s->replybuf, 0, sizeof s->replybuf);
+
 	/* don't try this at home! */
 #define mta_enter_state(_s, _st) do { newstate = _st; goto again; } while (0)
 
@@ -621,12 +620,14 @@ mta_enter_state(struct mta_session *s, int newstate)
 		break;
 
 	case MTA_STARTTLS:
+		if (s->flags & MTA_DOWNGRADE_PLAIN)
+			mta_enter_state(s, MTA_AUTH);
 		if (s->flags & MTA_TLS) /* already started */
 			mta_enter_state(s, MTA_AUTH);
 		else if ((s->ext & MTA_EXT_STARTTLS) == 0) {
 			if (s->flags & MTA_FORCE_TLS || s->flags & MTA_WANT_SECURE) {
 				mta_error(s, "TLS required but not supported by remote host");
-				mta_connect(s);
+				s->flags |= MTA_RECONN;
 			}
 			else
 				/* server doesn't support starttls, do not use it */
@@ -717,7 +718,8 @@ mta_enter_state(struct mta_session *s, int newstate)
 		}
 
 		if (s->msgtried >= MAX_TRYBEFOREDISABLE) {
-			log_info("smtp-out: Remote host seems to reject all mails on session %016"PRIx64,
+			log_info("smtp-out: session %016"PRIx64
+			    "remote host seems to reject all mails",
 			    s->id);
 			mta_route_down(s->relay, s->route);
 			mta_enter_state(s, MTA_QUIT);
@@ -732,7 +734,17 @@ mta_enter_state(struct mta_session *s, int newstate)
 			break;
 		}
 
-		s->task = mta_route_next_task(s->relay, s->route);
+		/*
+		 * When downgrading from opportunistic TLS, clear flag and
+		 * possibly reuse the same task (forbidden in other cases).
+		 */
+		if (s->flags & MTA_DOWNGRADE_PLAIN)
+			s->flags &= ~MTA_DOWNGRADE_PLAIN;
+		else if (s->task)
+			fatalx("task should be NULL at this point");
+
+		if (s->task == NULL)
+			s->task = mta_route_next_task(s->relay, s->route);
 		if (s->task == NULL) {
 			log_debug("debug: mta: %p: no task for relay %s",
 			    s, mta_relay_to_text(s->relay));
@@ -756,7 +768,7 @@ mta_enter_state(struct mta_session *s, int newstate)
 
 		stat_increment("mta.task.running", 1);
 
-		m_create(p_queue, IMSG_QUEUE_MESSAGE_FD, 0, 0, -1);
+		m_create(p_queue, IMSG_MTA_OPEN_MESSAGE, 0, 0, -1);
 		m_add_id(p_queue, s->id);
 		m_add_msgid(p_queue, s->task->msgid);
 		m_close(p_queue);
@@ -822,7 +834,7 @@ mta_enter_state(struct mta_session *s, int newstate)
 			break;
 		}
 
-		log_trace(TRACE_MTA, "mta: %p: >>> [...%zi bytes...]", s, q);
+		log_trace(TRACE_MTA, "mta: %p: >>> [...%zd bytes...]", s, q);
 		break;
 
 	case MTA_EOM:
@@ -863,12 +875,17 @@ mta_response(struct mta_session *s, char *line)
 	struct sockaddr		*sa;
 	const char		*domain;
 	socklen_t		 sa_len;
-	char			 buf[SMTPD_MAXLINESIZE];
+	char			 buf[LINE_MAX];
 	int			 delivery;
 
 	switch (s->state) {
 
 	case MTA_BANNER:
+		if (line[0] != '2') {
+			mta_error(s, "BANNER rejected: %s", line);
+			s->flags |= MTA_FREE;
+			return;
+		}
 		if (s->flags & MTA_LMTP)
 			mta_enter_state(s, MTA_LHLO);
 		else
@@ -965,9 +982,9 @@ mta_response(struct mta_session *s, char *line)
 	case MTA_MAIL:
 		if (line[0] != '2') {
 			if (line[0] == '5')
-				delivery = IMSG_DELIVERY_PERMFAIL;
+				delivery = IMSG_MTA_DELIVERY_PERMFAIL;
 			else
-				delivery = IMSG_DELIVERY_TEMPFAIL;
+				delivery = IMSG_MTA_DELIVERY_TEMPFAIL;
 			mta_flush_task(s, delivery, line, 0, 0);
 			mta_enter_state(s, MTA_RSET);
 			return;
@@ -996,9 +1013,9 @@ mta_response(struct mta_session *s, char *line)
 		}
 		else {
 			if (line[0] == '5')
-				delivery = IMSG_DELIVERY_PERMFAIL;
+				delivery = IMSG_MTA_DELIVERY_PERMFAIL;
 			else
-				delivery = IMSG_DELIVERY_TEMPFAIL;
+				delivery = IMSG_MTA_DELIVERY_TEMPFAIL;
 			s->failures++;
 
 			/* remove failed envelope from task list */
@@ -1006,7 +1023,7 @@ mta_response(struct mta_session *s, char *line)
 			stat_decrement("mta.envelope", 1);
 
 			/* log right away */
-			snprintf(buf, sizeof(buf), "%s",
+			(void)snprintf(buf, sizeof(buf), "%s",
 			    mta_host_to_text(s->route->dst));
 
 			e->session = s->id;
@@ -1029,7 +1046,7 @@ mta_response(struct mta_session *s, char *line)
 
 			if (s->relay->limits->max_failures_per_session &&
 			    s->failures == s->relay->limits->max_failures_per_session) {
-					mta_flush_task(s, IMSG_DELIVERY_TEMPFAIL,
+					mta_flush_task(s, IMSG_MTA_DELIVERY_TEMPFAIL,
 					    "Too many consecutive errors, closing connection", 0, 1);
 					mta_enter_state(s, MTA_QUIT);
 					break;
@@ -1039,7 +1056,7 @@ mta_response(struct mta_session *s, char *line)
 			 * if no more envelopes, flush failed queue
 			 */
 			if (TAILQ_EMPTY(&s->task->envelopes)) {
-				mta_flush_task(s, IMSG_DELIVERY_OK,
+				mta_flush_task(s, IMSG_MTA_DELIVERY_OK,
 				    "No envelope", 0, 0);
 				mta_enter_state(s, MTA_RSET);
 				break;
@@ -1058,9 +1075,9 @@ mta_response(struct mta_session *s, char *line)
 			break;
 		}
 		if (line[0] == '5')
-			delivery = IMSG_DELIVERY_PERMFAIL;
+			delivery = IMSG_MTA_DELIVERY_PERMFAIL;
 		else
-			delivery = IMSG_DELIVERY_TEMPFAIL;
+			delivery = IMSG_MTA_DELIVERY_TEMPFAIL;
 		mta_flush_task(s, delivery, line, 0, 0);
 		mta_enter_state(s, MTA_RSET);
 		break;
@@ -1068,14 +1085,14 @@ mta_response(struct mta_session *s, char *line)
 	case MTA_LMTP_EOM:
 	case MTA_EOM:
 		if (line[0] == '2') {
-			delivery = IMSG_DELIVERY_OK;
+			delivery = IMSG_MTA_DELIVERY_OK;
 			s->msgtried = 0;
 			s->msgcount++;
 		}
 		else if (line[0] == '5')
-			delivery = IMSG_DELIVERY_PERMFAIL;
+			delivery = IMSG_MTA_DELIVERY_PERMFAIL;
 		else
-			delivery = IMSG_DELIVERY_TEMPFAIL;
+			delivery = IMSG_MTA_DELIVERY_TEMPFAIL;
 		mta_flush_task(s, delivery, line, (s->flags & MTA_LMTP) ? 1 : 0, 0);
 		if (s->task) {
 			s->rcptcount--;
@@ -1132,8 +1149,6 @@ mta_io(struct io *io, int evt)
 	switch (evt) {
 
 	case IO_CONNECTED:
-		log_info("smtp-out: Connected on session %016"PRIx64, s->id);
-
 		if (s->use_smtps) {
 			io_set_write(io);
 			mta_start_tls(s);
@@ -1145,7 +1160,7 @@ mta_io(struct io *io, int evt)
 		break;
 
 	case IO_TLSREADY:
-		log_info("smtp-out: Started TLS on session %016"PRIx64": %s",
+		log_info("smtp-out: session %016"PRIx64": TLS started %s",
 		    s->id, ssl_to_text(s->io.ssl));
 		s->flags |= MTA_TLS;
 
@@ -1157,10 +1172,9 @@ mta_io(struct io *io, int evt)
 	case IO_TLSVERIFIED:
 		x = SSL_get_peer_certificate(s->io.ssl);
 		if (x) {
-			log_info("smtp-out: Server certificate verification %s "
-			    "on session %016"PRIx64,
-			    (s->flags & MTA_VERIFIED) ? "succeeded" : "failed",
-			    s->id);
+			log_info("smtp-out: session %016"PRIx64
+			    ": server certificate verification %s",
+			    s->id, (s->flags & MTA_VERIFIED) ? "succeeded" : "failed");
 			X509_free(x);
 		}
 
@@ -1176,7 +1190,7 @@ mta_io(struct io *io, int evt)
 	    nextline:
 		line = iobuf_getline(&s->iobuf, &len);
 		if (line == NULL) {
-			if (iobuf_len(&s->iobuf) >= SMTPD_MAXLINESIZE) {
+			if (iobuf_len(&s->iobuf) >= LINE_MAX) {
 				mta_error(s, "Input too long");
 				mta_free(s);
 				return;
@@ -1212,22 +1226,56 @@ mta_io(struct io *io, int evt)
 				s->ext |= MTA_EXT_DSN;
 		}
 
-		if (cont)
+		/* continuation reply, we parse out the repeating statuses and ESC */
+		if (cont) {
+			if (s->replybuf[0] == '\0')
+				(void)strlcat(s->replybuf, line, sizeof s->replybuf);
+			else {
+				line = line + 4;
+				if (isdigit((int)*line) && *(line + 1) == '.' &&
+				    isdigit((int)*line+2) && *(line + 3) == '.' &&
+				    isdigit((int)*line+4) && isspace((int)*(line + 5)))
+					(void)strlcat(s->replybuf, line+5, sizeof s->replybuf);
+				else
+					(void)strlcat(s->replybuf, line, sizeof s->replybuf);
+			}
 			goto nextline;
+		}
+
+		/* last line of a reply, check if we're on a continuation to parse out status and ESC.
+		 * if we overflow reply buffer or are not on continuation, log entire last line.
+		 */
+		if (s->replybuf[0] != '\0') {
+			p = line + 4;
+			if (isdigit((int)*p) && *(p + 1) == '.' &&
+			    isdigit((int)*p+2) && *(p + 3) == '.' &&
+			    isdigit((int)*p+4) && isspace((int)*(p + 5)))
+				p += 5;
+			if (strlcat(s->replybuf, p, sizeof s->replybuf) >= sizeof s->replybuf)
+				(void)strlcpy(s->replybuf, line, sizeof s->replybuf);
+		}
+		else
+			(void)strlcpy(s->replybuf, line, sizeof s->replybuf);
+
 
 		if (s->state == MTA_QUIT) {
-			log_info("smtp-out: Closing session %016"PRIx64
-			    ": %zu message%s sent.", s->id, s->msgcount,
+			log_info("smtp-out: session %016"PRIx64
+			    ": connection closed, %zu message%s sent.", s->id, s->msgcount,
 			    (s->msgcount > 1) ? "s" : "");
 			mta_free(s);
 			return;
 		}
 		io_set_write(io);
-		mta_response(s, line);
+		mta_response(s, s->replybuf);
 		if (s->flags & MTA_FREE) {
 			mta_free(s);
 			return;
 		}
+		if (s->flags & MTA_RECONN) {
+			s->flags &= ~MTA_RECONN;
+			mta_connect(s);
+			return;
+		}
 
 		iobuf_normalize(&s->iobuf);
 
@@ -1265,11 +1313,41 @@ mta_io(struct io *io, int evt)
 
 	case IO_ERROR:
 		log_debug("debug: mta: %p: IO error: %s", s, io->error);
-		mta_error(s, "IO Error: %s", io->error);
-		if (!s->ready)
+		if (!s->ready) {
+			mta_error(s, "IO Error: %s", io->error);
 			mta_connect(s);
-		else
-			mta_free(s);
+			break;
+		}
+		else if (!(s->flags & (MTA_FORCE_TLS|MTA_FORCE_ANYSSL))) {
+			/* error in non-strict SSL negotiation, downgrade to plain */
+			if (s->flags & MTA_TLS) {
+				log_info("smtp-out: session %016"PRIx64
+				    ": error: opportunistic TLS failed, "
+				    "downgrading to plain", s->id);
+				s->flags &= ~MTA_TLS;
+				s->flags |= MTA_DOWNGRADE_PLAIN;
+				mta_connect(s);
+				break;
+			}
+		}
+		mta_error(s, "IO Error: %s", io->error);
+		mta_free(s);
+		break;
+
+	case IO_TLSERROR:
+		log_debug("debug: mta: %p: TLS IO error: %s", s, io->error);
+		if (!(s->flags & (MTA_FORCE_TLS|MTA_FORCE_ANYSSL))) {
+			/* error in non-strict SSL negotiation, downgrade to plain */
+			log_info("smtp-out: session %016"PRIx64
+			    ": error: TLS failed, "
+			    "downgrading to plain", s->id);
+			s->flags &= ~MTA_TLS;
+			s->flags |= MTA_DOWNGRADE_PLAIN;
+			mta_connect(s);
+			break;
+		}
+		mta_error(s, "IO Error: %s", io->error);
+		mta_free(s);
 		break;
 
 	case IO_DISCONNECTED:
@@ -1327,7 +1405,7 @@ mta_queue_data(struct mta_session *s)
 	}
 
 	if (ferror(s->datafp)) {
-		mta_flush_task(s, IMSG_DELIVERY_TEMPFAIL,
+		mta_flush_task(s, IMSG_MTA_DELIVERY_TEMPFAIL,
 		    "Error reading content file", 0, 0);
 		return (-1);
 	}
@@ -1345,14 +1423,14 @@ mta_flush_task(struct mta_session *s, int delivery, const char *error, size_t co
 	int cache)
 {
 	struct mta_envelope	*e;
-	char			 relay[SMTPD_MAXLINESIZE];
+	char			 relay[LINE_MAX];
 	size_t			 n;
 	struct sockaddr_storage	 ss;
 	struct sockaddr		*sa;
 	socklen_t		 sa_len;
 	const char		*domain;
 
-	snprintf(relay, sizeof relay, "%s", mta_host_to_text(s->route->dst));
+	(void)snprintf(relay, sizeof relay, "%s", mta_host_to_text(s->route->dst));
 	n = 0;
 	while ((e = TAILQ_FIRST(&s->task->envelopes))) {
 
@@ -1420,11 +1498,11 @@ mta_error(struct mta_session *s, const char *fmt, ...)
 	va_end(ap);
 
 	if (s->msgcount)
-		log_info("smtp-out: Error on session %016"PRIx64
-		    " after %zu message%s sent: %s", s->id, s->msgcount,
+		log_info("smtp-out: session %016"PRIx64
+		    ": error after %zu message%s sent: %s", s->id, s->msgcount,
 		    (s->msgcount > 1) ? "s" : "", error);
 	else
-		log_info("smtp-out: Error on session %016"PRIx64 ": %s",
+		log_info("smtp-out: session %016"PRIx64 ": error: %s",
 		    s->id, error);
 	/*
 	 * If not connected yet, and the error is not local, just ignore it
@@ -1440,67 +1518,28 @@ mta_error(struct mta_session *s, const char *fmt, ...)
 	mta_route_error(s->relay, s->route);
 
 	if (s->task)
-		mta_flush_task(s, IMSG_DELIVERY_TEMPFAIL, error, 0, 0);
+		mta_flush_task(s, IMSG_MTA_DELIVERY_TEMPFAIL, error, 0, 0);
 
 	free(error);
 }
 
-static int
-mta_check_loop(FILE *fp)
-{
-	char	*buf, *lbuf;
-	size_t	 len;
-	uint32_t rcvcount = 0;
-	int	 ret = 0;
-
-	lbuf = NULL;
-	while ((buf = fgetln(fp, &len))) {
-		if (buf[len - 1] == '\n')
-			buf[len - 1] = '\0';
-		else {
-			/* EOF without EOL, copy and add the NUL */
-			lbuf = xmalloc(len + 1, "mta_check_loop");
-			memcpy(lbuf, buf, len);
-			lbuf[len] = '\0';
-			buf = lbuf;
-		}
-
-		if (strchr(buf, ':') == NULL && !isspace((unsigned char)*buf))
-			break;
-
-		if (strncasecmp("Received: ", buf, 10) == 0) {
-			rcvcount++;
-			if (rcvcount == MAX_HOPS_COUNT) {
-				ret = 1;
-				break;
-			}
-		}
-		if (lbuf) {
-			free(lbuf);
-			lbuf  = NULL;
-		}
-	}
-	if (lbuf)
-		free(lbuf);
-
-	fseek(fp, SEEK_SET, 0);
-	return ret;
-}
-
 static void
 mta_start_tls(struct mta_session *s)
 {
 	struct ca_cert_req_msg	req_ca_cert;
 	const char	       *certname;
 
-	if (s->relay->pki_name)
+	if (s->relay->pki_name) {
 		certname = s->relay->pki_name;
-	else
+		req_ca_cert.fallback = 0;
+	}
+	else {
 		certname = s->helo;
-
+		req_ca_cert.fallback = 1;
+	}
 	req_ca_cert.reqid = s->id;
-	strlcpy(req_ca_cert.name, certname, sizeof req_ca_cert.name);
-	m_compose(p_lka, IMSG_LKA_SSL_INIT, 0, 0, -1,
+	(void)strlcpy(req_ca_cert.name, certname, sizeof req_ca_cert.name);
+	m_compose(p_lka, IMSG_MTA_TLS_INIT, 0, 0, -1,
 	    &req_ca_cert, sizeof(req_ca_cert));
 	tree_xset(&wait_ssl_init, s->id, s);
 	s->flags |= MTA_WAIT;
@@ -1510,12 +1549,33 @@ mta_start_tls(struct mta_session *s)
 static int
 mta_verify_certificate(struct mta_session *s)
 {
+#define MAX_CERTS	16
+#define MAX_CERT_LEN	(MAX_IMSGSIZE - (IMSG_HEADER_SIZE + sizeof(req_ca_vrfy)))
 	struct ca_vrfy_req_msg	req_ca_vrfy;
 	struct iovec		iov[2];
 	X509		       *x;
 	STACK_OF(X509)	       *xchain;
-	int			i;
-	const char	       *pkiname;
+	const char	       *name;
+	unsigned char	       *cert_der[MAX_CERTS];
+	int			cert_len[MAX_CERTS];
+	int			i, cert_count, res;
+
+	res = 0;
+	memset(cert_der, 0, sizeof(cert_der));
+	memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
+
+	/* Send the client certificate */
+	if (s->relay->ca_name) {
+		name = s->relay->ca_name;
+		req_ca_vrfy.fallback = 0;
+	}
+	else {
+		name = s->helo;
+		req_ca_vrfy.fallback = 1;
+	}
+	if (strlcpy(req_ca_vrfy.name, name, sizeof req_ca_vrfy.name)
+	    >= sizeof req_ca_vrfy.name)
+		return 0;
 
 	x = SSL_get_peer_certificate(s->io.ssl);
 	if (x == NULL)
@@ -1530,56 +1590,83 @@ mta_verify_certificate(struct mta_session *s)
 	 *
 	 */
 
+	cert_len[0] = i2d_X509(x, &cert_der[0]);
+	X509_free(x);
+
+	if (cert_len[0] < 0) {
+		log_warnx("warn: failed to encode certificate");
+		goto end;
+	}
+	log_debug("debug: certificate 0: len=%d", cert_len[0]);
+	if (cert_len[0] > (int)MAX_CERT_LEN) {
+		log_warnx("warn: certificate too long");
+		goto end;
+	}
+
+	if (xchain) {
+		cert_count = sk_X509_num(xchain);
+		log_debug("debug: certificate chain len: %d", cert_count);
+		if (cert_count >= MAX_CERTS) {
+			log_warnx("warn: certificate chain too long");
+			goto end;
+		}
+	}
+	else
+		cert_count = 0;
+
+	for (i = 0; i < cert_count; ++i) {
+		x = sk_X509_value(xchain, i);
+		cert_len[i+1] = i2d_X509(x, &cert_der[i+1]);
+		if (cert_len[i+1] < 0) {
+			log_warnx("warn: failed to encode certificate");
+			goto end;
+		}
+		log_debug("debug: certificate %i: len=%d", i+1, cert_len[i+1]);
+		if (cert_len[i+1] > (int)MAX_CERT_LEN) {
+			log_warnx("warn: certificate too long");
+			goto end;
+		}
+	}
+
 	tree_xset(&wait_ssl_verify, s->id, s);
 	s->flags |= MTA_WAIT;
 
 	/* Send the client certificate */
-	memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
-	if (s->relay->pki_name)
-		pkiname = s->relay->pki_name;
-	else
-		pkiname = s->helo;
-	if (strlcpy(req_ca_vrfy.pkiname, pkiname, sizeof req_ca_vrfy.pkiname)
-	    >= sizeof req_ca_vrfy.pkiname)
-		return 0;
-
 	req_ca_vrfy.reqid = s->id;
-	req_ca_vrfy.cert_len = i2d_X509(x, &req_ca_vrfy.cert);
-	if (xchain)
-		req_ca_vrfy.n_chain = sk_X509_num(xchain);
+	req_ca_vrfy.cert_len = cert_len[0];
+	req_ca_vrfy.n_chain = cert_count;
 	iov[0].iov_base = &req_ca_vrfy;
 	iov[0].iov_len = sizeof(req_ca_vrfy);
-	iov[1].iov_base = req_ca_vrfy.cert;
-	iov[1].iov_len = req_ca_vrfy.cert_len;
-	m_composev(p_lka, IMSG_LKA_SSL_VERIFY_CERT, 0, 0, -1,
+	iov[1].iov_base = cert_der[0];
+	iov[1].iov_len = cert_len[0];
+	m_composev(p_lka, IMSG_MTA_TLS_VERIFY_CERT, 0, 0, -1,
 	    iov, nitems(iov));
-	free(req_ca_vrfy.cert);
-	X509_free(x);
 
-	if (xchain) {		
-		/* Send the chain, one cert at a time */
-		for (i = 0; i < sk_X509_num(xchain); ++i) {
-			memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
-			req_ca_vrfy.reqid = s->id;
-			x = sk_X509_value(xchain, i);
-			req_ca_vrfy.cert_len = i2d_X509(x, &req_ca_vrfy.cert);
-			iov[0].iov_base = &req_ca_vrfy;
-			iov[0].iov_len  = sizeof(req_ca_vrfy);
-			iov[1].iov_base = req_ca_vrfy.cert;
-			iov[1].iov_len  = req_ca_vrfy.cert_len;
-			m_composev(p_lka, IMSG_LKA_SSL_VERIFY_CHAIN, 0, 0, -1,
-			    iov, nitems(iov));
-			free(req_ca_vrfy.cert);
-		}
+	memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
+	req_ca_vrfy.reqid = s->id;
+
+	/* Send the chain, one cert at a time */
+	for (i = 0; i < cert_count; ++i) {
+		req_ca_vrfy.cert_len = cert_len[i+1];
+		iov[1].iov_base = cert_der[i+1];
+		iov[1].iov_len  = cert_len[i+1];
+		m_composev(p_lka, IMSG_MTA_TLS_VERIFY_CHAIN, 0, 0, -1,
+		    iov, nitems(iov));
 	}
 
 	/* Tell lookup process that it can start verifying, we're done */
 	memset(&req_ca_vrfy, 0, sizeof req_ca_vrfy);
 	req_ca_vrfy.reqid = s->id;
-	m_compose(p_lka, IMSG_LKA_SSL_VERIFY, 0, 0, -1,
+	m_compose(p_lka, IMSG_MTA_TLS_VERIFY, 0, 0, -1,
 	    &req_ca_vrfy, sizeof req_ca_vrfy);
 
-	return 1;
+	res = 1;
+
+    end:
+	for (i = 0; i < MAX_CERTS; ++i)
+		free(cert_der[i]);
+
+	return res;
 }
 
 static const char *
@@ -1601,17 +1688,18 @@ dsn_strnotify(uint8_t arg)
 	static char	buf[32];
 	size_t		sz;
 
+	buf[0] = '\0';
 	if (arg & DSN_SUCCESS)
-		strlcat(buf, "SUCCESS,", sizeof(buf));
+		(void)strlcat(buf, "SUCCESS,", sizeof(buf));
 
 	if (arg & DSN_FAILURE)
-		strlcat(buf, "FAILURE,", sizeof(buf));
+		(void)strlcat(buf, "FAILURE,", sizeof(buf));
 
 	if (arg & DSN_DELAY)
-		strlcat(buf, "DELAY,", sizeof(buf));
+		(void)strlcat(buf, "DELAY,", sizeof(buf));
 
 	if (arg & DSN_NEVER)
-		strlcat(buf, "NEVER,", sizeof(buf));
+		(void)strlcat(buf, "NEVER,", sizeof(buf));
 
 	/* trim trailing comma */
 	sz = strlen(buf);
