@@ -119,6 +119,7 @@ cmd_install(const char *pattern, int (*cmd)(int, struct parameter*))
 	return (n);
 }
 
+#if 0
 static void
 cmd_dump(struct node *node, int depth)
 {
@@ -132,6 +133,7 @@ cmd_dump(struct node *node, int depth)
 	TAILQ_FOREACH(n, &node->children, entry)
 		cmd_dump(n, depth + 1);
 }
+#endif
 
 static int
 cmd_check(const char *str, struct node *node, struct parameter *res)
@@ -280,8 +282,8 @@ text_to_sockaddr(struct sockaddr *sa, int family, const char *str)
 {
 	struct in_addr		 ina;
 	struct in6_addr		 in6a;
-	struct sockaddr_in	*sin;
-	struct sockaddr_in6	*sin6;
+	struct sockaddr_in	*in;
+	struct sockaddr_in6	*in6;
 	char			*cp, *str2;
 	const char		*errstr;
 
@@ -295,13 +297,13 @@ text_to_sockaddr(struct sockaddr *sa, int family, const char *str)
 		if (inet_pton(PF_INET, str, &ina) != 1)
 			return (-1);
 
-		sin = (struct sockaddr_in *)sa;
-		memset(sin, 0, sizeof *sin);
+		in = (struct sockaddr_in *)sa;
+		memset(in, 0, sizeof *in);
 #ifdef HAVE_STRUCT_SOCKADDR_IN_SIN_LEN
-		sin->sin_len = sizeof(struct sockaddr_in);
+		in->sin_len = sizeof(struct sockaddr_in);
 #endif
-		sin->sin_family = PF_INET;
-		sin->sin_addr.s_addr = ina.s_addr;
+		in->sin_family = PF_INET;
+		in->sin_addr.s_addr = ina.s_addr;
 		return (0);
 
 	case PF_INET6:
@@ -320,13 +322,13 @@ text_to_sockaddr(struct sockaddr *sa, int family, const char *str)
 		} else if (inet_pton(PF_INET6, str, &in6a) != 1)
 			return (-1);
 
-		sin6 = (struct sockaddr_in6 *)sa;
-		memset(sin6, 0, sizeof *sin6);
+		in6 = (struct sockaddr_in6 *)sa;
+		memset(in6, 0, sizeof *in6);
 #ifdef HAVE_STRUCT_SOCKADDR_IN6_SIN6_LEN
-		sin6->sin6_len = sizeof(struct sockaddr_in6);
+		in6->sin6_len = sizeof(struct sockaddr_in6);
 #endif
-		sin6->sin6_family = PF_INET6;
-		sin6->sin6_addr = in6a;
+		in6->sin6_family = PF_INET6;
+		in6->sin6_addr = in6a;
 
 		if (cp == NULL)
 			return (0);
@@ -334,10 +336,10 @@ text_to_sockaddr(struct sockaddr *sa, int family, const char *str)
 		if (IN6_IS_ADDR_LINKLOCAL(&in6a) ||
 		    IN6_IS_ADDR_MC_LINKLOCAL(&in6a) ||
 		    IN6_IS_ADDR_MC_NODELOCAL(&in6a))
-			if ((sin6->sin6_scope_id = if_nametoindex(cp)))
+			if ((in6->sin6_scope_id = if_nametoindex(cp)))
 				return (0);
 
-		sin6->sin6_scope_id = strtonum(cp, 0, UINT32_MAX, &errstr);
+		in6->sin6_scope_id = strtonum(cp, 0, UINT32_MAX, &errstr);
 		if (errstr)
 			return (-1);
 		return (0);
