@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: queue.c,v 1.162 2014/04/19 13:40:24 gilles Exp $	*/
 
 /*
  * Copyright (c) 2008 Gilles Chehade <gilles@poolp.org>
@@ -40,6 +40,7 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <limits.h>
 
 #include "smtpd.h"
 #include "log.h"
@@ -50,6 +51,7 @@ static void queue_bounce(struct envelope *, struct delivery_bounce *);
 static void queue_shutdown(void);
 static void queue_sig_handler(int, short, void *);
 static void queue_log(const struct envelope *, const char *, const char *);
+static void queue_msgid_walk(int, short, void *);
 
 static size_t	flow_agent_hiwat = 10 * 1024 * 1024;
 static size_t	flow_agent_lowat =   1 * 1024 * 1024;
@@ -65,6 +67,8 @@ static void
 queue_imsg(struct mproc *p, struct imsg *imsg)
 {
 	struct delivery_bounce	 bounce;
+	struct msg_walkinfo	*wi;
+	struct timeval		 tv;
 	struct bounce_req_msg	*req_bounce;
 	struct envelope		 evp;
 	struct msg		 m;
@@ -72,20 +76,22 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 	uint64_t		 reqid, evpid, holdq;
 	uint32_t		 msgid;
 	time_t			 nexttry;
+	size_t			 n_evp;
 	int			 fd, mta_ext, ret, v, flags, code;
+	char			 buf[sizeof(evp)];
 
 	memset(&bounce, 0, sizeof(struct delivery_bounce));
-	if (p->proc == PROC_SMTP) {
+	if (p->proc == PROC_PONY) {
 
 		switch (imsg->hdr.type) {
-		case IMSG_QUEUE_CREATE_MESSAGE:
+		case IMSG_SMTP_MESSAGE_CREATE:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_end(&m);
 
 			ret = queue_message_create(&msgid);
 
-			m_create(p, IMSG_QUEUE_CREATE_MESSAGE, 0, 0, -1);
+			m_create(p, IMSG_SMTP_MESSAGE_CREATE, 0, 0, -1);
 			m_add_id(p, reqid);
 			if (ret == 0)
 				m_add_int(p, 0);
@@ -96,20 +102,20 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			m_close(p);
 			return;
 
-		case IMSG_QUEUE_REMOVE_MESSAGE:
+		case IMSG_SMTP_MESSAGE_ROLLBACK:
 			m_msg(&m, imsg);
 			m_get_msgid(&m, &msgid);
 			m_end(&m);
 
 			queue_message_delete(msgid);
 
-			m_create(p_scheduler, IMSG_QUEUE_REMOVE_MESSAGE,
+			m_create(p_scheduler, IMSG_QUEUE_MESSAGE_ROLLBACK,
 			    0, 0, -1);
 			m_add_msgid(p_scheduler, msgid);
 			m_close(p_scheduler);
 			return;
 
-		case IMSG_QUEUE_COMMIT_MESSAGE:
+		case IMSG_SMTP_MESSAGE_COMMIT:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_get_msgid(&m, &msgid);
@@ -117,20 +123,20 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 
 			ret = queue_message_commit(msgid);
 
-			m_create(p,  IMSG_QUEUE_COMMIT_MESSAGE, 0, 0, -1);
+			m_create(p, IMSG_SMTP_MESSAGE_COMMIT, 0, 0, -1);
 			m_add_id(p, reqid);
 			m_add_int(p, (ret == 0) ? 0 : 1);
 			m_close(p);
 
 			if (ret) {
-				m_create(p_scheduler, IMSG_QUEUE_COMMIT_MESSAGE,
+				m_create(p_scheduler, IMSG_QUEUE_MESSAGE_COMMIT,
 				    0, 0, -1);
 				m_add_msgid(p_scheduler, msgid);
 				m_close(p_scheduler);
 			}
 			return;
 
-		case IMSG_QUEUE_MESSAGE_FILE:
+		case IMSG_SMTP_MESSAGE_OPEN:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_get_msgid(&m, &msgid);
@@ -138,13 +144,13 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 
 			fd = queue_message_fd_rw(msgid);
 
-			m_create(p, IMSG_QUEUE_MESSAGE_FILE, 0, 0, fd);
+			m_create(p, IMSG_SMTP_MESSAGE_OPEN, 0, 0, fd);
 			m_add_id(p, reqid);
 			m_add_int(p, (fd == -1) ? 0 : 1);
 			m_close(p);
 			return;
 
-		case IMSG_SMTP_ENQUEUE_FD:
+		case IMSG_QUEUE_SMTP_SESSION:
 			bounce_fd(imsg->fd);
 			return;
 		}
@@ -152,67 +158,76 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 
 	if (p->proc == PROC_LKA) {
 		switch (imsg->hdr.type) {
-		case IMSG_QUEUE_SUBMIT_ENVELOPE:
+		case IMSG_LKA_ENVELOPE_SUBMIT:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_get_envelope(&m, &evp);
 			m_end(&m);
-		    
+
 			if (evp.id == 0)
 				log_warnx("warn: imsg_queue_submit_envelope: evpid=0");
 			if (evpid_to_msgid(evp.id) == 0)
 				log_warnx("warn: imsg_queue_submit_envelope: msgid=0, "
 				    "evpid=%016"PRIx64, evp.id);
 			ret = queue_envelope_create(&evp);
-			m_create(p_smtp, IMSG_QUEUE_SUBMIT_ENVELOPE, 0, 0, -1);
-			m_add_id(p_smtp, reqid);
+			m_create(p_pony, IMSG_QUEUE_ENVELOPE_SUBMIT, 0, 0, -1);
+			m_add_id(p_pony, reqid);
 			if (ret == 0)
-				m_add_int(p_smtp, 0);
+				m_add_int(p_pony, 0);
 			else {
-				m_add_int(p_smtp, 1);
-				m_add_evpid(p_smtp, evp.id);
+				m_add_int(p_pony, 1);
+				m_add_evpid(p_pony, evp.id);
 			}
-			m_close(p_smtp);
+			m_close(p_pony);
 			if (ret) {
 				m_create(p_scheduler,
-				    IMSG_QUEUE_SUBMIT_ENVELOPE, 0, 0, -1);
+				    IMSG_QUEUE_ENVELOPE_SUBMIT, 0, 0, -1);
 				m_add_envelope(p_scheduler, &evp);
 				m_close(p_scheduler);
 
 			}
 			return;
 
-		case IMSG_QUEUE_COMMIT_ENVELOPES:
+		case IMSG_LKA_ENVELOPE_COMMIT:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_end(&m);
-			m_create(p_smtp, IMSG_QUEUE_COMMIT_ENVELOPES, 0, 0, -1);
-			m_add_id(p_smtp, reqid);
-			m_add_int(p_smtp, 1);
-			m_close(p_smtp);
+			m_create(p_pony, IMSG_QUEUE_ENVELOPE_COMMIT, 0, 0, -1);
+			m_add_id(p_pony, reqid);
+			m_add_int(p_pony, 1);
+			m_close(p_pony);
 			return;
 		}
 	}
 
 	if (p->proc == PROC_SCHEDULER) {
 		switch (imsg->hdr.type) {
-		case IMSG_QUEUE_REMOVE:
+		case IMSG_SCHED_ENVELOPE_REMOVE:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_end(&m);
 
+			m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_ACK, 0, 0, -1);
+			m_add_evpid(p_scheduler, evpid);
+			m_close(p_scheduler);
+
 			/* already removed by scheduler */
 			if (queue_envelope_load(evpid, &evp) == 0)
 				return;
+
 			queue_log(&evp, "Remove", "Removed by administrator");
 			queue_envelope_delete(evpid);
 			return;
 
-		case IMSG_QUEUE_EXPIRE:
+		case IMSG_SCHED_ENVELOPE_EXPIRE:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_end(&m);
 
+			m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_ACK, 0, 0, -1);
+			m_add_evpid(p_scheduler, evpid);
+			m_close(p_scheduler);
+
 			/* already removed by scheduler*/
 			if (queue_envelope_load(evpid, &evp) == 0)
 				return;
@@ -226,13 +241,13 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			queue_envelope_delete(evpid);
 			return;
 
-		case IMSG_QUEUE_BOUNCE:
+		case IMSG_SCHED_ENVELOPE_BOUNCE:
 			req_bounce = imsg->data;
 			evpid = req_bounce->evpid;
 
 			if (queue_envelope_load(evpid, &evp) == 0) {
 				log_warnx("queue: bounce: failed to load envelope");
-				m_create(p_scheduler, IMSG_QUEUE_REMOVE, 0, 0, -1);
+				m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_REMOVE, 0, 0, -1);
 				m_add_evpid(p_scheduler, evpid);
 				m_add_u32(p_scheduler, 0); /* not in-flight */
 				m_close(p_scheduler);
@@ -244,47 +259,47 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 				log_warnx("warn: could not update envelope %016"PRIx64, evpid);
 			return;
 
-		case IMSG_MDA_DELIVER:
+		case IMSG_SCHED_ENVELOPE_DELIVER:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_end(&m);
 			if (queue_envelope_load(evpid, &evp) == 0) {
 				log_warnx("queue: deliver: failed to load envelope");
-				m_create(p_scheduler, IMSG_QUEUE_REMOVE, 0, 0, -1);
+				m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_REMOVE, 0, 0, -1);
 				m_add_evpid(p_scheduler, evpid);
 				m_add_u32(p_scheduler, 1); /* in-flight */
 				m_close(p_scheduler);
 				return;
 			}
 			evp.lasttry = time(NULL);
-			m_create(p_mda, IMSG_MDA_DELIVER, 0, 0, -1);
-			m_add_envelope(p_mda, &evp);
-			m_close(p_mda);
+			m_create(p_pony, IMSG_QUEUE_DELIVER, 0, 0, -1);
+			m_add_envelope(p_pony, &evp);
+			m_close(p_pony);
 			return;
 
-		case IMSG_BOUNCE_INJECT:
+		case IMSG_SCHED_ENVELOPE_INJECT:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_end(&m);
 			bounce_add(evpid);
 			return;
 
-		case IMSG_MTA_TRANSFER:
+		case IMSG_SCHED_ENVELOPE_TRANSFER:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_end(&m);
 			if (queue_envelope_load(evpid, &evp) == 0) {
 				log_warnx("queue: failed to load envelope");
-				m_create(p_scheduler, IMSG_QUEUE_REMOVE, 0, 0, -1);
+				m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_REMOVE, 0, 0, -1);
 				m_add_evpid(p_scheduler, evpid);
 				m_add_u32(p_scheduler, 1); /* in-flight */
 				m_close(p_scheduler);
 				return;
 			}
 			evp.lasttry = time(NULL);
-			m_create(p_mta, IMSG_MTA_TRANSFER, 0, 0, -1);
-			m_add_envelope(p_mta, &evp);
-			m_close(p_mta);
+			m_create(p_pony, IMSG_QUEUE_TRANSFER, 0, 0, -1);
+			m_add_envelope(p_pony, &evp);
+			m_close(p_pony);
 			return;
 
 		case IMSG_CTL_LIST_ENVELOPES:
@@ -318,29 +333,34 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 				 */
 				evp.lasttry = nexttry;
 			}
+
+			(void)memcpy(buf, &evp.id, sizeof evp.id);
+			envelope_dump_buffer(&evp, buf+8, sizeof(buf)-8);
 			m_compose(p_control, IMSG_CTL_LIST_ENVELOPES,
-			    imsg->hdr.peerid, 0, -1, &evp, sizeof evp);
+			    imsg->hdr.peerid, 0, -1, buf, 8 + strlen(buf+8) + 1);
 			return;
 		}
 	}
 
-	if (p->proc == PROC_MTA || p->proc == PROC_MDA) {
+	if (p->proc == PROC_PONY) {
 		switch (imsg->hdr.type) {
-		case IMSG_QUEUE_MESSAGE_FD:
+		case IMSG_MDA_OPEN_MESSAGE:
+		case IMSG_MTA_OPEN_MESSAGE:
 			m_msg(&m, imsg);
 			m_get_id(&m, &reqid);
 			m_get_msgid(&m, &msgid);
 			m_end(&m);
 			fd = queue_message_fd_r(msgid);
-			m_create(p, IMSG_QUEUE_MESSAGE_FD, 0, 0, fd);
+			m_create(p, imsg->hdr.type, 0, 0, fd);
 			m_add_id(p, reqid);
 			m_close(p);
 			return;
 
-		case IMSG_DELIVERY_OK:
+		case IMSG_MDA_DELIVERY_OK:
+		case IMSG_MTA_DELIVERY_OK:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
-			if (p->proc == PROC_MTA)
+			if (imsg->hdr.type == IMSG_MTA_DELIVERY_OK)
 				m_get_int(&m, &mta_ext);
 			m_end(&m);
 			if (queue_envelope_load(evpid, &evp) == 0) {
@@ -351,21 +371,22 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 				bounce.type = B_DSN;
 				bounce.dsn_ret = evp.dsn_ret;
 
-				if (p->proc == PROC_MDA)
+				if (imsg->hdr.type == IMSG_MDA_DELIVERY_OK)
 					queue_bounce(&evp, &bounce);
-				else if (p->proc == PROC_MTA &&
+				else if (imsg->hdr.type == IMSG_MTA_DELIVERY_OK &&
 				    (mta_ext & MTA_EXT_DSN) == 0) {
 					bounce.mta_without_dsn = 1;
 					queue_bounce(&evp, &bounce);
 				}
 			}
 			queue_envelope_delete(evpid);
-			m_create(p_scheduler, IMSG_DELIVERY_OK, 0, 0, -1);
+			m_create(p_scheduler, IMSG_QUEUE_DELIVERY_OK, 0, 0, -1);
 			m_add_evpid(p_scheduler, evpid);
 			m_close(p_scheduler);
 			return;
 
-		case IMSG_DELIVERY_TEMPFAIL:
+		case IMSG_MDA_DELIVERY_TEMPFAIL:
+		case IMSG_MTA_DELIVERY_TEMPFAIL:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_get_string(&m, &reason);
@@ -373,7 +394,7 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			m_end(&m);
 			if (queue_envelope_load(evpid, &evp) == 0) {
 				log_warnx("queue: tempfail: failed to load envelope");
-				m_create(p_scheduler, IMSG_QUEUE_REMOVE, 0, 0, -1);
+				m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_REMOVE, 0, 0, -1);
 				m_add_evpid(p_scheduler, evpid);
 				m_add_u32(p_scheduler, 1); /* in-flight */
 				m_close(p_scheduler);
@@ -385,12 +406,13 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			evp.retry++;
 			if (!queue_envelope_update(&evp))
 				log_warnx("warn: could not update envelope %016"PRIx64, evpid);
-			m_create(p_scheduler, IMSG_DELIVERY_TEMPFAIL, 0, 0, -1);
+			m_create(p_scheduler, IMSG_QUEUE_DELIVERY_TEMPFAIL, 0, 0, -1);
 			m_add_envelope(p_scheduler, &evp);
 			m_close(p_scheduler);
 			return;
 
-		case IMSG_DELIVERY_PERMFAIL:
+		case IMSG_MDA_DELIVERY_PERMFAIL:
+		case IMSG_MTA_DELIVERY_PERMFAIL:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_get_string(&m, &reason);
@@ -398,7 +420,7 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			m_end(&m);
 			if (queue_envelope_load(evpid, &evp) == 0) {
 				log_warnx("queue: permfail: failed to load envelope");
-				m_create(p_scheduler, IMSG_QUEUE_REMOVE, 0, 0, -1);
+				m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_REMOVE, 0, 0, -1);
 				m_add_evpid(p_scheduler, evpid);
 				m_add_u32(p_scheduler, 1); /* in-flight */
 				m_close(p_scheduler);
@@ -410,18 +432,19 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			envelope_set_esc_code(&evp, code);
 			queue_bounce(&evp, &bounce);
 			queue_envelope_delete(evpid);
-			m_create(p_scheduler, IMSG_DELIVERY_PERMFAIL, 0, 0, -1);
+			m_create(p_scheduler, IMSG_QUEUE_DELIVERY_PERMFAIL, 0, 0, -1);
 			m_add_evpid(p_scheduler, evpid);
 			m_close(p_scheduler);
 			return;
 
-		case IMSG_DELIVERY_LOOP:
+		case IMSG_MDA_DELIVERY_LOOP:
+		case IMSG_MTA_DELIVERY_LOOP:
 			m_msg(&m, imsg);
 			m_get_evpid(&m, &evpid);
 			m_end(&m);
 			if (queue_envelope_load(evpid, &evp) == 0) {
 				log_warnx("queue: loop: failed to load envelope");
-				m_create(p_scheduler, IMSG_QUEUE_REMOVE, 0, 0, -1);
+				m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_REMOVE, 0, 0, -1);
 				m_add_evpid(p_scheduler, evpid);
 				m_add_u32(p_scheduler, 1); /* in-flight */
 				m_close(p_scheduler);
@@ -433,23 +456,30 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			bounce.type = B_ERROR;
 			queue_bounce(&evp, &bounce);
 			queue_envelope_delete(evp.id);
-			m_create(p_scheduler, IMSG_DELIVERY_LOOP, 0, 0, -1);
+			m_create(p_scheduler, IMSG_QUEUE_DELIVERY_LOOP, 0, 0, -1);
 			m_add_evpid(p_scheduler, evp.id);
 			m_close(p_scheduler);
 			return;
 
-		case IMSG_DELIVERY_HOLD:
-		case IMSG_MTA_SCHEDULE:
+		case IMSG_MTA_DELIVERY_HOLD:
+		case IMSG_MDA_DELIVERY_HOLD:
+			imsg->hdr.type = IMSG_QUEUE_HOLDQ_HOLD;
 			m_forward(p_scheduler, imsg);
 			return;
-		case IMSG_DELIVERY_RELEASE:
+
+		case IMSG_MTA_SCHEDULE:
+			imsg->hdr.type = IMSG_QUEUE_ENVELOPE_SCHEDULE;
+			m_forward(p_scheduler, imsg);
+			return;
+
+		case IMSG_MTA_HOLDQ_RELEASE:
+		case IMSG_MDA_HOLDQ_RELEASE:
 			m_msg(&m, imsg);
 			m_get_id(&m, &holdq);
 			m_get_int(&m, &v);
 			m_end(&m);
-
-			m_create(p_scheduler, IMSG_DELIVERY_RELEASE, 0, 0, -1);
-			if (p->proc == PROC_MTA)
+			m_create(p_scheduler, IMSG_QUEUE_HOLDQ_RELEASE, 0, 0, -1);
+			if (imsg->hdr.type == IMSG_MTA_HOLDQ_RELEASE)
 				m_add_int(p_scheduler, D_MTA);
 			else
 				m_add_int(p_scheduler, D_MDA);
@@ -466,20 +496,14 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 		case IMSG_CTL_PAUSE_MTA:
 		case IMSG_CTL_RESUME_MDA:
 		case IMSG_CTL_RESUME_MTA:
-		case IMSG_QUEUE_REMOVE:
 			m_forward(p_scheduler, imsg);
 			return;
-		}
-	}
 
-	if (p->proc == PROC_PARENT) {
-		switch (imsg->hdr.type) {
 		case IMSG_CTL_VERBOSE:
 			m_msg(&m, imsg);
 			m_get_int(&m, &v);
 			m_end(&m);
 			log_verbose(v);
-			m_forward(p_scheduler, imsg);
 			return;
 
 		case IMSG_CTL_PROFILE:
@@ -488,12 +512,100 @@ queue_imsg(struct mproc *p, struct imsg *imsg)
 			m_end(&m);
 			profiling = v;
 			return;
+
+		case IMSG_CTL_DISCOVER_EVPID:
+			m_msg(&m, imsg);
+			m_get_evpid(&m, &evpid);
+			m_end(&m);
+			if (queue_envelope_load(evpid, &evp) == 0) {
+				log_warnx("queue: discover: failed to load "
+				    "envelope %016" PRIx64, evpid);
+				n_evp = 0;
+				m_compose(p_control, imsg->hdr.type,
+				    imsg->hdr.peerid, 0, -1,
+				    &n_evp, sizeof n_evp);
+				return;
+			}
+
+			m_create(p_scheduler, IMSG_QUEUE_DISCOVER_EVPID,
+			    0, 0, -1);
+			m_add_envelope(p_scheduler, &evp);
+			m_close(p_scheduler);
+
+			m_create(p_scheduler, IMSG_QUEUE_DISCOVER_MSGID,
+			    0, 0, -1);
+			m_add_msgid(p_scheduler, evpid_to_msgid(evpid));
+			m_close(p_scheduler);
+			n_evp = 1;
+			m_compose(p_control, imsg->hdr.type, imsg->hdr.peerid,
+			    0, -1, &n_evp, sizeof n_evp);
+			return;
+
+		case IMSG_CTL_DISCOVER_MSGID:
+			m_msg(&m, imsg);
+			m_get_msgid(&m, &msgid);
+			m_end(&m);
+			/* handle concurrent walk requests */
+			wi = xcalloc(1, sizeof *wi, "queu_imsg");
+			wi->msgid = msgid;
+			wi->peerid = imsg->hdr.peerid;
+			evtimer_set(&wi->ev, queue_msgid_walk, wi);
+			tv.tv_sec = 0;
+			tv.tv_usec = 10;
+			evtimer_add(&wi->ev, &tv);
+			return;
+
+		case IMSG_CTL_UNCORRUPT_MSGID:
+			m_msg(&m, imsg);
+			m_get_msgid(&m, &msgid);
+			m_end(&m);
+			ret = queue_message_uncorrupt(msgid);
+			m_compose(p_control, imsg->hdr.type, imsg->hdr.peerid,
+			    0, -1, &ret, sizeof ret);
+			return;
 		}
 	}
 
 	errx(1, "queue_imsg: unexpected %s imsg", imsg_to_str(imsg->hdr.type));
 }
 
+static void
+queue_msgid_walk(int fd, short event, void *arg)
+{
+	struct envelope		 evp;
+	struct timeval		 tv;
+	struct msg_walkinfo	*wi = arg;
+	int			 r;
+
+	r = queue_message_walk(&evp, wi->msgid, &wi->done, &wi->data);
+	if (r == -1) {
+		if (wi->n_evp) {
+			m_create(p_scheduler, IMSG_QUEUE_DISCOVER_MSGID,
+			    0, 0, -1);
+			m_add_msgid(p_scheduler, wi->msgid);
+			m_close(p_scheduler);
+		}
+
+		m_compose(p_control, IMSG_CTL_DISCOVER_MSGID, wi->peerid, 0, -1,
+		    &wi->n_evp, sizeof wi->n_evp);
+		evtimer_del(&wi->ev);
+		free(wi);
+		return;
+	}
+
+	if (r) {
+		m_create(p_scheduler, IMSG_QUEUE_DISCOVER_EVPID, 0, 0, -1);
+		m_add_envelope(p_scheduler, &evp);
+		m_close(p_scheduler);
+		wi->n_evp += 1;
+	}
+
+	tv.tv_sec = 0;
+	tv.tv_usec = 10;
+	evtimer_set(&wi->ev, queue_msgid_walk, wi);
+	evtimer_add(&wi->ev, &tv);
+}
+
 static void
 queue_bounce(struct envelope *e, struct delivery_bounce *d)
 {
@@ -506,6 +618,7 @@ queue_bounce(struct envelope *e, struct delivery_bounce *d)
 	b.lasttry = 0;
 	b.creation = time(NULL);
 	b.expire = 3600 * 24 * 7;
+	b.dsn_ret = e->dsn_ret;
 
 	if (e->dsn_notify & DSN_NEVER)
 		return;
@@ -525,11 +638,11 @@ queue_bounce(struct envelope *e, struct delivery_bounce *d)
 		log_debug("debug: queue: bouncing evp:%016" PRIx64
 		    " as evp:%016" PRIx64, e->id, b.id);
 
-		m_create(p_scheduler, IMSG_QUEUE_SUBMIT_ENVELOPE, 0, 0, -1);
+		m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_SUBMIT, 0, 0, -1);
 		m_add_envelope(p_scheduler, &b);
 		m_close(p_scheduler);
 
-		m_create(p_scheduler, IMSG_QUEUE_COMMIT_MESSAGE, 0, 0, -1);
+		m_create(p_scheduler, IMSG_QUEUE_MESSAGE_COMMIT, 0, 0, -1);
 		m_add_msgid(p_scheduler, evpid_to_msgid(b.id));
 		m_close(p_scheduler);
 
@@ -554,6 +667,7 @@ static void
 queue_shutdown(void)
 {
 	log_info("info: queue handler exiting");
+	queue_close();
 	_exit(0);
 }
 
@@ -621,11 +735,9 @@ queue(void)
 
 	config_peer(PROC_PARENT);
 	config_peer(PROC_CONTROL);
-	config_peer(PROC_SMTP);
-	config_peer(PROC_MDA);
-	config_peer(PROC_MTA);
 	config_peer(PROC_LKA);
 	config_peer(PROC_SCHEDULER);
+	config_peer(PROC_PONY);
 	config_done();
 
 	/* setup queue loading task */
@@ -653,7 +765,7 @@ queue_timeout(int fd, short event, void *p)
 	r = queue_envelope_walk(&evp);
 	if (r == -1) {
 		if (msgid) {
-			m_create(p_scheduler, IMSG_QUEUE_COMMIT_MESSAGE,
+			m_create(p_scheduler, IMSG_QUEUE_MESSAGE_COMMIT,
 			    0, 0, -1);
 			m_add_msgid(p_scheduler, msgid);
 			m_close(p_scheduler);
@@ -664,13 +776,13 @@ queue_timeout(int fd, short event, void *p)
 
 	if (r) {
 		if (msgid && evpid_to_msgid(evp.id) != msgid) {
-			m_create(p_scheduler, IMSG_QUEUE_COMMIT_MESSAGE,
+			m_create(p_scheduler, IMSG_QUEUE_MESSAGE_COMMIT,
 			    0, 0, -1);
 			m_add_msgid(p_scheduler, msgid);
 			m_close(p_scheduler);
 		}
 		msgid = evpid_to_msgid(evp.id);
-		m_create(p_scheduler, IMSG_QUEUE_SUBMIT_ENVELOPE, 0, 0, -1);
+		m_create(p_scheduler, IMSG_QUEUE_ENVELOPE_SUBMIT, 0, 0, -1);
 		m_add_envelope(p_scheduler, &evp);
 		m_close(p_scheduler);
 	}
@@ -680,51 +792,15 @@ queue_timeout(int fd, short event, void *p)
 	evtimer_add(ev, &tv);
 }
 
-void
-queue_ok(uint64_t evpid)
-{
-	m_create(p_queue, IMSG_DELIVERY_OK, 0, 0, -1);
-	m_add_evpid(p_queue, evpid);
-	m_close(p_queue);
-}
-
-void
-queue_tempfail(uint64_t evpid, const char *reason, enum enhanced_status_code code)
-{
-	m_create(p_queue, IMSG_DELIVERY_TEMPFAIL, 0, 0, -1);
-	m_add_evpid(p_queue, evpid);
-	m_add_string(p_queue, reason);
-	m_add_int(p_queue, (int)code);
-	m_close(p_queue);
-}
-
-void
-queue_permfail(uint64_t evpid, const char *reason, enum enhanced_status_code code)
-{
-	m_create(p_queue, IMSG_DELIVERY_PERMFAIL, 0, 0, -1);
-	m_add_evpid(p_queue, evpid);
-	m_add_string(p_queue, reason);
-	m_add_int(p_queue, (int)code);
-	m_close(p_queue);
-}
-
-void
-queue_loop(uint64_t evpid)
-{
-	m_create(p_queue, IMSG_DELIVERY_LOOP, 0, 0, -1);
-	m_add_evpid(p_queue, evpid);
-	m_close(p_queue);
-}
-
 static void
 queue_log(const struct envelope *e, const char *prefix, const char *status)
 {
-	char rcpt[SMTPD_MAXLINESIZE];
+	char rcpt[LINE_MAX];
 	
-	strlcpy(rcpt, "-", sizeof rcpt);
+	(void)strlcpy(rcpt, "-", sizeof rcpt);
 	if (strcmp(e->rcpt.user, e->dest.user) ||
 	    strcmp(e->rcpt.domain, e->dest.domain))
-		snprintf(rcpt, sizeof rcpt, "%s@%s",
+		(void)snprintf(rcpt, sizeof rcpt, "%s@%s",
 		    e->rcpt.user, e->rcpt.domain);
 	
 	log_info("%s: %s for %016" PRIx64 ": from=<%s@%s>, to=<%s@%s>, "
@@ -745,7 +821,7 @@ queue_flow_control(void)
 	int	oldlimit = limit;
 	int	set, unset;
 
-	bufsz = p_mda->bytes_queued + p_mta->bytes_queued;
+	bufsz = p_pony->bytes_queued;
 	if (bufsz <= flow_agent_lowat)
 		limit &= ~LIMIT_AGENT;
 	else if (bufsz > flow_agent_hiwat)
@@ -762,15 +838,13 @@ queue_flow_control(void)
 	if (set & LIMIT_SCHEDULER) {
 		log_warnx("warn: queue: Hiwat reached on scheduler buffer: "
 		    "suspending transfer, delivery and lookup input");
-		mproc_disable(p_mta);
-		mproc_disable(p_mda);
+		mproc_disable(p_pony);
 		mproc_disable(p_lka);
 	}
 	else if (unset & LIMIT_SCHEDULER) {
 		log_warnx("warn: queue: Down to lowat on scheduler buffer: "
 		    "resuming transfer, delivery and lookup input");
-		mproc_enable(p_mta);
-		mproc_enable(p_mda);
+		mproc_enable(p_pony);
 		mproc_enable(p_lka);
 	}
 
