@@ -31,6 +31,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <limits.h>
 
 #include "smtpd.h"
 #include "log.h"
@@ -44,12 +45,10 @@ scheduler_backend_lookup(const char *name)
 {
 	if (!strcmp(name, "null"))
 		return &scheduler_backend_null;
-	if (!strcmp(name, "proc"))
-		return &scheduler_backend_proc;
 	if (!strcmp(name, "ramqueue"))
 		return &scheduler_backend_ramqueue;
 
-	return NULL;
+	return &scheduler_backend_proc;
 }
 
 void
@@ -64,20 +63,3 @@ scheduler_info(struct scheduler_info *sched, struct envelope *evp)
 	sched->lastbounce = evp->lastbounce;
 	sched->nexttry	= 0;
 }
-
-time_t
-scheduler_compute_schedule(struct scheduler_info *sched)
-{
-	time_t		delay;
-	uint32_t	retry;
-
-	if (sched->type == D_MTA)
-		delay = 800;
-	else
-		delay = 10;
-
-	retry = sched->retry;
-	delay = ((delay * retry) * retry) / 2;
-
-	return (sched->creation + delay);
-}
