@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: smtpctl.c,v 1.124 2014/07/20 01:38:40 guenther Exp $	*/
 
 /*
  * Copyright (c) 2013 Eric Faurot <eric@openbsd.org>
@@ -48,6 +48,7 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <limits.h>
 
 #include "smtpd.h"
 #include "parser.h"
@@ -123,7 +124,7 @@ srv_connect(void)
 
 	memset(&sun, 0, sizeof(sun));
 	sun.sun_family = AF_UNIX;
-	strlcpy(sun.sun_path, SMTPD_SOCKET, sizeof(sun.sun_path));
+	(void)strlcpy(sun.sun_path, SMTPD_SOCKET, sizeof(sun.sun_path));
 	if (connect(ctl_sock, (struct sockaddr *)&sun, sizeof(sun)) == -1) {
 		saved_errno = errno;
 		close(ctl_sock);
@@ -271,6 +272,9 @@ srv_iter_envelopes(uint32_t msgid, struct envelope *evp)
 	static uint32_t	currmsgid = 0;
 	static uint64_t	from = 0;
 	static int	done = 0, need_send = 1, found;
+	char		buf[sizeof(*evp)];
+	size_t		buflen;
+	uint64_t	evpid;
 
 	if (currmsgid != msgid) {
 		if (currmsgid != 0 && !done)
@@ -303,7 +307,12 @@ srv_iter_envelopes(uint32_t msgid, struct envelope *evp)
 		goto again;
 	}
 
-	srv_read(evp, sizeof(*evp));
+	srv_read(&evpid, sizeof evpid);
+	buflen = rlen;
+	srv_read(buf, rlen);	
+	envelope_load_buffer(evp, buf, buflen - 1);
+	evp->id = evpid;
+
 	srv_end();
 	from = evp->id + 1;
 	found++;
@@ -319,9 +328,9 @@ srv_iter_evpids(uint32_t msgid, uint64_t *evpid, int *offset)
 
 	if (evpids == NULL) {
 		alloc = 1000;
-		evpids = malloc(alloc * sizeof(*evpids));
+		evpids = calloc(alloc, sizeof(*evpids));
 		if (evpids == NULL)
-			err(1, "malloc");
+			err(1, "calloc");
 	}
 
 	if (*offset == 0) {
@@ -329,9 +338,10 @@ srv_iter_evpids(uint32_t msgid, uint64_t *evpid, int *offset)
 		while (srv_iter_envelopes(msgid, &evp)) {
 			if (n == alloc) {
 				alloc += 256;
-				evpids = realloc(evpids, alloc * sizeof(*evpids));
+				evpids = reallocarray(evpids, alloc,
+				    sizeof(*evpids));
 				if (evpids == NULL)
-					err(1, "realloc");
+					err(1, "reallocarray");
 			}
 			evpids[n++] = evp.id;
 		}
@@ -426,8 +436,8 @@ do_monitor(int argc, struct parameter *argv)
 	count = 0;
 
 	while (1) {
-		srv_send(IMSG_DIGEST, NULL, 0);
-		srv_recv(IMSG_DIGEST);
+		srv_send(IMSG_CTL_GET_DIGEST, NULL, 0);
+		srv_recv(IMSG_CTL_GET_DIGEST);
 		srv_read(&digest, sizeof(digest));
 		srv_end();
 
@@ -511,7 +521,7 @@ do_profile(int argc, struct parameter *argv)
 
 	v = str_to_profile(argv[0].u.u_str);
 
-	srv_send(IMSG_CTL_PROFILE, &v, sizeof(v));
+	srv_send(IMSG_CTL_PROFILE_ENABLE, &v, sizeof(v));
 	return srv_check_result(1);
 }
 
@@ -586,7 +596,7 @@ do_schedule(int argc, struct parameter *argv)
 static int
 do_show_envelope(int argc, struct parameter *argv)
 {
-	char	 buf[SMTPD_MAXPATHLEN];
+	char	 buf[PATH_MAX];
 
 	if (! bsnprintf(buf, sizeof(buf), "%s%s/%02x/%08x/%016" PRIx64,
 	    PATH_SPOOL,
@@ -612,7 +622,7 @@ do_show_hoststats(int argc, struct parameter *argv)
 static int
 do_show_message(int argc, struct parameter *argv)
 {
-	char	 buf[SMTPD_MAXPATHLEN];
+	char	 buf[PATH_MAX];
 	uint32_t msgid;
 
 	if (argv[0].type == P_EVPID)
@@ -669,11 +679,6 @@ do_show_queue(int argc, struct parameter *argv)
 		}
 
 		fts_close(fts);
-		/*
-		while ((r = queue_envelope_walk(&evp)) != -1)
-			if (r)
-				show_queue_envelope(&evp, 0);
-		*/
 		return (0);
 	}
 
@@ -714,6 +719,31 @@ do_show_routes(int argc, struct parameter *argv)
 	return (0);
 }
 
+#if 1
+static int
+do_show_sizes(int argc, struct parameter *argv)
+{
+	printf("struct userinfo=%ld\n", sizeof (struct userinfo));
+	printf("struct netaddr=%ld\n", sizeof (struct netaddr));
+	printf("struct relayhost=%ld\n", sizeof (struct relayhost));
+	printf("struct credentials=%ld\n", sizeof (struct credentials));
+	printf("struct destination=%ld\n", sizeof (struct destination));
+	printf("struct source=%ld\n", sizeof (struct source));
+	printf("struct addrname=%ld\n", sizeof (struct addrname));
+	printf("union lookup=%ld\n", sizeof (union lookup));
+	printf("struct delivery_mda=%ld\n", sizeof (struct delivery_mda));
+	printf("struct delivery_mta=%ld\n", sizeof (struct delivery_mta));
+	printf("struct envelope=%ld\n", sizeof (struct envelope));
+	printf("struct forward_req=%ld\n", sizeof (struct forward_req));
+	printf("struct deliver=%ld\n", sizeof (struct deliver));
+	printf("struct bounce_req_msg=%ld\n", sizeof (struct bounce_req_msg));
+	printf("struct ca_cert_req_msg=%ld\n", sizeof (struct ca_cert_req_msg));
+	printf("struct ca_vrfy_req_msg=%ld\n", sizeof (struct ca_vrfy_req_msg));
+	return 0;
+}
+#endif
+
+
 static int
 do_show_stats(int argc, struct parameter *argv)
 {
@@ -723,8 +753,8 @@ do_show_stats(int argc, struct parameter *argv)
 	memset(&kv, 0, sizeof kv);
 
 	while (1) {
-		srv_send(IMSG_STATS_GET, &kv, sizeof kv);
-		srv_recv(IMSG_STATS_GET);
+		srv_send(IMSG_CTL_GET_STATS, &kv, sizeof kv);
+		srv_recv(IMSG_CTL_GET_STATS);
 		srv_read(&kv, sizeof(kv));
 		srv_end();
 
@@ -741,7 +771,7 @@ do_show_stats(int argc, struct parameter *argv)
 			switch (kv.val.type) {
 			case STAT_COUNTER:
 				printf("%s=%zd\n",
-				    kv.key, kv.val.u.counter);
+				    kv.key, (ssize_t)kv.val.u.counter);
 				break;
 			case STAT_TIMESTAMP:
 				printf("%s=%" PRId64 "\n",
@@ -798,7 +828,7 @@ do_trace(int argc, struct parameter *argv)
 
 	v = str_to_trace(argv[0].u.u_str);
 
-	srv_send(IMSG_CTL_TRACE, &v, sizeof(v));
+	srv_send(IMSG_CTL_TRACE_ENABLE, &v, sizeof(v));
 	return srv_check_result(1);
 }
 
@@ -809,7 +839,7 @@ do_unprofile(int argc, struct parameter *argv)
 
 	v = str_to_profile(argv[0].u.u_str);
 
-	srv_send(IMSG_CTL_UNPROFILE, &v, sizeof(v));
+	srv_send(IMSG_CTL_PROFILE_DISABLE, &v, sizeof(v));
 	return srv_check_result(1);
 }
 
@@ -820,7 +850,7 @@ do_untrace(int argc, struct parameter *argv)
 
 	v = str_to_trace(argv[0].u.u_str);
 
-	srv_send(IMSG_CTL_UNTRACE, &v, sizeof(v));
+	srv_send(IMSG_CTL_TRACE_DISABLE, &v, sizeof(v));
 	return srv_check_result(1);
 }
 
@@ -829,7 +859,7 @@ do_update_table(int argc, struct parameter *argv)
 {
 	const char	*name = argv[0].u.u_str;
 
-	srv_send(IMSG_LKA_UPDATE_TABLE, name, strlen(name) + 1);
+	srv_send(IMSG_CTL_UPDATE_TABLE, name, strlen(name) + 1);
 	return srv_check_result(1);
 }
 
@@ -889,6 +919,63 @@ do_show_mta_block(int argc, struct parameter *argv)
 	return (0);
 }
 
+static int
+do_discover(int argc, struct parameter *argv)
+{
+	uint64_t evpid;
+	uint32_t msgid;
+	size_t	 n_evp;
+
+	if (ibuf == NULL && !srv_connect())
+		errx(1, "smtpd doesn't seem to be running");
+
+	if (argv[0].type == P_EVPID) {
+		evpid = argv[0].u.u_evpid;
+		srv_send(IMSG_CTL_DISCOVER_EVPID, &evpid, sizeof evpid);
+		srv_recv(IMSG_CTL_DISCOVER_EVPID);
+	} else {
+		msgid = argv[0].u.u_msgid;
+		srv_send(IMSG_CTL_DISCOVER_MSGID, &msgid, sizeof msgid);
+		srv_recv(IMSG_CTL_DISCOVER_MSGID);
+	}
+
+	if (rlen == 0) {
+		srv_end();
+		return (0);
+	} else {
+		srv_read(&n_evp, sizeof n_evp);
+		srv_end();
+	}
+	
+	printf("%zu envelope%s discovered\n", n_evp, (n_evp != 1) ? "s" : "");
+	return (0);
+}
+
+static int
+do_uncorrupt(int argc, struct parameter *argv)
+{
+	uint32_t msgid;
+	int	 ret;
+
+	if (ibuf == NULL && !srv_connect())
+		errx(1, "smtpd doesn't seem to be running");
+
+	msgid = argv[0].u.u_msgid;
+	srv_send(IMSG_CTL_UNCORRUPT_MSGID, &msgid, sizeof msgid);
+	srv_recv(IMSG_CTL_UNCORRUPT_MSGID);
+
+	if (rlen == 0) {
+		srv_end();
+		return (0);
+	} else {
+		srv_read(&ret, sizeof ret);
+		srv_end();
+	}
+
+	printf("command %s\n", ret ? "succeeded" : "failed");
+	return (0);
+}
+
 int
 main(int argc, char **argv)
 {
@@ -903,6 +990,8 @@ main(int argc, char **argv)
 	if (geteuid())
 		errx(1, "need root privileges");
 
+	cmd_install("discover <evpid>",		do_discover);
+	cmd_install("discover <msgid>",		do_discover);
 	cmd_install("encrypt",			do_encrypt);
 	cmd_install("encrypt <str>",		do_encrypt);
 	cmd_install("pause mta from <addr> for <str>", do_block_mta);
@@ -941,10 +1030,17 @@ main(int argc, char **argv)
 	cmd_install("show status",		do_show_status);
 	cmd_install("stop",			do_stop);
 	cmd_install("trace <str>",		do_trace);
+	cmd_install("uncorrupt <msgid>",	do_uncorrupt);
 	cmd_install("unprofile <str>",		do_unprofile);
 	cmd_install("untrace <str>",		do_untrace);
 	cmd_install("update table <str>",	do_update_table);
 
+#if 1
+	/* print size of various structures */
+	cmd_install("show sizes",		do_show_sizes);
+#endif
+
+
 	if (strcmp(__progname, "mailq") == 0)
 		return cmd_run(2, argv_mailq);
 	if (strcmp(__progname, "smtpctl") == 0)
@@ -971,17 +1067,17 @@ show_queue_envelope(struct envelope *e, int online)
 
 	if (online) {
 		if (e->flags & EF_PENDING)
-			snprintf(runstate, sizeof runstate, "pending|%zi",
+			(void)snprintf(runstate, sizeof runstate, "pending|%zd",
 			    (ssize_t)(e->nexttry - now));
 		else if (e->flags & EF_INFLIGHT)
-			snprintf(runstate, sizeof runstate, "inflight|%zi",
+			(void)snprintf(runstate, sizeof runstate, "inflight|%zd",
 			    (ssize_t)(now - e->lasttry));
 		else
-			snprintf(runstate, sizeof runstate, "invalid|");
+			(void)snprintf(runstate, sizeof runstate, "invalid|");
 		e->flags &= ~(EF_PENDING|EF_INFLIGHT);
 	}
 	else
-		strlcpy(runstate, "offline|", sizeof runstate);
+		(void)strlcpy(runstate, "offline|", sizeof runstate);
 
 	if (e->flags)
 		errx(1, "%016" PRIx64 ": unexpected flags 0x%04x", e->id,
@@ -1030,8 +1126,8 @@ getflag(uint *bitmap, int bit, char *bitstr, char *buf, size_t len)
 {
 	if (*bitmap & bit) {
 		*bitmap &= ~bit;
-		strlcat(buf, bitstr, len);
-		strlcat(buf, ",", len);
+		(void)strlcat(buf, bitstr, len);
+		(void)strlcat(buf, ",", len);
 	}
 }
 
@@ -1039,7 +1135,7 @@ static void
 show_offline_envelope(uint64_t evpid)
 {
 	FILE   *fp = NULL;
-	char	pathname[SMTPD_MAXPATHLEN];
+	char	pathname[PATH_MAX];
 	size_t	plen;
 	char   *p;
 	size_t	buflen;
@@ -1103,11 +1199,12 @@ display(const char *s)
 
 		if ((fd = mkstemp(sfn)) == -1 ||
 		    (ofp = fdopen(fd, "w+")) == NULL) {
+			int saved_errno = errno;
 			if (fd != -1) {
 				unlink(sfn);
 				close(fd);
 			}
-			err(1, "mkstemp");
+			errc(1, saved_errno, "mkstemp");
 		}
 		unlink(sfn);
 
@@ -1152,8 +1249,8 @@ str_to_trace(const char *str)
 		return TRACE_IO;
 	if (!strcmp(str, "smtp"))
 		return TRACE_SMTP;
-	if (!strcmp(str, "mfa"))
-		return TRACE_MFA;
+	if (!strcmp(str, "filters"))
+		return TRACE_FILTERS;
 	if (!strcmp(str, "mta"))
 		return TRACE_MTA;
 	if (!strcmp(str, "bounce"))
