@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: smtpd-defines.h,v 1.3 2014/04/30 09:17:29 gilles Exp $	*/
 
 /*
  * Copyright (c) 2013 Gilles Chehade <gilles@poolp.org>
@@ -20,13 +20,16 @@
 #define nitems(_a) (sizeof((_a)) / sizeof((_a)[0]))
 #endif
 
-#define SMTPD_MAXLOCALPARTSIZE	 (255 + 1)
-#define SMTPD_MAXDOMAINPARTSIZE	 (255 + 1)
+#define	SMTPD_TABLENAME_SIZE	(64 + 1)
+#define	SMTPD_TAG_SIZE		(32 + 1)
 
-#define	SMTPD_MAXLOGNAME	32
-#define	SMTPD_MAXPATHLEN	1024
-#define	SMTPD_MAXHOSTNAMELEN	256
-#define	SMTPD_MAXLINESIZE	2048
+/* buffer sizes for email address components */
+#define SMTPD_MAXLOCALPARTSIZE	(255 + 1)
+#define SMTPD_MAXDOMAINPARTSIZE	(255 + 1)
+#define	SMTPD_MAXMAILADDRSIZE	(255 + 1)
+
+/* buffer size for virtual username (can be email addresses) */
+#define	SMTPD_VUSERNAME_SIZE	SMTPD_MAXMAILADDRSIZE
 
 #ifndef SMTPD_USER
 #define SMTPD_USER		"_smtpd"
@@ -40,3 +43,5 @@
 #ifndef PATH_SPOOL
 #define PATH_SPOOL		"/var/spool/smtpd"
 #endif
+
+#define TAG_CHAR		'+'
