@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: smtpd.h,v 1.455 2014/04/19 16:55:15 gilles Exp $	*/
 
 /*
  * Copyright (c) 2008 Gilles Chehade <gilles@poolp.org>
@@ -18,11 +18,7 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#if LIBEVENT_MAJOR_VERSION < 2
 #include <event.h>
-#else
-#include <event2/event_struct.h>
-#endif
 
 #include			 <imsg.h>
 
@@ -37,19 +33,23 @@
 #include "ioev.h"
 #include "iobuf.h"
 
+#include "rfc2822.h"
 
 #ifndef SMTPD_CONFDIR
 #define SMTPD_CONFDIR		 "/etc"
 #endif
 #define CONF_FILE		 SMTPD_CONFDIR "/smtpd.conf"
 #define MAILNAME_FILE		 SMTPD_CONFDIR "/mailname"
+#ifndef CA_FILE
+#define CA_FILE                  "/etc/ssl/cert.pem"
+#endif
 
-#define PROC_COUNT		 10
+#define PROC_COUNT		 7
 
 #define MAX_HOPS_COUNT		 100
 #define	DEFAULT_MAX_BODY_SIZE	(35*1024*1024)
-#define MAX_TAG_SIZE		 32
 #define	MAX_FILTER_NAME		 32
+#define	MAX_FILTER_ARGS		 255
 
 #define	EXPAND_BUFFER		 1024
 
@@ -67,8 +67,7 @@
 #ifndef SMTPD_NAME
 #define	SMTPD_NAME		 "OpenSMTPD"
 #endif
-#define	SMTPD_VERSION		 "5.4.2p1"
-#define SMTPD_BANNER		 "220 %s ESMTP %s"
+#define	SMTPD_VERSION		 "5.7.3p2"
 #define SMTPD_SESSION_TIMEOUT	 300
 #define SMTPD_BACKLOG		 5
 
@@ -80,11 +79,8 @@
 #define PATH_PURGE		"/purge"
 #define PATH_TEMPORARY		"/temporary"
 
-#ifndef	PATH_FILTERS
-#define	PATH_FILTERS		"/usr/libexec/smtpd"
-#endif
-#ifndef	PATH_TABLES
-#define	PATH_TABLES		"/usr/libexec/smtpd"
+#ifndef	PATH_LIBEXEC
+#define	PATH_LIBEXEC		"/usr/libexec/smtpd"
 #endif
 
 
@@ -107,6 +103,11 @@
 #define	F_LMTP			0x80
 #define	F_MASK_SOURCE		0x100
 #define	F_TLS_VERIFY		0x200
+#define	F_EXT_DSN		0x400
+#define	F_RECEIVEDAUTH		0x800
+#define	F_MASQUERADE		0x1000
+#define	F_DSNNOTIFY_DISABLE    	0x2000
+#define	F_DSNRET_HEADERS	0x4000
 
 /* must match F_* for mta */
 #define RELAY_STARTTLS		0x01
@@ -118,12 +119,12 @@
 #define RELAY_MX		0x20
 #define RELAY_LMTP		0x80
 #define	RELAY_TLS_VERIFY	0x200
-
 #define MTA_EXT_DSN		0x400
 
+
 struct userinfo {
-	char username[SMTPD_MAXLOGNAME];
-	char directory[SMTPD_MAXPATHLEN];
+	char username[SMTPD_VUSERNAME_SIZE];
+	char directory[PATH_MAX];
 	uid_t uid;
 	gid_t gid;
 };
@@ -135,23 +136,24 @@ struct netaddr {
 
 struct relayhost {
 	uint16_t flags;
-	char hostname[SMTPD_MAXHOSTNAMELEN];
+	char hostname[HOST_NAME_MAX+1];
 	uint16_t port;
-	char pki_name[SMTPD_MAXPATHLEN];
-	char authtable[SMTPD_MAXPATHLEN];
-	char authlabel[SMTPD_MAXPATHLEN];
-	char sourcetable[SMTPD_MAXPATHLEN];
-	char heloname[SMTPD_MAXHOSTNAMELEN];
-	char helotable[SMTPD_MAXPATHLEN];
+	char pki_name[HOST_NAME_MAX+1];
+	char ca_name[HOST_NAME_MAX+1];
+	char authtable[SMTPD_TABLENAME_SIZE];
+	char authlabel[PATH_MAX];
+	char sourcetable[PATH_MAX];
+	char heloname[HOST_NAME_MAX+1];
+	char helotable[PATH_MAX];
 };
 
 struct credentials {
-	char username[SMTPD_MAXLINESIZE];
-	char password[SMTPD_MAXLINESIZE];
+	char username[LINE_MAX];
+	char password[LINE_MAX];
 };
 
 struct destination {
-	char	name[SMTPD_MAXHOSTNAMELEN];
+	char	name[HOST_NAME_MAX+1];
 };
 
 struct source {
@@ -160,7 +162,7 @@ struct source {
 
 struct addrname {
 	struct sockaddr_storage	addr;
-	char			name[SMTPD_MAXHOSTNAMELEN];
+	char			name[HOST_NAME_MAX+1];
 };
 
 union lookup {
@@ -172,40 +174,25 @@ union lookup {
 	struct userinfo		 userinfo;
 	struct mailaddr		 mailaddr;
 	struct addrname		 addrname;
+	struct maddrmap		*maddrmap;
 };
 
 /*
  * Bump IMSG_VERSION whenever a change is made to enum imsg_type.
  * This will ensure that we can never use a wrong version of smtpctl with smtpd.
  */
-#define	IMSG_VERSION		9
+#define	IMSG_VERSION		14
 
 enum imsg_type {
 	IMSG_NONE,
-	IMSG_CTL_OK,		/* answer to smtpctl requests */
+
+	IMSG_CTL_OK,
 	IMSG_CTL_FAIL,
-	IMSG_CTL_SHUTDOWN,
-	IMSG_CTL_VERBOSE,
-	IMSG_CTL_PAUSE_EVP,
-	IMSG_CTL_PAUSE_MDA,
-	IMSG_CTL_PAUSE_MTA,
-	IMSG_CTL_PAUSE_SMTP,
-	IMSG_CTL_RESUME_EVP,
-	IMSG_CTL_RESUME_MDA,
-	IMSG_CTL_RESUME_MTA,
-	IMSG_CTL_RESUME_SMTP,
-	IMSG_CTL_RESUME_ROUTE,
+
+	IMSG_CTL_GET_DIGEST,
+	IMSG_CTL_GET_STATS,
 	IMSG_CTL_LIST_MESSAGES,
 	IMSG_CTL_LIST_ENVELOPES,
-	IMSG_CTL_REMOVE,
-	IMSG_CTL_SCHEDULE,
-	IMSG_CTL_SHOW_STATUS,
-
-	IMSG_CTL_TRACE,
-	IMSG_CTL_UNTRACE,
-	IMSG_CTL_PROFILE,
-	IMSG_CTL_UNPROFILE,
-
 	IMSG_CTL_MTA_SHOW_HOSTS,
 	IMSG_CTL_MTA_SHOW_RELAYS,
 	IMSG_CTL_MTA_SHOW_ROUTES,
@@ -213,91 +200,132 @@ enum imsg_type {
 	IMSG_CTL_MTA_BLOCK,
 	IMSG_CTL_MTA_UNBLOCK,
 	IMSG_CTL_MTA_SHOW_BLOCK,
+	IMSG_CTL_PAUSE_EVP,
+	IMSG_CTL_PAUSE_MDA,
+	IMSG_CTL_PAUSE_MTA,
+	IMSG_CTL_PAUSE_SMTP,
+	IMSG_CTL_PROFILE,
+	IMSG_CTL_PROFILE_DISABLE,
+	IMSG_CTL_PROFILE_ENABLE,
+	IMSG_CTL_RESUME_EVP,
+	IMSG_CTL_RESUME_MDA,
+	IMSG_CTL_RESUME_MTA,
+	IMSG_CTL_RESUME_SMTP,
+	IMSG_CTL_RESUME_ROUTE,
+	IMSG_CTL_REMOVE,
+	IMSG_CTL_SCHEDULE,
+	IMSG_CTL_SHOW_STATUS,
+	IMSG_CTL_SHUTDOWN,
+	IMSG_CTL_TRACE_DISABLE,
+	IMSG_CTL_TRACE_ENABLE,
+	IMSG_CTL_UPDATE_TABLE,
+	IMSG_CTL_VERBOSE,
+	IMSG_CTL_DISCOVER_EVPID,
+	IMSG_CTL_DISCOVER_MSGID,
+	IMSG_CTL_UNCORRUPT_MSGID,
+
+	IMSG_CTL_SMTP_SESSION,
 
 	IMSG_CONF_START,
-	IMSG_CONF_SSL,
-	IMSG_CONF_LISTENER,
-	IMSG_CONF_TABLE,
-	IMSG_CONF_TABLE_CONTENT,
-	IMSG_CONF_RULE,
-	IMSG_CONF_RULE_SOURCE,
-	IMSG_CONF_RULE_SENDER,
-	IMSG_CONF_RULE_DESTINATION,
-	IMSG_CONF_RULE_RECIPIENT,
-	IMSG_CONF_RULE_MAPPING,
-	IMSG_CONF_RULE_USERS,
-	IMSG_CONF_FILTER,
 	IMSG_CONF_END,
 
-	IMSG_LKA_UPDATE_TABLE,
-	IMSG_LKA_EXPAND_RCPT,
-	IMSG_LKA_SECRET,
-	IMSG_LKA_SOURCE,
-	IMSG_LKA_HELO,
-	IMSG_LKA_USERINFO,
-	IMSG_LKA_AUTHENTICATE,
-	IMSG_LKA_SSL_INIT,
-	IMSG_LKA_SSL_VERIFY_CERT,
-	IMSG_LKA_SSL_VERIFY_CHAIN,
-	IMSG_LKA_SSL_VERIFY,
-
-	IMSG_DELIVERY_OK,
-	IMSG_DELIVERY_TEMPFAIL,
-	IMSG_DELIVERY_PERMFAIL,
-	IMSG_DELIVERY_LOOP,
-	IMSG_DELIVERY_HOLD,
-	IMSG_DELIVERY_RELEASE,
-
-	IMSG_BOUNCE_INJECT,
-
-	IMSG_MDA_DELIVER,
-	IMSG_MDA_DONE,
-
-	IMSG_MFA_REQ_CONNECT,
-	IMSG_MFA_REQ_HELO,
-	IMSG_MFA_REQ_MAIL,
-	IMSG_MFA_REQ_RCPT,
-	IMSG_MFA_REQ_DATA,
-	IMSG_MFA_REQ_EOM,
-	IMSG_MFA_EVENT_RSET,
-	IMSG_MFA_EVENT_COMMIT,
-	IMSG_MFA_EVENT_ROLLBACK,
-	IMSG_MFA_EVENT_DISCONNECT,
-	IMSG_MFA_SMTP_RESPONSE,
-
-	IMSG_MTA_TRANSFER,
-	IMSG_MTA_SCHEDULE,
-
-	IMSG_QUEUE_CREATE_MESSAGE,
-	IMSG_QUEUE_SUBMIT_ENVELOPE,
-	IMSG_QUEUE_COMMIT_ENVELOPES,
-	IMSG_QUEUE_REMOVE_MESSAGE,
-	IMSG_QUEUE_COMMIT_MESSAGE,
-	IMSG_QUEUE_MESSAGE_FD,
-	IMSG_QUEUE_MESSAGE_FILE,
-	IMSG_QUEUE_REMOVE,
-	IMSG_QUEUE_EXPIRE,
-	IMSG_QUEUE_BOUNCE,
-
-	IMSG_PARENT_FORWARD_OPEN,
-	IMSG_PARENT_FORK_MDA,
-	IMSG_PARENT_KILL_MDA,
-
-	IMSG_SMTP_ENQUEUE_FD,
-
-	IMSG_DNS_HOST,
-	IMSG_DNS_HOST_END,
-	IMSG_DNS_PTR,
-	IMSG_DNS_MX,
-	IMSG_DNS_MX_PREFERENCE,
-
 	IMSG_STAT_INCREMENT,
 	IMSG_STAT_DECREMENT,
 	IMSG_STAT_SET,
 
-	IMSG_DIGEST,
-	IMSG_STATS,
-	IMSG_STATS_GET,
+	IMSG_LKA_AUTHENTICATE,
+	IMSG_LKA_OPEN_FORWARD,
+	IMSG_LKA_ENVELOPE_SUBMIT,
+	IMSG_LKA_ENVELOPE_COMMIT,
+
+	IMSG_QUEUE_DELIVER,
+	IMSG_QUEUE_DELIVERY_OK,
+	IMSG_QUEUE_DELIVERY_TEMPFAIL,
+	IMSG_QUEUE_DELIVERY_PERMFAIL,
+	IMSG_QUEUE_DELIVERY_LOOP,
+	IMSG_QUEUE_DISCOVER_EVPID,
+	IMSG_QUEUE_DISCOVER_MSGID,
+	IMSG_QUEUE_ENVELOPE_ACK,
+	IMSG_QUEUE_ENVELOPE_COMMIT,
+	IMSG_QUEUE_ENVELOPE_REMOVE,
+	IMSG_QUEUE_ENVELOPE_SCHEDULE,
+	IMSG_QUEUE_ENVELOPE_SUBMIT,
+	IMSG_QUEUE_HOLDQ_HOLD,
+	IMSG_QUEUE_HOLDQ_RELEASE,
+	IMSG_QUEUE_MESSAGE_COMMIT,
+	IMSG_QUEUE_MESSAGE_ROLLBACK,
+	IMSG_QUEUE_SMTP_SESSION,
+	IMSG_QUEUE_TRANSFER,
+
+	IMSG_MDA_DELIVERY_OK,
+	IMSG_MDA_DELIVERY_TEMPFAIL,
+	IMSG_MDA_DELIVERY_PERMFAIL,
+	IMSG_MDA_DELIVERY_LOOP,
+	IMSG_MDA_DELIVERY_HOLD,
+	IMSG_MDA_DONE,
+	IMSG_MDA_FORK,
+	IMSG_MDA_HOLDQ_RELEASE,
+	IMSG_MDA_LOOKUP_USERINFO,
+	IMSG_MDA_KILL,
+	IMSG_MDA_OPEN_MESSAGE,
+
+	IMSG_MFA_SMTP_RESPONSE,
+
+	IMSG_MTA_DELIVERY_OK,
+	IMSG_MTA_DELIVERY_TEMPFAIL,
+	IMSG_MTA_DELIVERY_PERMFAIL,
+	IMSG_MTA_DELIVERY_LOOP,
+	IMSG_MTA_DELIVERY_HOLD,
+	IMSG_MTA_DNS_HOST,
+	IMSG_MTA_DNS_HOST_END,
+	IMSG_MTA_DNS_PTR,
+	IMSG_MTA_DNS_MX,
+	IMSG_MTA_DNS_MX_PREFERENCE,
+	IMSG_MTA_HOLDQ_RELEASE,
+	IMSG_MTA_LOOKUP_CREDENTIALS,
+	IMSG_MTA_LOOKUP_SOURCE,
+	IMSG_MTA_LOOKUP_HELO,
+	IMSG_MTA_OPEN_MESSAGE,
+	IMSG_MTA_SCHEDULE,
+	IMSG_MTA_TLS_INIT,
+	IMSG_MTA_TLS_VERIFY_CERT,
+	IMSG_MTA_TLS_VERIFY_CHAIN,
+	IMSG_MTA_TLS_VERIFY,
+
+	IMSG_SCHED_ENVELOPE_BOUNCE,
+	IMSG_SCHED_ENVELOPE_DELIVER,
+	IMSG_SCHED_ENVELOPE_EXPIRE,
+	IMSG_SCHED_ENVELOPE_INJECT,
+	IMSG_SCHED_ENVELOPE_REMOVE,
+	IMSG_SCHED_ENVELOPE_TRANSFER,
+
+	IMSG_SMTP_AUTHENTICATE,
+	IMSG_SMTP_DNS_PTR,
+	IMSG_SMTP_MESSAGE_COMMIT,
+	IMSG_SMTP_MESSAGE_CREATE,
+	IMSG_SMTP_MESSAGE_ROLLBACK,
+	IMSG_SMTP_MESSAGE_OPEN,
+	IMSG_SMTP_CHECK_SENDER,
+	IMSG_SMTP_EXPAND_RCPT,
+	IMSG_SMTP_LOOKUP_HELO,
+	IMSG_SMTP_TLS_INIT,
+	IMSG_SMTP_TLS_VERIFY_CERT,
+	IMSG_SMTP_TLS_VERIFY_CHAIN,
+	IMSG_SMTP_TLS_VERIFY,
+
+	IMSG_SMTP_REQ_CONNECT,
+	IMSG_SMTP_REQ_HELO,
+	IMSG_SMTP_REQ_MAIL,
+	IMSG_SMTP_REQ_RCPT,
+	IMSG_SMTP_REQ_DATA,
+	IMSG_SMTP_REQ_EOM,
+	IMSG_SMTP_EVENT_RSET,
+	IMSG_SMTP_EVENT_COMMIT,
+	IMSG_SMTP_EVENT_ROLLBACK,
+	IMSG_SMTP_EVENT_DISCONNECT,
+
+	IMSG_CA_PRIVENC,
+	IMSG_CA_PRIVDEC
 };
 
 enum blockmodes {
@@ -307,14 +335,12 @@ enum blockmodes {
 
 enum smtp_proc_type {
 	PROC_PARENT = 0,
-	PROC_SMTP,
-	PROC_MFA,
 	PROC_LKA,
 	PROC_QUEUE,
-	PROC_MDA,
-	PROC_MTA,
 	PROC_CONTROL,
 	PROC_SCHEDULER,
+	PROC_PONY,
+	PROC_CA,
 
 	PROC_FILTER,
 	PROC_CLIENT,
@@ -328,9 +354,9 @@ enum table_type {
 };
 
 struct table {
-	char				 t_name[SMTPD_MAXLINESIZE];
+	char				 t_name[LINE_MAX];
 	enum table_type			 t_type;
-	char				 t_config[SMTPD_MAXPATHLEN];
+	char				 t_config[PATH_MAX];
 
 	struct dict			 t_dict;
 
@@ -345,8 +371,8 @@ struct table_backend {
 	void   *(*open)(struct table *);
 	int	(*update)(struct table *);
 	void	(*close)(void *);
-	int	(*lookup)(void *, const char *, enum table_service, union lookup *);
-	int	(*fetch)(void *, enum table_service, union lookup *);
+	int	(*lookup)(void *, struct dict *, const char *, enum table_service, union lookup *);
+	int	(*fetch)(void *, struct dict *, enum table_service, union lookup *);
 };
 
 
@@ -375,7 +401,7 @@ struct rule {
 	TAILQ_ENTRY(rule)		r_entry;
 	enum decision			r_decision;
 	uint8_t				r_nottag;
-	char				r_tag[MAX_TAG_SIZE];
+	char				r_tag[SMTPD_TAG_SIZE];
 
 	uint8_t				r_notsources;
 	struct table		       *r_sources;
@@ -401,13 +427,15 @@ struct rule {
 	struct table		       *r_userbase;
 	time_t				r_qexpire;
 	uint8_t				r_forwardonly;
+	char				r_delivery_user[LINE_MAX];
 };
 
 struct delivery_mda {
 	enum action_type	method;
-	char			usertable[SMTPD_MAXPATHLEN];
-	char			username[SMTPD_MAXLOGNAME];
+	char			usertable[SMTPD_TABLENAME_SIZE];
+	char			username[SMTPD_VUSERNAME_SIZE];
 	char			buffer[EXPAND_BUFFER];
+	char			delivery_user[SMTPD_VUSERNAME_SIZE];
 };
 
 struct delivery_mta {
@@ -440,7 +468,8 @@ enum expand_type {
 	EXPAND_FILTER,
 	EXPAND_INCLUDE,
 	EXPAND_ADDRESS,
-	EXPAND_ERROR
+	EXPAND_ERROR,
+	EXPAND_MAILDIR
 };
 
 struct expandnode {
@@ -474,6 +503,15 @@ struct expand {
 	struct expandnode		*parent;
 };
 
+struct maddrnode {
+	TAILQ_ENTRY(maddrnode)		entries;
+	struct mailaddr			mailaddr;
+};
+
+struct maddrmap {
+	TAILQ_HEAD(xmaddr, maddrnode)	queue;
+};
+
 #define DSN_SUCCESS 0x01
 #define DSN_FAILURE 0x02
 #define DSN_DELAY   0x04
@@ -485,16 +523,16 @@ struct expand {
 struct envelope {
 	TAILQ_ENTRY(envelope)		entry;
 
-	char				tag[MAX_TAG_SIZE];
+	char				tag[SMTPD_TAG_SIZE];
 
 	uint32_t			version;
 	uint64_t			id;
 	enum envelope_flags		flags;
 
-	char				smtpname[SMTPD_MAXHOSTNAMELEN];
-	char				helo[SMTPD_MAXHOSTNAMELEN];
-	char				hostname[SMTPD_MAXHOSTNAMELEN];
-	char				errorline[SMTPD_MAXLINESIZE];
+	char				smtpname[HOST_NAME_MAX+1];
+	char				helo[HOST_NAME_MAX+1];
+	char				hostname[HOST_NAME_MAX+1];
+	char				errorline[LINE_MAX];
 	struct sockaddr_storage		ss;
 
 	struct mailaddr			sender;
@@ -531,16 +569,20 @@ struct listener {
 	in_port_t		 port;
 	struct timeval		 timeout;
 	struct event		 ev;
-	char			 pki_name[SMTPD_MAXPATHLEN];
-	char			 tag[MAX_TAG_SIZE];
-	char			 authtable[SMTPD_MAXLINESIZE];
-	char			 hostname[SMTPD_MAXHOSTNAMELEN];
-	char			 hostnametable[SMTPD_MAXPATHLEN];
+	char			 pki_name[PATH_MAX];
+	char			 ca_name[PATH_MAX];
+	char			 tag[SMTPD_TAG_SIZE];
+	char			 filter[PATH_MAX];
+	char			 authtable[LINE_MAX];
+	char			 hostname[HOST_NAME_MAX+1];
+	char			 hostnametable[PATH_MAX];
+	char			 sendertable[PATH_MAX];
+
 	TAILQ_ENTRY(listener)	 entry;
 };
 
 struct smtpd {
-	char				sc_conffile[SMTPD_MAXPATHLEN];
+	char				sc_conffile[PATH_MAX];
 	size_t				sc_maxsize;
 
 #define SMTPD_OPT_VERBOSE		0x00000001
@@ -564,6 +606,9 @@ struct smtpd {
 	char			       *sc_queue_key;
 	size_t				sc_queue_evpcache_size;
 
+	size_t				sc_session_max_rcpt;
+	size_t				sc_session_max_mails;
+
 	size_t				sc_mda_max_session;
 	size_t				sc_mda_max_user_session;
 	size_t				sc_mda_task_hiwat;
@@ -580,7 +625,7 @@ struct smtpd {
 	int				sc_qexpire;
 #define MAX_BOUNCE_WARN			4
 	time_t				sc_bounce_warn[MAX_BOUNCE_WARN];
-	char				sc_hostname[SMTPD_MAXHOSTNAMELEN];
+	char				sc_hostname[HOST_NAME_MAX+1];
 	struct stat_backend	       *sc_stat;
 	struct compress_backend	       *sc_comp;
 
@@ -589,7 +634,8 @@ struct smtpd {
 	TAILQ_HEAD(listenerlist, listener)	*sc_listeners;
 
 	TAILQ_HEAD(rulelist, rule)		*sc_rules;
-	
+
+       	struct dict			       *sc_ca_dict;
 	struct dict			       *sc_pki_dict;
 	struct dict			       *sc_ssl_dict;
 
@@ -599,13 +645,18 @@ struct smtpd {
 
 	struct dict				sc_filters;
 	uint32_t				filtermask;
+
+	char					sc_enqueue_filter[PATH_MAX];
+
+	char			       *sc_tls_ciphers;
+	char			       *sc_tls_curve;
 };
 
 #define	TRACE_DEBUG	0x0001
 #define	TRACE_IMSG	0x0002
 #define	TRACE_IO	0x0004
 #define	TRACE_SMTP	0x0008
-#define	TRACE_MFA	0x0010
+#define	TRACE_FILTERS	0x0010
 #define	TRACE_MTA	0x0020
 #define	TRACE_BOUNCE	0x0040
 #define	TRACE_SCHEDULER	0x0080
@@ -620,33 +671,36 @@ struct smtpd {
 #define PROFILE_TOSTAT	0x0001
 #define PROFILE_IMSG	0x0002
 #define PROFILE_QUEUE	0x0004
+#define PROFILE_BUFFERS	0x0008
 
 struct forward_req {
 	uint64_t			id;
 	uint8_t				status;
 
-	char				user[SMTPD_MAXLOGNAME];
+	char				user[LINE_MAX];
 	uid_t				uid;
 	gid_t				gid;
-	char				directory[SMTPD_MAXPATHLEN];
+	char				directory[PATH_MAX];
 };
 
 struct deliver {
-	char			to[SMTPD_MAXPATHLEN];
-	char			from[SMTPD_MAXPATHLEN];
-	char			user[SMTPD_MAXLOGNAME];
+	char			to[EXPAND_BUFFER];
+	char			from[SMTPD_MAXMAILADDRSIZE];
+	char			dest[SMTPD_MAXMAILADDRSIZE];
+	char			user[SMTPD_VUSERNAME_SIZE];
 	short			mode;
 
 	struct userinfo		userinfo;
 };
 
 #define MAX_FILTER_PER_CHAIN	16
-struct filter {
-	int			chain;
-	int			done;
-	char			name[MAX_FILTER_NAME];
-	char			path[SMTPD_MAXPATHLEN];
-	char			filters[MAX_FILTER_NAME][MAX_FILTER_PER_CHAIN];
+struct filter_conf {
+	int		 chain;
+	int		 done;
+	int		 argc;
+	char		*name;
+	char		*argv[MAX_FILTER_ARGS + 1];
+	char		*path;
 };
 
 struct mta_host {
@@ -781,6 +835,7 @@ struct mta_relay {
 	char			*sourcetable;
 	uint16_t		 port;
 	char			*pki_name;
+	char			*ca_name;
 	char			*authtable;
 	char			*authlabel;
 	char			*helotable;
@@ -830,7 +885,7 @@ struct mta_envelope {
 	uint8_t				dsn_notify;
 	enum dsn_ret			dsn_ret;
 
-	char				 status[SMTPD_MAXLINESIZE];
+	char				 status[LINE_MAX];
 };
 
 struct mta_task {
@@ -844,7 +899,7 @@ struct mta_task {
 struct passwd;
 
 struct queue_backend {
-	int	(*init)(struct passwd *, int);
+	int	(*init)(struct passwd *, int, const char *);
 };
 
 struct compress_backend {
@@ -872,7 +927,7 @@ struct delivery_backend {
 };
 
 struct scheduler_backend {
-	int	(*init)(void);
+	int	(*init)(const char *);
 
 	int	(*insert)(struct scheduler_info *);
 	size_t	(*commit)(uint32_t);
@@ -883,7 +938,7 @@ struct scheduler_backend {
 	int	(*hold)(uint64_t, uint64_t);
 	int	(*release)(int, uint64_t, int);
 
-	int	(*batch)(int, struct scheduler_batch *);
+	int	(*batch)(int, int*, size_t*, uint64_t*, int*);
 
 	size_t	(*messages)(uint32_t, uint32_t *, size_t);
 	size_t	(*envelopes)(uint64_t, struct evpstate *, size_t);
@@ -891,6 +946,7 @@ struct scheduler_backend {
 	int	(*remove)(uint64_t);
 	int	(*suspend)(uint64_t);
 	int	(*resume)(uint64_t);
+	int	(*query)(uint64_t);
 };
 
 enum stat_type {
@@ -988,12 +1044,10 @@ extern int profiling;
 extern struct mproc *p_control;
 extern struct mproc *p_parent;
 extern struct mproc *p_lka;
-extern struct mproc *p_mda;
-extern struct mproc *p_mfa;
-extern struct mproc *p_mta;
 extern struct mproc *p_queue;
 extern struct mproc *p_scheduler;
-extern struct mproc *p_smtp;
+extern struct mproc *p_pony;
+extern struct mproc *p_ca;
 
 extern struct smtpd	*env;
 extern void (*imsg_callback)(struct mproc *, struct imsg *);
@@ -1016,12 +1070,6 @@ struct bounce_req_msg {
 	struct delivery_bounce	bounce;
 };
 
-enum mfa_resp_status {
-	MFA_OK,
-	MFA_FAIL,
-	MFA_CLOSE,
-};
-
 enum dns_error {
 	DNS_OK = 0,
 	DNS_RETRY,
@@ -1043,21 +1091,22 @@ enum ca_resp_status {
 
 struct ca_cert_req_msg {
 	uint64_t		reqid;
-	char			name[SMTPD_MAXPATHLEN];
+	char			name[HOST_NAME_MAX+1];
+	int			fallback;
 };
 
 struct ca_cert_resp_msg {
 	uint64_t		reqid;
 	enum ca_resp_status	status;
+	char			name[HOST_NAME_MAX+1];
 	char		       *cert;
 	off_t			cert_len;
-	char		       *key;
-	off_t			key_len;
 };
 
 struct ca_vrfy_req_msg {
 	uint64_t		reqid;
-	char			pkiname[SMTPD_MAXHOSTNAMELEN];
+	char			name[HOST_NAME_MAX+1];
+	int			fallback;
 	unsigned char  	       *cert;
 	off_t			cert_len;
 	size_t			n_chain;
@@ -1071,6 +1120,14 @@ struct ca_vrfy_resp_msg {
 	enum ca_resp_status	status;
 };
 
+struct msg_walkinfo {
+	struct event	 ev;
+	uint32_t	 msgid;
+	uint32_t	 peerid;
+	size_t		 n_evp;
+	void		*data;
+	int		 done;
+};
 
 /* aliases.c */
 int aliases_get(struct expand *, const char *);
@@ -1089,8 +1146,11 @@ void bounce_fd(int);
 
 
 /* ca.c */
-int	ca_X509_verify(void *, void *, const char *, const char *, const char **);
-
+pid_t	 ca(void);
+int	 ca_X509_verify(void *, void *, const char *, const char *, const char **);
+void	 ca_imsg(struct mproc *, struct imsg *);
+void	 ca_init(void);
+void	 ca_engine_init(void);
 
 /* compress_backend.c */
 struct compress_backend *compress_backend_lookup(const char *);
@@ -1104,7 +1164,8 @@ int	uncompress_file(FILE *, FILE *);
 #define PURGE_TABLES		0x02
 #define PURGE_RULES		0x04
 #define PURGE_PKI		0x08
-#define PURGE_EVERYTHING	0xff
+#define PURGE_PKI_KEYS		0x10
+#define PURGE_EVERYTHING	0x0f
 void purge_config(uint8_t);
 void init_pipes(void);
 void config_process(enum smtp_proc_type);
@@ -1130,10 +1191,6 @@ struct delivery_backend *delivery_backend_lookup(enum action_type);
 
 
 /* dns.c */
-void dns_query_host(uint64_t, const char *);
-void dns_query_ptr(uint64_t, const struct sockaddr *);
-void dns_query_mx(uint64_t, const char *);
-void dns_query_mx_preference(uint64_t, const char *, const char *);
 void dns_imsg(struct mproc *, struct imsg *);
 
 
@@ -1188,34 +1245,39 @@ void lka_session_forward_reply(struct forward_req *, int);
 
 /* log.c */
 void vlog(int, const char *, va_list);
-
+void logit(int, const char *, ...) __attribute__((format (printf, 2, 3)));
 
 /* mda.c */
-pid_t mda(void);
+void mda_postfork(void);
+void mda_postprivdrop(void);
+void mda_imsg(struct mproc *, struct imsg *);
 
 
-/* mfa.c */
-pid_t mfa(void);
-void mfa_ready(void);
+/* filter.c */
+void filter_postfork(void);
+void filter_configure(void);
+void filter_connect(uint64_t, const struct sockaddr *,
+    const struct sockaddr *, const char *, const char *);
+void filter_mailaddr(uint64_t, int, const struct mailaddr *);
+void filter_line(uint64_t, int, const char *);
+void filter_eom(uint64_t, int, size_t);
+void filter_event(uint64_t, int);
+void filter_build_fd_chain(uint64_t, int);
 
-/* mfa_session.c */
-void mfa_filter_prepare(void);
-void mfa_filter_init(void);
-void mfa_filter_connect(uint64_t, const struct sockaddr *,
-    const struct sockaddr *, const char *);
-void mfa_filter_mailaddr(uint64_t, int, const struct mailaddr *);
-void mfa_filter_line(uint64_t, int, const char *);
-void mfa_filter_eom(uint64_t, int, size_t);
-void mfa_filter(uint64_t, int);
-void mfa_filter_event(uint64_t, int);
-void mfa_build_fd_chain(uint64_t, int);
+
+/* mailaddr.c */
+int mailaddr_line(struct maddrmap *, const char *);
+void maddrmap_init(struct maddrmap *);
+void maddrmap_insert(struct maddrmap *, struct maddrnode *);
+void maddrmap_free(struct maddrmap *);
 
 /* mproc.c */
-int mproc_fork(struct mproc *, const char*, const char *);
+int mproc_fork(struct mproc *, const char*, char **);
 void mproc_init(struct mproc *, int);
 void mproc_clear(struct mproc *);
 void mproc_enable(struct mproc *);
 void mproc_disable(struct mproc *);
+void mproc_event_add(struct mproc *);
 void m_compose(struct mproc *, uint32_t, uint32_t, pid_t, int, void *, size_t);
 void m_composev(struct mproc *, uint32_t, uint32_t, pid_t, int,
     const struct iovec *, int);
@@ -1224,6 +1286,7 @@ void m_create(struct mproc *, uint32_t, uint32_t, pid_t, int);
 void m_add(struct mproc *, const void *, size_t);
 void m_add_int(struct mproc *, int);
 void m_add_u32(struct mproc *, uint32_t);
+void m_add_size(struct mproc *, size_t);
 void m_add_time(struct mproc *, time_t);
 void m_add_string(struct mproc *, const char *);
 void m_add_data(struct mproc *, const void *, size_t);
@@ -1233,12 +1296,15 @@ void m_add_id(struct mproc *, uint64_t);
 void m_add_sockaddr(struct mproc *, const struct sockaddr *);
 void m_add_mailaddr(struct mproc *, const struct mailaddr *);
 void m_add_envelope(struct mproc *, const struct envelope *);
+void m_add_params(struct mproc *, struct dict *);
 void m_close(struct mproc *);
+void m_flush(struct mproc *);
 
 void m_msg(struct msg *, struct imsg *);
 int  m_is_eom(struct msg *);
 void m_end(struct msg *);
 void m_get_int(struct msg *, int *);
+void m_get_size(struct msg *, size_t *);
 void m_get_u32(struct msg *, uint32_t *);
 void m_get_time(struct msg *, time_t *);
 void m_get_string(struct msg *, const char **);
@@ -1249,10 +1315,14 @@ void m_get_id(struct msg *, uint64_t *);
 void m_get_sockaddr(struct msg *, struct sockaddr *);
 void m_get_mailaddr(struct msg *, struct mailaddr *);
 void m_get_envelope(struct msg *, struct envelope *);
+void m_get_params(struct msg *, struct dict *);
+void m_clear_params(struct dict *);
 
 
 /* mta.c */
-pid_t mta(void);
+void mta_postfork(void);
+void mta_postprivdrop(void);
+void mta_imsg(struct mproc *, struct imsg *);
 void mta_route_ok(struct mta_relay *, struct mta_route *);
 void mta_route_error(struct mta_relay *, struct mta_route *);
 void mta_route_down(struct mta_relay *, struct mta_route *);
@@ -1276,10 +1346,6 @@ int cmdline_symset(char *);
 
 /* queue.c */
 pid_t queue(void);
-void queue_ok(uint64_t);
-void queue_tempfail(uint64_t, const char *, enum enhanced_status_code);
-void queue_permfail(uint64_t, const char *, enum enhanced_status_code);
-void queue_loop(uint64_t);
 void queue_flow_control(void);
 
 
@@ -1287,18 +1353,20 @@ void queue_flow_control(void);
 uint32_t queue_generate_msgid(void);
 uint64_t queue_generate_evpid(uint32_t);
 int queue_init(const char *, int);
+int queue_close(void);
 int queue_message_create(uint32_t *);
 int queue_message_delete(uint32_t);
 int queue_message_commit(uint32_t);
 int queue_message_fd_r(uint32_t);
 int queue_message_fd_rw(uint32_t);
 int queue_message_corrupt(uint32_t);
+int queue_message_uncorrupt(uint32_t);
 int queue_envelope_create(struct envelope *);
 int queue_envelope_delete(uint64_t);
 int queue_envelope_load(uint64_t, struct envelope *);
 int queue_envelope_update(struct envelope *);
 int queue_envelope_walk(struct envelope *);
-
+int queue_message_walk(struct envelope *, uint32_t, int *, void **);
 
 /* ruleset.c */
 struct rule *ruleset_match(const struct envelope *);
@@ -1311,11 +1379,18 @@ pid_t scheduler(void);
 /* scheduler_bakend.c */
 struct scheduler_backend *scheduler_backend_lookup(const char *);
 void scheduler_info(struct scheduler_info *, struct envelope *);
-time_t scheduler_compute_schedule(struct scheduler_info *);
+
+
+/* pony.c */
+pid_t pony(void);
+void pony_imsg(struct mproc *, struct imsg *);
 
 
 /* smtp.c */
-pid_t smtp(void);
+void smtp_postfork(void);
+void smtp_postprivdrop(void);
+void smtp_imsg(struct mproc *, struct imsg *);
+void smtp_configure(void);
 void smtp_collect(void);
 
 
@@ -1323,6 +1398,8 @@ void smtp_collect(void);
 int smtp_session(struct listener *, int, const struct sockaddr_storage *,
     const char *);
 void smtp_session_imsg(struct mproc *, struct imsg *);
+void smtp_filter_response(uint64_t, int, int, uint32_t, const char *);
+void smtp_filter_fd(uint64_t, int);
 
 
 /* smtpd.c */
@@ -1331,11 +1408,13 @@ void post_fork(int);
 const char *proc_name(enum smtp_proc_type);
 const char *proc_title(enum smtp_proc_type);
 const char *imsg_to_str(int);
+void log_imsg(int, int, struct imsg *);
+int fork_proc_backend(const char *, const char *, const char *);
 
 
 /* ssl_smtpd.c */
-void   *ssl_mta_init(char *, off_t, char *, off_t);
-void   *ssl_smtp_init(void *, char *, off_t, char *, off_t, void *, void *);
+void   *ssl_mta_init(void *, char *, off_t, const char *);
+void   *ssl_smtp_init(void *);
 
 
 /* stat_backend.c */
@@ -1360,19 +1439,17 @@ void	table_close(struct table *);
 int	table_check_use(struct table *, uint32_t, uint32_t);
 int	table_check_type(struct table *, uint32_t);
 int	table_check_service(struct table *, uint32_t);
-int	table_lookup(struct table *, const char *, enum table_service,
+int	table_lookup(struct table *, struct dict *, const char *, enum table_service,
     union lookup *);
-int	table_fetch(struct table *, enum table_service, union lookup *);
+int	table_fetch(struct table *, struct dict *, enum table_service, union lookup *);
 void table_destroy(struct table *);
 void table_add(struct table *, const char *, const char *);
-void table_delete(struct table *, const char *);
 int table_domain_match(const char *, const char *);
 int table_netaddr_match(const char *, const char *);
 int table_mailaddr_match(const char *, const char *);
 void	table_open_all(void);
 void	table_dump_all(void);
 void	table_close_all(void);
-const void	*table_get(struct table *, const char *);
 int table_parse_lookup(enum table_service, const char *, const char *,
     union lookup *);
 
@@ -1411,6 +1488,7 @@ int bsnprintf(char *, size_t, const char *, ...)
 int mkdirs(char *, mode_t);
 int safe_fclose(FILE *);
 int hostname_match(const char *, const char *);
+int mailaddr_match(const struct mailaddr *, const struct mailaddr *);
 int valid_localpart(const char *);
 int valid_domainpart(const char *);
 int secure_file(int, char *, char *, uid_t, int);
@@ -1419,6 +1497,7 @@ void xlowercase(char *, const char *, size_t);
 int  uppercase(char *, const char *, size_t);
 uint64_t generate_uid(void);
 int availdesc(void);
+int ckdir_quiet(const char *, mode_t, uid_t, gid_t);
 int ckdir(const char *, mode_t, uid_t, gid_t, int);
 int rmtree(char *, int);
 int mvpurge(char *, char *);
