@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: ssl.h,v 1.9 2014/05/20 17:33:36 reyk Exp $	*/
 /*
  * Copyright (c) 2013 Gilles Chehade <gilles@poolp.org>
  *
@@ -20,11 +20,7 @@
 #define	SSL_SESSION_TIMEOUT	300
 
 struct pki {
-	char			 pki_name[PATH_MAX];
-
-	char			*pki_ca_file;
-	char			*pki_ca;
-	off_t			 pki_ca_len;
+	char			 pki_name[HOST_NAME_MAX+1];
 
 	char			*pki_cert_file;
 	char			*pki_cert;
@@ -34,21 +30,33 @@ struct pki {
 	char			*pki_key;
 	off_t			 pki_key_len;
 
+	EVP_PKEY		*pki_pkey;
+
 	char			*pki_dhparams_file;
 	char			*pki_dhparams;
 	off_t			 pki_dhparams_len;
 };
 
+struct ca {
+       	char			 ca_name[HOST_NAME_MAX+1];
+
+	char			*ca_cert_file;
+	char			*ca_cert;
+	off_t			 ca_cert_len;
+};
+
 /* ssl.c */
 void		ssl_init(void);
-int		ssl_setup(SSL_CTX **, struct pki *);
-SSL_CTX	       *ssl_ctx_create(void);
+int		ssl_setup(SSL_CTX **, struct pki *, int (*)(SSL *, int *, void *),
+	    const char *, const char *);
+SSL_CTX	       *ssl_ctx_create(const char *, char *, off_t, const char *);
 int	        ssl_cmp(struct pki *, struct pki *);
+DH	       *get_dh(void);
 DH	       *get_dh1024(void);
+DH	       *get_dh2048(void);
 DH	       *get_dh_from_memory(char *, size_t);
 void		ssl_set_ephemeral_key_exchange(SSL_CTX *, DH *);
 void		ssl_set_ecdh_curve(SSL_CTX *, const char *);
-extern int	ssl_ctx_load_verify_memory(SSL_CTX *, char *, off_t);
 char	       *ssl_load_file(const char *, off_t *, mode_t);
 char	       *ssl_load_key(const char *, off_t *, char *, mode_t, const char *);
 
@@ -57,12 +65,13 @@ void		ssl_error(const char *);
 
 int		ssl_load_certificate(struct pki *, const char *);
 int		ssl_load_keyfile(struct pki *, const char *, const char *);
-int		ssl_load_cafile(struct pki *, const char *);
 int		ssl_load_dhparams(struct pki *, const char *);
+int		ssl_load_pkey(const void *, size_t, char *, off_t,
+		    X509 **, EVP_PKEY **);
+int		ssl_ctx_fake_private_key(SSL_CTX *, const void *, size_t,
+		    char *, off_t, X509 **, EVP_PKEY **);
 
+int		ssl_load_cafile(struct ca *, const char *);
 
-/* ssl_privsep.c */
-int	 ssl_ctx_use_private_key(SSL_CTX *, char *, off_t);
-int	 ssl_ctx_use_certificate_chain(SSL_CTX *, char *, off_t);
-int      ssl_ctx_load_verify_memory(SSL_CTX *, char *, off_t);
-int      ssl_by_mem_ctrl(X509_LOOKUP *, int, const char *, long, char **);
+/* libressl.c */
+int SSL_CTX_use_certificate_chain_mem(SSL_CTX *, void *, int);
