@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: table.c,v 1.16 2014/05/09 21:30:11 tedu Exp $	*/
 
 /*
  * Copyright (c) 2013 Eric Faurot <eric@openbsd.org>
@@ -36,6 +36,8 @@
 #include <imsg.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <netdb.h>
+#include <limits.h>
 #include <string.h>
 #include <unistd.h>
 
@@ -45,7 +47,9 @@
 struct table_backend *table_backend_lookup(const char *);
 
 extern struct table_backend table_backend_static;
+#ifdef HAVE_DB_API
 extern struct table_backend table_backend_db;
+#endif
 extern struct table_backend table_backend_getpwnam;
 extern struct table_backend table_backend_proc;
 
@@ -61,8 +65,10 @@ table_backend_lookup(const char *backend)
 {
 	if (!strcmp(backend, "static") || !strcmp(backend, "file"))
 		return &table_backend_static;
+#ifdef HAVE_DB_API
 	if (!strcmp(backend, "db"))
 		return &table_backend_db;
+#endif
 	if (!strcmp(backend, "getpwnam"))
 		return &table_backend_getpwnam;
 	if (!strcmp(backend, "proc"))
@@ -75,8 +81,10 @@ table_backend_name(struct table_backend *backend)
 {
 	if (backend == &table_backend_static)
 		return "static";
+#ifdef HAVE_DB_API
 	if (backend == &table_backend_db)
 		return "db";
+#endif
 	if (backend == &table_backend_getpwnam)
 		return "getpwnam";
 	if (backend == &table_backend_proc)
@@ -97,6 +105,7 @@ table_service_name(enum table_service s)
 	case K_SOURCE:		return "SOURCE";
 	case K_MAILADDR:	return "MAILADDR";
 	case K_ADDRNAME:	return "ADDRNAME";
+	case K_MAILADDRMAP:	return "MAILADDRMAP";
 	default:		return "???";
 	}
 }
@@ -104,12 +113,12 @@ table_service_name(enum table_service s)
 struct table *
 table_find(const char *name, const char *tag)
 {
-	char buf[SMTPD_MAXLINESIZE];
+	char buf[LINE_MAX];
 
 	if (tag == NULL)
 		return dict_get(env->sc_tables_dict, name);
 
-	if (snprintf(buf, sizeof(buf), "%s#%s", name, tag) >= (int)sizeof(buf)) {
+	if ((size_t)snprintf(buf, sizeof(buf), "%s#%s", name, tag) >= sizeof(buf)) {
 		log_warnx("warn: table name too long: %s#%s", name, tag);
 		return (NULL);
 	}
@@ -118,7 +127,7 @@ table_find(const char *name, const char *tag)
 }
 
 int
-table_lookup(struct table *table, const char *key, enum table_service kind,
+table_lookup(struct table *table, struct dict *params, const char *key, enum table_service kind,
     union lookup *lk)
 {
 	int	r;
@@ -132,7 +141,7 @@ table_lookup(struct table *table, const char *key, enum table_service kind,
 		return -1;
 	}
 
-	r = table->t_backend->lookup(table->t_handle, lkey, kind, lk);
+	r = table->t_backend->lookup(table->t_handle, params, lkey, kind, lk);
 
 	if (r == 1)
 		log_trace(TRACE_LOOKUP, "lookup: %s \"%s\" as %s in table %s:%s -> %s%s%s",
@@ -157,14 +166,14 @@ table_lookup(struct table *table, const char *key, enum table_service kind,
 }
 
 int
-table_fetch(struct table *table, enum table_service kind, union lookup *lk)
+table_fetch(struct table *table, struct dict *params, enum table_service kind, union lookup *lk)
 {
 	int 	r;
 
 	if (table->t_backend->fetch == NULL)
 		return (-1);
 
-	r = table->t_backend->fetch(table->t_handle, kind, lk);
+	r = table->t_backend->fetch(table->t_handle, params, kind, lk);
 
 	if (r == 1)
 		log_trace(TRACE_LOOKUP, "lookup: fetch %s from table %s:%s -> %s%s%s",
@@ -190,40 +199,43 @@ table_create(const char *backend, const char *name, const char *tag,
 {
 	struct table		*t;
 	struct table_backend	*tb;
-	char			 buf[SMTPD_MAXLINESIZE];
-	char			 path[SMTPD_MAXLINESIZE];
+	char			 buf[LINE_MAX];
+	char			 path[LINE_MAX];
 	size_t			 n;
 	struct stat		 sb;
 
 	if (name && tag) {
-		if (snprintf(buf, sizeof(buf), "%s#%s", name, tag)
-		    >= (int)sizeof(buf))
-			errx(1, "table_create: name too long \"%s#%s\"",
+		if ((size_t)snprintf(buf, sizeof(buf), "%s#%s", name, tag) >=
+		    sizeof(buf))
+			fatalx("table_create: name too long \"%s#%s\"",
 			    name, tag);
 		name = buf;
 	}
 
 	if (name && table_find(name, NULL))
-		errx(1, "table_create: table \"%s\" already defined", name);
+		fatalx("table_create: table \"%s\" already defined", name);
 
 	if ((tb = table_backend_lookup(backend)) == NULL) {
-		if (snprintf(path, sizeof(path), PATH_TABLES "/table-%s",
-		    backend) >= (int)sizeof(path)) {
-			errx(1, "table_create: path too long \""
-			    PATH_TABLES "/table-%s\"", backend);
+		if ((size_t)snprintf(path, sizeof(path), PATH_LIBEXEC "/table-%s",
+		    backend) >= sizeof(path)) {
+			fatalx("table_create: path too long \""
+			    PATH_LIBEXEC "/table-%s\"", backend);
 		}
 		if (stat(path, &sb) == 0) {
 			tb = table_backend_lookup("proc");
+			(void)strlcpy(path, backend, sizeof(path));
 			if (config) {
-				strlcat(path, " ", sizeof(path));
-				strlcat(path, config, sizeof(path));
+				(void)strlcat(path, ":", sizeof(path));
+				if (strlcat(path, config, sizeof(path))
+				    >= sizeof(path))
+					fatalx("table_create: config file path too long");
 			}
 			config = path;
 		}
 	}
 
 	if (tb == NULL)
-		errx(1, "table_create: backend \"%s\" does not exist", backend);
+		fatalx("table_create: backend \"%s\" does not exist", backend);
 
 	t = xcalloc(1, sizeof(*t), "table_create");
 	t->t_backend = tb;
@@ -238,7 +250,7 @@ table_create(const char *backend, const char *name, const char *tag,
 	if (config) {
 		if (strlcpy(t->t_config, config, sizeof t->t_config)
 		    >= sizeof t->t_config)
-			errx(1, "table_create: table config \"%s\" too large",
+			fatalx("table_create: table config \"%s\" too large",
 			    t->t_config);
 	}
 
@@ -246,12 +258,12 @@ table_create(const char *backend, const char *name, const char *tag,
 		t->t_type = T_DYNAMIC;
 
 	if (name == NULL)
-		snprintf(t->t_name, sizeof(t->t_name), "<dynamic:%u>",
+		(void)snprintf(t->t_name, sizeof(t->t_name), "<dynamic:%u>",
 		    last_table_id++);
 	else {
 		n = strlcpy(t->t_name, name, sizeof(t->t_name));
 		if (n >= sizeof(t->t_name))
-			errx(1, "table_create: table name too long");
+			fatalx("table_create: table name too long");
 	}
 
 	dict_init(&t->t_dict);
@@ -286,7 +298,7 @@ table_add(struct table *t, const char *key, const char *val)
 	char	lkey[1024], *old;
 
 	if (t->t_type & T_DYNAMIC)
-		errx(1, "table_add: cannot add to table");
+		fatalx("table_add: cannot add to table");
 
 	if (! lowercase(lkey, key, sizeof lkey)) {
 		log_warnx("warn: lookup key too long: %s", key);
@@ -301,22 +313,6 @@ table_add(struct table *t, const char *key, const char *val)
 	}
 }
 
-const void *
-table_get(struct table *t, const char *key)
-{
-	if (t->t_type & T_DYNAMIC)
-		errx(1, "table_get: cannot get from table");
-	return dict_get(&t->t_dict, key);
-}
-
-void
-table_delete(struct table *t, const char *key)
-{
-	if (t->t_type & T_DYNAMIC)
-		errx(1, "table_delete: cannot delete from table");
-	free(dict_pop(&t->t_dict, key));
-}
-
 int
 table_check_type(struct table *t, uint32_t mask)
 {
@@ -362,6 +358,12 @@ table_update(struct table *t)
 	return (t->t_backend->update(t));
 }
 
+
+/*
+ * quick reminder:
+ * in *_match() s1 comes from session, s2 comes from table
+ */
+
 int
 table_domain_match(const char *s1, const char *s2)
 {
@@ -378,14 +380,7 @@ table_mailaddr_match(const char *s1, const char *s2)
 		return 0;
 	if (! text_to_mailaddr(&m2, s2))
 		return 0;
-
-	if (! table_domain_match(m1.domain, m2.domain))
-		return 0;
-
-	if (m2.user[0])
-		if (strcasecmp(m1.user, m2.user))
-			return 0;
-	return 1;
+	return mailaddr_match(&m1, &m2);
 }
 
 static int table_match_mask(struct sockaddr_storage *, struct netaddr *);
@@ -485,18 +480,17 @@ table_dump_all(void)
 		sep = "";
  		buf[0] = '\0';
 		if (t->t_type & T_DYNAMIC) {
-			strlcat(buf, "DYNAMIC", sizeof(buf));
+			(void)strlcat(buf, "DYNAMIC", sizeof(buf));
 			sep = ",";
 		}
 		if (t->t_type & T_LIST) {
-			strlcat(buf, sep, sizeof(buf));
-			strlcat(buf, "LIST", sizeof(buf));
+			(void)strlcat(buf, sep, sizeof(buf));
+			(void)strlcat(buf, "LIST", sizeof(buf));
 			sep = ",";
 		}
 		if (t->t_type & T_HASH) {
-			strlcat(buf, sep, sizeof(buf));
-			strlcat(buf, "HASH", sizeof(buf));
-			sep = ",";
+			(void)strlcat(buf, sep, sizeof(buf));
+			(void)strlcat(buf, "HASH", sizeof(buf));
 		}
 		log_debug("TABLE \"%s\" type=%s config=\"%s\"",
 		    t->t_name, buf, t->t_config);
@@ -518,7 +512,7 @@ table_open_all(void)
 	iter = NULL;
 	while (dict_iter(env->sc_tables_dict, &iter, NULL, (void **)&t))
 		if (! table_open(t))
-			errx(1, "failed to open table %s", t->t_name);
+			fatalx("failed to open table %s", t->t_name);
 }
 
 void
@@ -536,7 +530,7 @@ int
 table_parse_lookup(enum table_service service, const char *key,
     const char *line, union lookup *lk)
 {
-	char	buffer[SMTPD_MAXLINESIZE], *p;
+	char	buffer[LINE_MAX], *p;
 	size_t	len;
 
 	len = strlen(line);
@@ -565,7 +559,7 @@ table_parse_lookup(enum table_service service, const char *key,
 			return (-1);
 
 		/* too big to fit in a smtp session line */
-		if (len >= SMTPD_MAXLINESIZE)
+		if (len >= LINE_MAX)
 			return (-1);
 
 		p = strchr(line, ':');
@@ -614,6 +608,17 @@ table_parse_lookup(enum table_service service, const char *key,
 			return (-1);
 		return (1);
 
+	case K_MAILADDRMAP:
+		lk->maddrmap = calloc(1, sizeof(*lk->maddrmap));
+		if (lk->maddrmap == NULL)
+			return (-1);
+		maddrmap_init(lk->maddrmap);
+		if (! mailaddr_line(lk->maddrmap, line)) {
+			maddrmap_free(lk->maddrmap);
+			return (-1);
+		}
+		return (1);
+		
 	case K_ADDRNAME:
 		if (parse_sockaddr((struct sockaddr *)&lk->addrname.addr,
 		    PF_UNSPEC, key) == -1)
@@ -631,7 +636,8 @@ table_parse_lookup(enum table_service service, const char *key,
 static const char *
 table_dump_lookup(enum table_service s, union lookup *lk)
 {
-	static char	buf[SMTPD_MAXLINESIZE];
+	static char	buf[LINE_MAX];
+	int		ret;
 
 	switch (s) {
 	case K_NONE:
@@ -642,42 +648,56 @@ table_dump_lookup(enum table_service s, union lookup *lk)
 		break;
 
 	case K_DOMAIN:
-		snprintf(buf, sizeof(buf), "%s", lk->domain.name);
+		ret = snprintf(buf, sizeof(buf), "%s", lk->domain.name);
+		if (ret == -1 || (size_t)ret >= sizeof (buf))
+			goto err;
 		break;
 
 	case K_CREDENTIALS:
-		snprintf(buf, sizeof(buf), "%s:%s",
+		ret = snprintf(buf, sizeof(buf), "%s:%s",
 		    lk->creds.username, lk->creds.password);
+		if (ret == -1 || (size_t)ret >= sizeof (buf))
+			goto err;
 		break;
 
 	case K_NETADDR:
-		snprintf(buf, sizeof(buf), "%s/%d",
+		ret = snprintf(buf, sizeof(buf), "%s/%d",
 		    sockaddr_to_text((struct sockaddr *)&lk->netaddr.ss),
 		    lk->netaddr.bits);
+		if (ret == -1 || (size_t)ret >= sizeof (buf))
+			goto err;
 		break;
 
 	case K_USERINFO:
-		snprintf(buf, sizeof(buf), "%s:%d:%d:%s",
+		ret = snprintf(buf, sizeof(buf), "%s:%d:%d:%s",
 		    lk->userinfo.username,
 		    lk->userinfo.uid,
 		    lk->userinfo.gid,
 		    lk->userinfo.directory);
+		if (ret == -1 || (size_t)ret >= sizeof (buf))
+			goto err;
 		break;
 
 	case K_SOURCE:
-		snprintf(buf, sizeof(buf), "%s",
+		ret = snprintf(buf, sizeof(buf), "%s",
 		    ss_to_text(&lk->source.addr));
+		if (ret == -1 || (size_t)ret >= sizeof (buf))
+			goto err;
 		break;
 
 	case K_MAILADDR:
-		snprintf(buf, sizeof(buf), "%s@%s",
+		ret = snprintf(buf, sizeof(buf), "%s@%s",
 		    lk->mailaddr.user,
 		    lk->mailaddr.domain);
+		if (ret == -1 || (size_t)ret >= sizeof (buf))
+			goto err;
 		break;
 
 	case K_ADDRNAME:
-		snprintf(buf, sizeof(buf), "%s",
+		ret = snprintf(buf, sizeof(buf), "%s",
 		    lk->addrname.name);
+		if (ret == -1 || (size_t)ret >= sizeof (buf))
+			goto err;
 		break;
 
 	default:
@@ -685,6 +705,9 @@ table_dump_lookup(enum table_service s, union lookup *lk)
 	}
 
 	return (buf);
+
+err:
+	return (NULL);
 }
 
 
