@@ -20,7 +20,11 @@
 
 #include <sys/types.h>
 #include <sys/queue.h>
-#include <sys/uio.h>
+#include <sys/tree.h>
+#include <sys/socket.h>
+
+#include <netinet/in.h>
+#include <netdb.h>
 
 #include <event.h>
 #include <fcntl.h>
@@ -31,15 +35,16 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <limits.h>
 
 #include "smtpd-defines.h"
 #include "smtpd-api.h"
 #include "log.h"
 
 static int (*handler_update)(void);
-static int (*handler_check)(int, const char *);
-static int (*handler_lookup)(int, const char *, char *, size_t);
-static int (*handler_fetch)(int, char *, size_t);
+static int (*handler_check)(int, struct dict *, const char *);
+static int (*handler_lookup)(int, struct dict *, const char *, char *, size_t);
+static int (*handler_fetch)(int, struct dict *, char *, size_t);
 
 static int		 quit;
 static struct imsgbuf	 ibuf;
@@ -104,13 +109,44 @@ table_msg_close(void)
 	buf = NULL;
 }
 
+static int
+table_read_params(struct dict *params)
+{
+	size_t	count;
+	char	*key;
+	char	*value;
+
+	dict_init(params);
+
+	table_msg_get(&count, sizeof(count));
+
+	for (;count; count--) {
+		key = rdata;
+		table_msg_get(NULL, strlen(key) + 1);
+		value = rdata;
+		table_msg_get(NULL, strlen(value) + 1);
+		dict_set(params, key, value);
+	}
+
+	return (0);
+}
+
+static void
+table_clear_params(struct dict *params)
+{
+	while (dict_poproot(params, NULL))
+		;
+}
+
 static void
 table_msg_dispatch(void)
 {
 	struct table_open_params op;
+	struct dict	 params;
 	char		 res[4096];
 	int		 type, r;
 
+	memset(res, 0, sizeof res);
 	switch (imsg.hdr.type) {
 	case PROC_TABLE_OPEN:
 		table_msg_get(&op, sizeof op);
@@ -145,6 +181,7 @@ table_msg_dispatch(void)
 
 	case PROC_TABLE_CHECK:
 		table_msg_get(&type, sizeof(type));
+		table_read_params(&params);
 		if (rlen == 0) {
 			log_warnx("warn: table-api: no key");
 			fatalx("table-api: exiting");
@@ -155,9 +192,10 @@ table_msg_dispatch(void)
 		}
 
 		if (handler_check)
-			r = handler_check(type, rdata);
+			r = handler_check(type, &params, rdata);
 		else
 			r = -1;
+		table_clear_params(&params);
 		table_msg_get(NULL, rlen);
 		table_msg_end();
 
@@ -167,7 +205,7 @@ table_msg_dispatch(void)
 
 	case PROC_TABLE_LOOKUP:
 		table_msg_get(&type, sizeof(type));
-
+		table_read_params(&params);
 		if (rlen == 0) {
 			log_warnx("warn: table-api: no key");
 			fatalx("table-api: exiting");
@@ -178,10 +216,10 @@ table_msg_dispatch(void)
 		}
 
 		if (handler_lookup)
-			r = handler_lookup(type, rdata, res, sizeof(res));
+			r = handler_lookup(type, &params, rdata, res, sizeof(res));
 		else
 			r = -1;
-
+		table_clear_params(&params);
 		table_msg_get(NULL, rlen);
 		table_msg_end();
 
@@ -194,12 +232,13 @@ table_msg_dispatch(void)
 
 	case PROC_TABLE_FETCH:
 		table_msg_get(&type, sizeof(type));
-		table_msg_end();
-
+		table_read_params(&params);
 		if (handler_fetch)
-			r = handler_fetch(type, res, sizeof(res));
+			r = handler_fetch(type, &params, res, sizeof(res));
 		else
 			r = -1;
+		table_clear_params(&params);
+		table_msg_end();
 
 		table_msg_add(&r, sizeof(r));
 		if (r == 1)
@@ -220,19 +259,19 @@ table_api_on_update(int(*cb)(void))
 }
 
 void
-table_api_on_check(int(*cb)(int, const char *))
+table_api_on_check(int(*cb)(int, struct dict *, const char *))
 {
 	handler_check = cb;
 }
 
 void
-table_api_on_lookup(int(*cb)(int, const char *, char *, size_t))
+table_api_on_lookup(int(*cb)(int, struct dict  *, const char *, char *, size_t))
 {
 	handler_lookup = cb;
 }
 
 void
-table_api_on_fetch(int(*cb)(int, char *, size_t))
+table_api_on_fetch(int(*cb)(int, struct dict *, char *, size_t))
 {
 	handler_fetch = cb;
 }
