@@ -21,7 +21,6 @@
 #include <sys/types.h>
 #include <sys/queue.h>
 #include <sys/tree.h>
-#include <sys/param.h>
 #include <sys/socket.h>
 
 #include <ctype.h>
@@ -33,6 +32,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <limits.h>
 #include <unistd.h>
 
 #include "smtpd.h"
@@ -121,40 +121,17 @@ table_proc_end(void)
 static void *
 table_proc_open(struct table *table)
 {
-	int			 sp[2];
 	struct table_proc_priv	*priv;
-	char			*environ_new[2];
 	struct table_open_params op;
+	int			 fd;
 
-	errno = 0;
-
-	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, sp) < 0) {
-		log_warn("warn: table-proc: socketpair");
-		return (NULL);
-	}
 	priv = xcalloc(1, sizeof(*priv), "table_proc_open");
 
-	if ((priv->pid = fork()) == -1) {
-		log_warn("warn: table-proc: fork");
-		goto err;
-	}
+	fd = fork_proc_backend("table", table->t_config, table->t_name);
+	if (fd == -1)
+		fatalx("table-proc: exiting");
 
-	if (priv->pid == 0) {
-		/* child process */
-		dup2(sp[0], STDIN_FILENO);
-		closefrom(STDERR_FILENO + 1);
-
-		environ_new[0] = "PATH=" _PATH_DEFPATH;
-		environ_new[1] = (char *)NULL;
-		environ = environ_new;
-		execle("/bin/sh", "/bin/sh", "-c", table->t_config, (char *)NULL,
-		    environ_new);
-		fatal("execl");
-	}
-
-	/* parent process */
-	close(sp[0]);
-	imsg_init(&priv->ibuf, sp[1]);
+	imsg_init(&priv->ibuf, fd);
 
 	memset(&op, 0, sizeof op);
 	op.version = PROC_TABLE_API_VERSION;
@@ -165,11 +142,6 @@ table_proc_open(struct table *table)
 	table_proc_end();
 
 	return (priv);
-err:
-	free(priv);
-	close(sp[0]);
-	close(sp[1]);
-	return (NULL);
 }
 
 static int
@@ -197,7 +169,36 @@ table_proc_close(void *arg)
 }
 
 static int
-table_proc_lookup(void *arg, const char *k, enum table_service s,
+imsg_add_params(struct ibuf *buf, struct dict *params)
+{
+	size_t count;
+	const char *key;
+	char *value;
+	void *iter;
+
+	count = 0;
+	if (params)
+		count = dict_count(params);
+
+	if (imsg_add(buf, &count, sizeof(count)) == -1)
+		return (-1);
+
+	if (count == 0)
+		return (0);
+
+	iter = NULL;
+	while (dict_iter(params, &iter, &key, (void **)&value)) {
+		if (imsg_add(buf, key, strlen(key) + 1) == -1)
+			return (-1);
+		if (imsg_add(buf, value, strlen(value) + 1) == -1)
+			return (-1);
+	}
+
+	return (0);
+}
+
+static int
+table_proc_lookup(void *arg, struct dict *params, const char *k, enum table_service s,
     union lookup *lk)
 {
 	struct table_proc_priv	*priv = arg;
@@ -212,6 +213,8 @@ table_proc_lookup(void *arg, const char *k, enum table_service s,
 		return (-1);
 	if (imsg_add(buf, &s, sizeof(s)) == -1)
 		return (-1);
+	if (imsg_add_params(buf, params) == -1)
+		return (-1);
 	if (imsg_add(buf, k, strlen(k) + 1) == -1)
 		return (-1);
 	imsg_close(&priv->ibuf, buf);
@@ -238,7 +241,7 @@ table_proc_lookup(void *arg, const char *k, enum table_service s,
 }
 
 static int
-table_proc_fetch(void *arg, enum table_service s, union lookup *lk)
+table_proc_fetch(void *arg, struct dict *params, enum table_service s, union lookup *lk)
 {
 	struct table_proc_priv	*priv = arg;
 	struct ibuf		*buf;
@@ -249,6 +252,8 @@ table_proc_fetch(void *arg, enum table_service s, union lookup *lk)
 		return (-1);
 	if (imsg_add(buf, &s, sizeof(s)) == -1)
 		return (-1);
+	if (imsg_add_params(buf, params) == -1)
+		return (-1);
 	imsg_close(&priv->ibuf, buf);
 
 	table_proc_call(priv);
