@@ -1,4 +1,4 @@
-/*	$OpenBSD$	*/
+/*	$OpenBSD: to.c,v 1.17 2014/04/19 14:27:29 gilles Exp $	*/
 
 /*
  * Copyright (c) 2009 Jacek Masiulaniec <jacekm@dobremiasto.net>
@@ -37,6 +37,7 @@
 #include <fcntl.h>
 #include <fts.h>
 #include <imsg.h>
+#include <limits.h>
 #include <inttypes.h>
 #include <libgen.h>
 #include <netdb.h>
@@ -52,6 +53,7 @@
 #include "log.h"
 
 static const char *in6addr_to_text(const struct in6_addr *);
+static int alias_is_maildir(struct expandnode *, const char *, size_t);
 static int alias_is_filter(struct expandnode *, const char *, size_t);
 static int alias_is_username(struct expandnode *, const char *, size_t);
 static int alias_is_address(struct expandnode *, const char *, size_t);
@@ -103,7 +105,7 @@ text_to_mailaddr(struct mailaddr *maddr, const char *email)
 {
 	char *username;
 	char *hostname;
-	char  buffer[SMTPD_MAXLINESIZE];
+	char  buffer[LINE_MAX];
 
 	if (strlcpy(buffer, email, sizeof buffer) >= sizeof buffer)
 		return 0;
@@ -140,10 +142,10 @@ text_to_mailaddr(struct mailaddr *maddr, const char *email)
 const char *
 mailaddr_to_text(const struct mailaddr *maddr)
 {
-	static char  buffer[SMTPD_MAXLINESIZE];
+	static char  buffer[LINE_MAX];
 
-	strlcpy(buffer, maddr->user, sizeof buffer);
-	strlcat(buffer, "@", sizeof buffer);
+	(void)strlcpy(buffer, maddr->user, sizeof buffer);
+	(void)strlcat(buffer, "@", sizeof buffer);
 	if (strlcat(buffer, maddr->domain, sizeof buffer) >= sizeof buffer)
 		return NULL;
 
@@ -161,13 +163,13 @@ sa_to_text(const struct sockaddr *sa)
 	p = buf;
 
 	if (sa->sa_family == AF_LOCAL)
-		strlcpy(buf, "local", sizeof buf);
+		(void)strlcpy(buf, "local", sizeof buf);
 	else if (sa->sa_family == AF_INET) {
 		in_addr_t addr;
 
 		addr = ((const struct sockaddr_in *)sa)->sin_addr.s_addr;
 		addr = ntohl(addr);
-		bsnprintf(p, NI_MAXHOST, "%d.%d.%d.%d",
+		(void)bsnprintf(p, NI_MAXHOST, "%d.%d.%d.%d",
 		    (addr >> 24) & 0xff, (addr >> 16) & 0xff,
 		    (addr >> 8) & 0xff, addr & 0xff);
 	}
@@ -176,10 +178,10 @@ sa_to_text(const struct sockaddr *sa)
 		const struct in6_addr	*in6_addr;
 
 		in6 = (const struct sockaddr_in6 *)sa;
-		strlcpy(buf, "IPv6:", sizeof(buf));
+		(void)strlcpy(buf, "IPv6:", sizeof(buf));
 		p = buf + 5;
 		in6_addr = &in6->sin6_addr;
-		bsnprintf(p, NI_MAXHOST, "%s", in6addr_to_text(in6_addr));
+		(void)bsnprintf(p, NI_MAXHOST, "%s", in6addr_to_text(in6_addr));
 	}
 
 	return (buf);
@@ -228,13 +230,13 @@ duration_to_text(time_t t)
 	long long	d;
 
 	if (t == 0) {
-		strlcpy(dst, "0s", sizeof dst);
+		(void)strlcpy(dst, "0s", sizeof dst);
 		return (dst);
 	}
 
 	dst[0] = '\0';
 	if (t < 0) {
-		strlcpy(dst, "-", sizeof dst);
+		(void)strlcpy(dst, "-", sizeof dst);
 		t = -t;
 	}
 
@@ -246,20 +248,20 @@ duration_to_text(time_t t)
 	d = t / 24;
 
 	if (d) {
-		snprintf(buf, sizeof buf, "%lldd", d);
-		strlcat(dst, buf, sizeof dst);
+		(void)snprintf(buf, sizeof buf, "%lldd", d);
+		(void)strlcat(dst, buf, sizeof dst);
 	}
 	if (h) {
-		snprintf(buf, sizeof buf, "%dh", h);
-		strlcat(dst, buf, sizeof dst);
+		(void)snprintf(buf, sizeof buf, "%dh", h);
+		(void)strlcat(dst, buf, sizeof dst);
 	}
 	if (m) {
-		snprintf(buf, sizeof buf, "%dm", m);
-		strlcat(dst, buf, sizeof dst);
+		(void)snprintf(buf, sizeof buf, "%dm", m);
+		(void)strlcat(dst, buf, sizeof dst);
 	}
 	if (s) {
-		snprintf(buf, sizeof buf, "%ds", s);
-		strlcat(dst, buf, sizeof dst);
+		(void)snprintf(buf, sizeof buf, "%ds", s);
+		(void)strlcat(dst, buf, sizeof dst);
 	}
 
 	return (dst);
@@ -362,14 +364,15 @@ text_to_relayhost(struct relayhost *relay, const char *s)
 		{ "tls+auth://",	F_STARTTLS|F_AUTH		},
 		{ "secure://",		F_SMTPS|F_STARTTLS		},
 		{ "secure+auth://",	F_SMTPS|F_STARTTLS|F_AUTH	},
-		{ "backup://",		F_BACKUP       			}
+		{ "backup://",		F_BACKUP       			},
+		{ "tls+backup://",	F_BACKUP|F_STARTTLS    		}
 	};
 	const char     *errstr = NULL;
 	char	       *p, *q;
 	char		buffer[1024];
-	char	       *sep;
+	char	       *beg, *end;
 	size_t		i;
-	int		len;
+	size_t		len;
 
 	memset(buffer, 0, sizeof buffer);
 	if (strlcpy(buffer, s, sizeof buffer) >= sizeof buffer)
@@ -398,40 +401,53 @@ text_to_relayhost(struct relayhost *relay, const char *s)
 	if (relay->flags & F_LMTP)
 		relay->port = 0;
 
-	if ((sep = strrchr(p, ':')) != NULL) {
-		*sep = 0;
-		relay->port = strtonum(sep+1, 1, 0xffff, &errstr);
-		if (errstr)
-			return 0;
-		len = sep - p;
-	}
-	else
-		len = strlen(p);
-
-	if ((relay->flags & F_LMTP) && (relay->port == 0))
-		return 0;
-
-	relay->hostname[len] = 0;
-
-	q = strchr(p, '@');
-	if (q == NULL && relay->flags & F_AUTH)
-		return 0;
-	if (q && !(relay->flags & F_AUTH))
-		return 0;
-
-	if (q == NULL) {
-		if (strlcpy(relay->hostname, p, sizeof (relay->hostname))
-		    >= sizeof (relay->hostname))
-			return 0;
-	} else {
+	/* first, we extract the label if any */
+	if ((q = strchr(p, '@')) != NULL) {
 		*q = 0;
 		if (strlcpy(relay->authlabel, p, sizeof (relay->authlabel))
 		    >= sizeof (relay->authlabel))
 			return 0;
-		if (strlcpy(relay->hostname, q + 1, sizeof (relay->hostname))
-		    >= sizeof (relay->hostname))
+		p = q + 1;
+	}
+
+	/* then, we extract the mail exchanger */
+	beg = end = p;
+	if (*beg == '[') {
+		if ((end = strchr(beg, ']')) == NULL)
+			return 0;
+		/* skip ']', it has to be included in the relay hostname */
+		++end;
+		len = end - beg;
+	}
+	else {
+		for (end = beg; *end; ++end)
+			if (! isalnum((unsigned char)*end) &&
+			    *end != '_' && *end != '.' && *end != '-')
+				break;
+		len = end - beg;
+	}
+	if (len >= sizeof relay->hostname)
+		return 0;
+	for (i = 0; i < len; ++i)
+		relay->hostname[i] = beg[i];
+	relay->hostname[i] = 0;
+
+	/* finally, we extract the port */
+	p = beg + len;
+	if (*p == ':') {
+		relay->port = strtonum(p+1, 1, 0xffff, &errstr);
+		if (errstr)
 			return 0;
 	}
+
+	if (! valid_domainpart(relay->hostname))
+		return 0;
+	if ((relay->flags & F_LMTP) && (relay->port == 0))
+		return 0;
+	if (relay->authlabel[0] == '\0' && relay->flags & F_AUTH)
+		return 0;
+	if (relay->authlabel[0] != '\0' && !(relay->flags & F_AUTH))
+		return 0;
 	return 1;
 }
 
@@ -445,45 +461,48 @@ relayhost_to_text(const struct relayhost *relay)
 	memset(buf, 0, sizeof buf);
 	switch (relay->flags & mask) {
 	case F_SMTPS|F_STARTTLS|F_AUTH:
-		strlcat(buf, "secure+auth://", sizeof buf);
+		(void)strlcat(buf, "secure+auth://", sizeof buf);
 		break;
 	case F_SMTPS|F_STARTTLS:
-		strlcat(buf, "secure://", sizeof buf);
+		(void)strlcat(buf, "secure://", sizeof buf);
 		break;
 	case F_STARTTLS|F_AUTH:
-		strlcat(buf, "tls+auth://", sizeof buf);
+		(void)strlcat(buf, "tls+auth://", sizeof buf);
 		break;
 	case F_SMTPS|F_AUTH:
-		strlcat(buf, "smtps+auth://", sizeof buf);
+		(void)strlcat(buf, "smtps+auth://", sizeof buf);
 		break;
 	case F_STARTTLS:
-		strlcat(buf, "tls://", sizeof buf);
+		(void)strlcat(buf, "tls://", sizeof buf);
 		break;
 	case F_SMTPS:
-		strlcat(buf, "smtps://", sizeof buf);
+		(void)strlcat(buf, "smtps://", sizeof buf);
+		break;
+	case F_BACKUP|F_STARTTLS:
+		(void)strlcat(buf, "tls+backup://", sizeof buf);
 		break;
 	case F_BACKUP:
-		strlcat(buf, "backup://", sizeof buf);
+		(void)strlcat(buf, "backup://", sizeof buf);
 		break;
 	case F_TLS_OPTIONAL:
-		strlcat(buf, "smtp+tls://", sizeof buf);
+		(void)strlcat(buf, "smtp+tls://", sizeof buf);
 		break;
 	case F_LMTP:
-		strlcat(buf, "lmtp://", sizeof buf);
+		(void)strlcat(buf, "lmtp://", sizeof buf);
 		break;
 	default:
-		strlcat(buf, "smtp://", sizeof buf);
+		(void)strlcat(buf, "smtp://", sizeof buf);
 		break;
 	}
 	if (relay->authlabel[0]) {
-		strlcat(buf, relay->authlabel, sizeof buf);
-		strlcat(buf, "@", sizeof buf);
+		(void)strlcat(buf, relay->authlabel, sizeof buf);
+		(void)strlcat(buf, "@", sizeof buf);
 	}
-	strlcat(buf, relay->hostname, sizeof buf);
+	(void)strlcat(buf, relay->hostname, sizeof buf);
 	if (relay->port) {
-		strlcat(buf, ":", sizeof buf);
-		snprintf(port, sizeof port, "%d", relay->port);
-		strlcat(buf, port, sizeof buf);
+		(void)strlcat(buf, ":", sizeof buf);
+		(void)snprintf(port, sizeof port, "%d", relay->port);
+		(void)strlcat(buf, port, sizeof buf);
 	}
 	return buf;
 }
@@ -530,77 +549,77 @@ rule_to_text(struct rule *r)
 	static char buf[4096];
 
 	memset(buf, 0, sizeof buf);
-	strlcpy(buf, r->r_decision == R_ACCEPT  ? "accept" : "reject", sizeof buf);
+	(void)strlcpy(buf, r->r_decision == R_ACCEPT  ? "accept" : "reject", sizeof buf);
 	if (r->r_tag[0]) {
-		strlcat(buf, " tagged ", sizeof buf);
+		(void)strlcat(buf, " tagged ", sizeof buf);
 		if (r->r_nottag)
-			strlcat(buf, "! ", sizeof buf);
-		strlcat(buf, r->r_tag, sizeof buf);
+			(void)strlcat(buf, "! ", sizeof buf);
+		(void)strlcat(buf, r->r_tag, sizeof buf);
 	}
-	strlcat(buf, " from ", sizeof buf);
+	(void)strlcat(buf, " from ", sizeof buf);
 	if (r->r_notsources)
-		strlcat(buf, "! ", sizeof buf);
-	strlcat(buf, r->r_sources->t_name, sizeof buf);
+		(void)strlcat(buf, "! ", sizeof buf);
+	(void)strlcat(buf, r->r_sources->t_name, sizeof buf);
 
-	strlcat(buf, " for ", sizeof buf);
+	(void)strlcat(buf, " for ", sizeof buf);
 	if (r->r_notdestination)
-		strlcat(buf, "! ", sizeof buf);
+		(void)strlcat(buf, "! ", sizeof buf);
 	switch (r->r_desttype) {
 	case DEST_DOM:
 		if (r->r_destination == NULL) {
-			strlcat(buf, " any", sizeof buf);
+			(void)strlcat(buf, " any", sizeof buf);
 			break;
 		}
-		strlcat(buf, " domain ", sizeof buf);
-		strlcat(buf, r->r_destination->t_name, sizeof buf);
+		(void)strlcat(buf, " domain ", sizeof buf);
+		(void)strlcat(buf, r->r_destination->t_name, sizeof buf);
 		if (r->r_mapping) {
-			strlcat(buf, " alias ", sizeof buf);
-			strlcat(buf, r->r_mapping->t_name, sizeof buf);
+			(void)strlcat(buf, " alias ", sizeof buf);
+			(void)strlcat(buf, r->r_mapping->t_name, sizeof buf);
 		}
 		break;
 	case DEST_VDOM:
 		if (r->r_destination == NULL) {
-			strlcat(buf, " any virtual ", sizeof buf);
-			strlcat(buf, r->r_mapping->t_name, sizeof buf);
+			(void)strlcat(buf, " any virtual ", sizeof buf);
+			(void)strlcat(buf, r->r_mapping->t_name, sizeof buf);
 			break;
 		}
-		strlcat(buf, " domain ", sizeof buf);
-		strlcat(buf, r->r_destination->t_name, sizeof buf);
-		strlcat(buf, " virtual ", sizeof buf);
-		strlcat(buf, r->r_mapping->t_name, sizeof buf);
+		(void)strlcat(buf, " domain ", sizeof buf);
+		(void)strlcat(buf, r->r_destination->t_name, sizeof buf);
+		(void)strlcat(buf, " virtual ", sizeof buf);
+		(void)strlcat(buf, r->r_mapping->t_name, sizeof buf);
 		break;
 	}
 
 	switch (r->r_action) {
 	case A_RELAY:
-		strlcat(buf, " relay", sizeof buf);
+		(void)strlcat(buf, " relay", sizeof buf);
 		break;
 	case A_RELAYVIA:
-		strlcat(buf, " relay via ", sizeof buf);
-		strlcat(buf, relayhost_to_text(&r->r_value.relayhost), sizeof buf);
+		(void)strlcat(buf, " relay via ", sizeof buf);
+		(void)strlcat(buf, relayhost_to_text(&r->r_value.relayhost), sizeof buf);
 		break;
 	case A_MAILDIR:
-		strlcat(buf, " deliver to maildir \"", sizeof buf);
-		strlcat(buf, r->r_value.buffer, sizeof buf);
-		strlcat(buf, "\"", sizeof buf);
+		(void)strlcat(buf, " deliver to maildir \"", sizeof buf);
+		(void)strlcat(buf, r->r_value.buffer, sizeof buf);
+		(void)strlcat(buf, "\"", sizeof buf);
 		break;
 	case A_MBOX:
-		strlcat(buf, " deliver to mbox", sizeof buf);
+		(void)strlcat(buf, " deliver to mbox", sizeof buf);
 		break;
 	case A_FILENAME:
-		strlcat(buf, " deliver to filename \"", sizeof buf);
-		strlcat(buf, r->r_value.buffer, sizeof buf);
-		strlcat(buf, "\"", sizeof buf);
+		(void)strlcat(buf, " deliver to filename \"", sizeof buf);
+		(void)strlcat(buf, r->r_value.buffer, sizeof buf);
+		(void)strlcat(buf, "\"", sizeof buf);
 		break;
 	case A_MDA:
-		strlcat(buf, " deliver to mda \"", sizeof buf);
-		strlcat(buf, r->r_value.buffer, sizeof buf);
-		strlcat(buf, "\"", sizeof buf);
+		(void)strlcat(buf, " deliver to mda \"", sizeof buf);
+		(void)strlcat(buf, r->r_value.buffer, sizeof buf);
+		(void)strlcat(buf, "\"", sizeof buf);
 		break;
 	case A_LMTP:
-		strlcat(buf, " deliver to lmtp \"", sizeof buf);
-		strlcat(buf, r->r_value.buffer, sizeof buf);
-		strlcat(buf, "\"", sizeof buf);
+		(void)strlcat(buf, " deliver to lmtp \"", sizeof buf);
+		(void)strlcat(buf, r->r_value.buffer, sizeof buf);
+		(void)strlcat(buf, "\"", sizeof buf);
 		break;
 	case A_NONE:
 		break;
@@ -612,7 +631,7 @@ rule_to_text(struct rule *r)
 int
 text_to_userinfo(struct userinfo *userinfo, const char *s)
 {
-	char		buf[SMTPD_MAXPATHLEN];
+	char		buf[PATH_MAX];
 	char	       *p;
 	const char     *errstr;
 
@@ -661,7 +680,7 @@ int
 text_to_credentials(struct credentials *creds, const char *s)
 {
 	char   *p;
-	char	buffer[SMTPD_MAXLINESIZE];
+	char	buffer[LINE_MAX];
 	size_t	offset;
 
 	p = strchr(s, ':');
@@ -702,6 +721,7 @@ text_to_expandnode(struct expandnode *expandnode, const char *s)
 	    alias_is_filter(expandnode, s, l) ||
 	    alias_is_filename(expandnode, s, l) ||
 	    alias_is_address(expandnode, s, l) ||
+	    alias_is_maildir(expandnode, s, l) ||
 	    alias_is_username(expandnode, s, l))
 		return (1);
 
@@ -716,6 +736,7 @@ expandnode_to_text(struct expandnode *expandnode)
 	case EXPAND_FILENAME:
 	case EXPAND_INCLUDE:
 	case EXPAND_ERROR:
+	case EXPAND_MAILDIR:
 		return expandnode->u.buffer;
 	case EXPAND_USERNAME:
 		return expandnode->u.user;
@@ -728,6 +749,21 @@ expandnode_to_text(struct expandnode *expandnode)
 	return NULL;
 }
 
+static int
+alias_is_maildir(struct expandnode *alias, const char *line, size_t len)
+{
+	if (strncasecmp("maildir:", line, 8) != 0)
+		return (0);
+
+	line += 8;
+	memset(alias, 0, sizeof *alias);
+	alias->type = EXPAND_MAILDIR;
+	if (strlcpy(alias->u.buffer, line,
+	    sizeof(alias->u.buffer)) >= sizeof(alias->u.buffer))
+		return (0);
+
+	return (1);
+}
 
 /******/
 static int
@@ -766,7 +802,7 @@ alias_is_username(struct expandnode *alias, const char *line, size_t len)
 
 	while (*line) {
 		if (!isalnum((unsigned char)*line) &&
-		    *line != '_' && *line != '.' && *line != '-')
+		    *line != '_' && *line != '.' && *line != '-' && *line != '+')
 			return 0;
 		++line;
 	}
@@ -795,8 +831,8 @@ alias_is_address(struct expandnode *alias, const char *line, size_t len)
 
 	/* scan pre @ for disallowed chars */
 	*domain++ = '\0';
-	strlcpy(alias->u.mailaddr.user, line, sizeof(alias->u.mailaddr.user));
-	strlcpy(alias->u.mailaddr.domain, domain,
+	(void)strlcpy(alias->u.mailaddr.user, line, sizeof(alias->u.mailaddr.user));
+	(void)strlcpy(alias->u.mailaddr.domain, domain,
 	    sizeof(alias->u.mailaddr.domain));
 
 	while (*line) {
